//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "boids.h"

static
void
detectFirstRock(const SpatialBasis *location,
                real64_t range,
                Rock **rockOut,
                real64_t *distanceOut,
                real64_t *yawAngleOut,
                real64_t *pitchAngleOut)
{
NearestVisible nv;
NearestVisible_start(&nv,location,range,M_2PI,M_PI);
real64_t px,py,pz;
SpatialBasis_getPosition(location,&px,&py,&pz);
for(intptr_t l=0,l_last=SpatialGrid_discreteDistance(&simu.rockGrid,range);
    l<=l_last;++l)
  {
  SpatialGridLayer sgl;
  for(intptr_t key=SpatialGridLayer_start(&sgl,&simu.rockGrid,px,py,pz,l);
      key>=0;key=SpatialGridLayer_next(&sgl))
    {
    Cell *cell=simu.rockCells+key;
    PtrVector *cellContent=&cell->content;
    for(intptr_t i=0,i_end=PtrVector_size(cellContent);i<i_end;++i)
      {
      Rock *candidate=(Rock *)PtrVector_get(cellContent,i);
      NearestVisible_add(&nv,&candidate->location,candidate);
      }
    }
  *rockOut=(Rock *)NearestVisible_get(&nv,distanceOut,
                                      yawAngleOut,pitchAngleOut);
  if(*rockOut) break;
  }
}

static
bool // success
detectFirstShark(const SpatialBasis *location,
                 Activity *act,
                 real64_t range,
                 real64_t yawVision,
                 Shark **sharkOut,
                 real64_t *distanceOut,
                 real64_t *yawAngleOut,
                 real64_t *pitchAngleOut)
{
const int cpuCtx=Activity_cpuCtx(act);
#if !defined ASYNC_BOIDS
  int fromState=Scheduler_fromState(&simu.sched);
#endif
NearestVisible nv;
NearestVisible_start(&nv,location,range,yawVision,M_PI);
real64_t px,py,pz;
SpatialBasis_getPosition(location,&px,&py,&pz);
for(intptr_t l=0,l_last=SpatialGrid_discreteDistance(&simu.sharkGrid,range);
    l<=l_last;++l)
  {
  SpatialGridLayer sgl;
  for(intptr_t key=SpatialGridLayer_start(&sgl,&simu.sharkGrid,px,py,pz,l);
      key>=0;key=SpatialGridLayer_next(&sgl))
    {
#if defined ASYNC_BOIDS
    Cell *cell=simu.sharkCells+key;
#else
    Cell *cell=simu.sharkCells[fromState]+key;
#endif
    PtrVector *cellContent=&cell->content;
    intptr_t contentSize=PtrVector_size(cellContent);
    if(contentSize)
      {
#if defined ASYNC_BOIDS
      if(!SpinSynchro_tryLockR(&cell->sync))
        {
        return false;
        }
#endif
      for(intptr_t i=0;i<contentSize;++i)
        {
        Shark *candidate=(Shark *)PtrVector_get(cellContent,i);
#if defined ASYNC_BOIDS
        const SpatialBasis *sharkLocation=&candidate->location;
#else
        const SpatialBasis *sharkLocation=&candidate->location[fromState];
#endif
        if(sharkLocation==location)
          {
          continue;
          }
        if(simu.neighbourAffinity)
          {
          const int candidateCpu=Activity_cpuCtx(&candidate->act);
          Activity_adjustAffinity(act,&simu.sched,candidateCpu,
                                  Platform_cpuProximity(cpuCtx,candidateCpu));
          }
#if defined ASYNC_BOIDS
        if(!SpinSynchro_tryLockR(&candidate->sync))
          {
          SpinSynchro_unlockR(&cell->sync);
          return false;
          }
#endif
        NearestVisible_add(&nv,sharkLocation,candidate);
#if defined ASYNC_BOIDS
        SpinSynchro_unlockR(&candidate->sync);
#endif
        }
#if defined ASYNC_BOIDS
      SpinSynchro_unlockR(&cell->sync);
#endif
      }
    }
  *sharkOut=(Shark *)NearestVisible_get(&nv,distanceOut,
                                        yawAngleOut,pitchAngleOut);
  if(*sharkOut) break;
  }
return true;
}

static
bool // success
collectFish(const SpatialBasis *location,
            SpatialGrid *sg,
            Cell *cells,
            real64_t range,
            PtrVector *contentOut)
{
PtrVector_clear(contentOut);
real64_t px,py,pz;
SpatialBasis_getPosition(location,&px,&py,&pz);
SpatialGridNeighbourhood sgn;
for(intptr_t key=SpatialGridNeighbourhood_start(&sgn,sg,px,py,pz,range);
    key>=0;key=SpatialGridNeighbourhood_next(&sgn))
  {
  Cell *cell=cells+key;
  PtrVector *cellContent=&cell->content;
  intptr_t contentSize=PtrVector_size(cellContent);
  if(contentSize)
    {
#if defined ASYNC_BOIDS
    if(!SpinSynchro_tryLockR(&cell->sync))
      {
      return false;
      }
#endif
    PtrVector_pushMany(contentOut,PtrVector_data(cellContent),contentSize);
#if defined ASYNC_BOIDS
    SpinSynchro_unlockR(&cell->sync);
#endif
    }
  }
return true;
}

//----------------------------------------------------------------------------

Rock *
Rock_new(int school)
{
Rock *rock=Memory_alloc(1,Rock);
SpatialBasis_init(&rock->location);
real64_t x=R64Vector_get(&simu.schools,3*school);
real64_t y=R64Vector_get(&simu.schools,3*school+1);
real64_t z=R64Vector_get(&simu.schools,3*school+2);
int cpuCtx=Platform_cpuCtx();
x+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
y+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
z+=0.5*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5)+0.1*LIMIT_SIZE;
SpatialBasis_setPosition(&rock->location,x,y,z);
rock->key=SpatialGrid_coordToKey(&simu.rockGrid,x,y,z);
Cell_add(simu.rockCells+rock->key,rock);
PtrVector_push(&simu.rocks,rock);
return rock;
}

//----------------------------------------------------------------------------

static
void
Shark_delete(Shark *s)
{
Activity_destroy(&s->act,&simu.sched);
#if defined ASYNC_BOIDS
  SpatialBasis_destroy(&s->location);
  Cell_remove(simu.sharkCells+s->key,s);
  SpinSynchro_destroy(&s->sync);
#else
  SpatialBasis_destroy(&s->location[0]);
  SpatialBasis_destroy(&s->location[1]);
  if(s->key[0]>=0)
    {
    Cell_remove(simu.sharkCells[0]+s->key[0],s);
    }
  if(s->key[1]>=0)
    {
    Cell_remove(simu.sharkCells[1]+s->key[1],s);
    }
#endif
PtrVector_destroy(&s->memo);
SpinSynchro_lockW(&simu.sharksSync);
PtrVector_fastErase(&simu.sharks,
                    PtrVector_find(&simu.sharks,
                                   (PtrVector_CmpFnct)0,s));
SpinSynchro_unlockW(&simu.sharksSync);
Memory_free(s);
}

static
Shark * // clone or 0
Shark_numaMigration(Shark *s)
{
Shark *clone=(Shark *)0;
int cpuCtx=Activity_cpuCtx(&s->act);
if(simu.numaMigration&&!Random_int(cpuCtx,simu.numaMigration))
  {
  intptr_t *numaUsage=Scheduler_initNumaUsage(&simu.sched,cpuCtx);
  Memory_numaUsage(s,numaUsage);
  PtrVector_numaUsage(&s->memo,numaUsage);
  if(Memory_numaBalance(Platform_memNumaCtx(cpuCtx),numaUsage)<0.5)
    {
    clone=Memory_alloc(1,Shark);
    clone->dead=s->dead;
    clone->school=s->school;
    clone->memoCount=s->memoCount;
    PtrVector_init(&clone->memo);
    Activity_initCopy(&clone->act,&simu.sched,&s->act);
#if defined ASYNC_BOIDS
    clone->location=s->location; // bitwise copy is correct here
    SpinSynchro_init(&clone->sync);
    clone->key=s->key;
    Cell_add(simu.sharkCells+clone->key,clone);
#else
    clone->location[0]=s->location[0]; // bitwise copy is correct here
    clone->location[1]=s->location[1]; // bitwise copy is correct here
    int fromState=Scheduler_fromState(&simu.sched);
    clone->key[fromState]=-1;
    int toState=Scheduler_toState(&simu.sched);
    clone->key[toState]=s->key[toState];
    Cell_add(simu.sharkCells[toState]+clone->key[toState],clone);
#endif
    SpinSynchro_lockW(&simu.sharksSync);
    PtrVector_push(&simu.sharks,clone);
    SpinSynchro_unlockW(&simu.sharksSync);
    Shark_delete(s);
    }
  }
return clone;
}

static
ActivityResult
Shark_activate(Activity *a,
               Scheduler *s)
{
const int cpuCtx=Activity_cpuCtx(a);
Shark *shark=STRUCT_OF_MEMBER(a,Shark,act);
#if defined ASYNC_BOIDS
  if(!SpinSynchro_tryLockW(&shark->sync))
    {
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
#endif
if(shark->dead)
  {
  Shark_delete(shark);
  return ACT_DEAD;
  }
real64_t yaw1=0.0,pitch1=0.0,wy1=0.0,wp1=0.0;
real64_t yaw2=0.0,pitch2=0.0,wy2=0.0,wp2=0.0;
real64_t yaw3=0.0,pitch3=0.0,wy3=0.0,wp3=0.0;
real64_t yaw4=0.0,pitch4=0.0,wy4=0.0,wp4=0.0;
real64_t distance=0.0;
#if defined ASYNC_BOIDS
  SpatialBasis *sharkLocation=&shark->location;
#else
  int fromState=Scheduler_fromState(&simu.sched);
  SpatialBasis *sharkLocation=&shark->location[fromState];
#endif
real64_t px,py,pz;
SpatialBasis_getPosition(sharkLocation,&px,&py,&pz);
real64_t rx,ry,rz;
SpatialBasis_extractZyxOrientation(sharkLocation,&rx,&ry,&rz);
//-- avoid ground --
if(ry>0.0)
  {
  distance=pz/sin(ry);
  pitch1=-M_PI_2;
  }
else
  {
  distance=(pz-0.5*LIMIT_SIZE)/sin(ry); // div by 0 ?
  pitch1=M_PI_2;
  }
if(distance<SHARK_SIZE)
  {
  wp1=20.0;
  }
//-- avoid rocks --
Rock *rock;
detectFirstRock(sharkLocation,5.0*ROCK_SIZE,
                &rock,&distance,&yaw2,&pitch2);
if(rock)
  {
  yaw2=yaw2<0.0 ? yaw2+M_PI_2 : yaw2-M_PI_2;
  pitch2=pitch2<0.0 ? pitch2+M_PI_2 : pitch2-M_PI_2;
  wy2=wp2=10.0;
  }
//-- avoid neighbours --
Shark *other;
if(!detectFirstShark(sharkLocation,&shark->act,SHARK_SIZE,M_PI,
                     &other,&distance,&yaw3,&pitch3))
  {
#if defined ASYNC_BOIDS
  SpinSynchro_unlockW(&shark->sync);
  return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
#endif
  }
if(other)
  {
  yaw3=yaw3<0.0 ? yaw3+M_PI_2 : yaw3-M_PI_2;
  pitch3=pitch3<0.0 ? pitch3+M_PI_2 : pitch3-M_PI_2;
  wy3=wp3=SHARK_SIZE/distance;
  }
//-- purchase fish --
if(shark->memoCount&&PtrVector_size(&shark->memo))
  {
  --shark->memoCount;
  }
else
  {
  if(!collectFish(sharkLocation,&simu.fineFishGrid,
#if defined ASYNC_BOIDS
                  simu.fineFishCells,
#else
                  simu.fineFishCells[fromState],
#endif
                  FISH_RANGE_MIN,&shark->memo))
    {
#if defined ASYNC_BOIDS
    SpinSynchro_unlockW(&shark->sync);
#endif
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
  if(!PtrVector_size(&shark->memo))
    {
    if(!collectFish(sharkLocation,&simu.coarseFishGrid,
#if defined ASYNC_BOIDS
                    simu.coarseFishCells,
#else
                    simu.coarseFishCells[fromState],
#endif
                    SHARK_RANGE,&shark->memo))
      {
#if defined ASYNC_BOIDS
      SpinSynchro_unlockW(&shark->sync);
#endif
      return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
      }
    }
  shark->memoCount=simu.memo;
  }
bool purchasing=false;
real64_t posX=0.0,posY=0.0,posZ=0.0;
intptr_t candidateCount=PtrVector_size(&shark->memo);
if(candidateCount)
  {
  real64_t r2=SHARK_RANGE*SHARK_RANGE;
  for(intptr_t i=0;i<candidateCount;++i)
    {
    Fish *fish=(Fish *)PtrVector_get(&shark->memo,i);
    if(simu.neighbourAffinity)
      {
      const int fishCpu=Activity_cpuCtx(&fish->act);
      Activity_adjustAffinity(&shark->act,s,fishCpu,
                              Platform_cpuProximity(cpuCtx,fishCpu));
      }
#if defined ASYNC_BOIDS
    if(!SpinSynchro_tryLockR(&fish->sync))
      {
      SpinSynchro_unlockW(&shark->sync);
      return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
      }
    const SpatialBasis *fishLocation=&fish->location;
#else
    const SpatialBasis *fishLocation=&fish->location[fromState];
#endif
    real64_t x,y,z;
    SpatialBasis_getPosition(fishLocation,&x,&y,&z);
    SpatialBasis_globalToLocalPosition(sharkLocation,&x,&y,&z);
    real64_t d=x*x+y*y+z*z;
    if(d<=r2)
      {
      real64_t dxy=sqrt(d-z*z);
      real64_t yaw=dxy ? atan2(y,x) : 0.0;
      if(fabs(yaw)<=0.4*M_PI)
        {
        posX+=x; posY+=y; posZ+=z;
        purchasing=true;
        }
      }
#if defined ASYNC_BOIDS
    SpinSynchro_unlockR(&fish->sync);
#endif
    }
  }
if(purchasing)
  {
  cart3DToAngles64(posX,posY,posZ,&yaw4,&pitch4);
  wy4=wp4=1.0;
  }
else //-- stay around --
  {
  real64_t x,y,z;
  x=R64Vector_get(&simu.schools,3*shark->school);
  y=R64Vector_get(&simu.schools,3*shark->school+1);
  z=R64Vector_get(&simu.schools,3*shark->school+2);
  SpatialBasis_globalToLocalPosition(sharkLocation,&x,&y,&z);
  real64_t p;
  cart3DToSpherical64(x,y,z,&distance,&yaw1,&p);
  wy1=distance;
  }
//-- compute final direction and move --
#if !defined ASYNC_BOIDS
  int toState=Scheduler_toState(&simu.sched);
  shark->location[toState]=*sharkLocation; // bitwise copy is correct here
  sharkLocation=&shark->location[toState];
#endif
real64_t sum=wy1+wy2+wy3+wy4;
if(sum)
  {
  SpatialBasis_yaw(sharkLocation,
                   simu.dt*(yaw1*wy1+yaw2*wy2+
                            yaw3*wy3+yaw4*wy4)/sum);
  }
sum=wp1+wp2+wp3+wp4;
if(sum)
  {
  SpatialBasis_pitch(sharkLocation,
                     simu.dt*(pitch1*wp1+pitch2*wp2+
                              pitch3*wp3+pitch4*wp4)/sum);
  }
SpatialBasis_roll(sharkLocation,-simu.dt*rx);
SpatialBasis_translate(sharkLocation,
                       simu.dt*SHARK_SPEED,0.0,0.0);
SpatialBasis_getPosition(sharkLocation,&px,&py,&pz);
intptr_t key=SpatialGrid_coordToKey(&simu.sharkGrid,px,py,pz);
#if defined ASYNC_BOIDS
  if(key!=shark->key)
    {
    Cell_add(simu.sharkCells+key,shark);
    Cell_remove(simu.sharkCells+shark->key,shark);
    shark->key=key;
    }
#else
  if(key!=shark->key[toState])
    {
    Cell_add(simu.sharkCells[toState]+key,shark);
    if(shark->key[toState]>=0)
      {
      Cell_remove(simu.sharkCells[toState]+shark->key[toState],shark);
      }
    shark->key[toState]=key;
    }
#endif
// if(Shark_numaMigration(shark)) return ACT_DEAD;
#if defined ASYNC_BOIDS
  SpinSynchro_unlockW(&shark->sync);
#endif
return ACT_DONE;
}

Shark *
Shark_new(int school)
{
int cpuCtx=Platform_cpuCtx();
Shark *shark=Memory_alloc(1,Shark);
shark->dead=false;
shark->school=school;
shark->memoCount=simu.memo ? Random_int(cpuCtx,simu.memo) : 0;
PtrVector_init(&shark->memo);
#if defined ASYNC_BOIDS
  SpatialBasis *location=&shark->location;
#else
  int toState=Scheduler_toState(&simu.sched);
  SpatialBasis *location=&shark->location[toState];
#endif
SpatialBasis_init(location);
real64_t x=R64Vector_get(&simu.schools,3*school);
real64_t y=R64Vector_get(&simu.schools,3*school+1);
real64_t z=R64Vector_get(&simu.schools,3*school+2);
x+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
y+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
z+=0.5*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5)+0.5*LIMIT_SIZE;
SpatialBasis_setPosition(location,x,y,z);
SpatialBasis_yaw(location,2.0*M_PI*(Random_real64(cpuCtx)-0.5));
SpatialBasis_pitch(location,M_PI*(Random_real64(cpuCtx)-0.5));
#if defined ASYNC_BOIDS
  SpinSynchro_init(&shark->sync);
  shark->key=SpatialGrid_coordToKey(&simu.sharkGrid,x,y,z);
  Cell_add(simu.sharkCells+shark->key,shark);
#else
  int fromState=Scheduler_fromState(&simu.sched);
  shark->location[fromState]=*location; // bitwise copy is correct here
  shark->key[fromState]=-1;
  shark->key[toState]=SpatialGrid_coordToKey(&simu.sharkGrid,x,y,z);
  Cell_add(simu.sharkCells[toState]+shark->key[toState],shark);
#endif
Activity_init(&shark->act,&Shark_activate,&simu.sched,cpuCtx);
SpinSynchro_lockW(&simu.sharksSync);
PtrVector_push(&simu.sharks,shark);
SpinSynchro_unlockW(&simu.sharksSync);
return shark;
}

//----------------------------------------------------------------------------

static
void
Fish_delete(Fish *f)
{
Activity_destroy(&f->act,&simu.sched);
#if defined ASYNC_BOIDS
  SpatialBasis_destroy(&f->location);
  Cell_remove(simu.fineFishCells+f->fineKey,f);
  Cell_remove(simu.coarseFishCells+f->coarseKey,f);
  SpinSynchro_destroy(&f->sync);
#else
  SpatialBasis_destroy(&f->location[0]);
  SpatialBasis_destroy(&f->location[1]);
  if(f->fineKey[0]>=0)
    {
    Cell_remove(simu.fineFishCells[0]+f->fineKey[0],f);
    }
  if(f->fineKey[1]>=0)
    {
    Cell_remove(simu.fineFishCells[1]+f->fineKey[1],f);
    }
  if(f->coarseKey[0]>=0)
    {
    Cell_remove(simu.coarseFishCells[0]+f->coarseKey[0],f);
    }
  if(f->coarseKey[1]>=0)
    {
    Cell_remove(simu.coarseFishCells[1]+f->coarseKey[1],f);
    }
#endif
PtrVector_destroy(&f->memo);
SpinSynchro_lockW(&simu.fishSync);
PtrVector_fastErase(&simu.fish,
                    PtrVector_find(&simu.fish,
                                   (PtrVector_CmpFnct)0,f));
SpinSynchro_unlockW(&simu.fishSync);
Memory_free(f);
}

static
Fish * // clone or 0
Fish_numaMigration(Fish *f)
{
Fish *clone=(Fish *)0;
int cpuCtx=Activity_cpuCtx(&f->act);
if(simu.numaMigration&&!Random_int(cpuCtx,simu.numaMigration))
  {
  intptr_t *numaUsage=Scheduler_initNumaUsage(&simu.sched,cpuCtx);
  Memory_numaUsage(f,numaUsage);
  PtrVector_numaUsage(&f->memo,numaUsage);
  if(Memory_numaBalance(Platform_memNumaCtx(cpuCtx),numaUsage)<0.5)
    {
    clone=Memory_alloc(1,Fish);
    clone->dead=f->dead;
    clone->speed=f->speed;
    clone->range=f->range;
    clone->school=f->school;
    clone->memoCount=f->memoCount;
    PtrVector_init(&clone->memo);
    Activity_initCopy(&clone->act,&simu.sched,&f->act);
#if defined ASYNC_BOIDS
    clone->location=f->location; // bitwise copy is correct here
    SpinSynchro_init(&clone->sync);
    clone->fineKey=f->fineKey;
    Cell_add(simu.fineFishCells+clone->fineKey,clone);
    clone->coarseKey=f->coarseKey;
    Cell_add(simu.coarseFishCells+clone->coarseKey,clone);
#else
    clone->location[0]=f->location[0]; // bitwise copy is correct here
    clone->location[1]=f->location[1]; // bitwise copy is correct here
    int fromState=Scheduler_fromState(&simu.sched);
    clone->fineKey[fromState]=-1;
    clone->coarseKey[fromState]=-1;
    int toState=Scheduler_toState(&simu.sched);
    clone->fineKey[toState]=f->fineKey[toState];
    Cell_add(simu.fineFishCells[toState]+clone->fineKey[toState],clone);
    clone->coarseKey[toState]=f->coarseKey[toState];
    Cell_add(simu.coarseFishCells[toState]+clone->coarseKey[toState],clone);
#endif
    SpinSynchro_lockW(&simu.fishSync);
    PtrVector_push(&simu.fish,clone);
    SpinSynchro_unlockW(&simu.fishSync);
    Fish_delete(f);
    }
  }
return clone;
}

static
ActivityResult
Fish_activate(Activity *a,
              Scheduler *s)
{
const int cpuCtx=Activity_cpuCtx(a);
Fish *fish=STRUCT_OF_MEMBER(a,Fish,act);
#if defined ASYNC_BOIDS
  if(!SpinSynchro_tryLockW(&fish->sync))
    {
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
#endif
if(fish->dead)
  {
  Fish_delete(fish);
  return ACT_DEAD;
  }
#if !defined AVOID_RENDERING
  Rendering_showFish(fish);
#endif
real64_t yaw1=0.0,pitch1=0.0,s1=0.0,wy1=0.0,wp1=0.0,ws1=0.0;
real64_t yaw2=0.0,pitch2=0.0,s2=0.0,wy2=0.0,wp2=0.0,ws2=0.0;
real64_t yaw3=0.0,pitch3=0.0,s3=0.0,wy3=0.0,wp3=0.0,ws3=0.0;
real64_t yaw4=0.0,pitch4=0.0,s4=0.0,wy4=0.0,wp4=0.0,ws4=0.0;
real64_t yaw5=0.0,pitch5=0.0,s5=0.0,wy5=0.0,wp5=0.0,ws5=0.0;
real64_t yaw6=0.0,pitch6=0.0,s6=0.0,wy6=0.0,wp6=0.0,ws6=0.0;
real64_t yaw7=0.0,pitch7=0.0,s7=0.0,wy7=0.0,wp7=0.0,ws7=0.0;
real64_t distance=0.0;
#if defined ASYNC_BOIDS
  SpatialBasis *fishLocation=&fish->location;
#else
  int fromState=Scheduler_fromState(&simu.sched);
  SpatialBasis *fishLocation=&fish->location[fromState];
#endif
real64_t px,py,pz;
SpatialBasis_getPosition(fishLocation,&px,&py,&pz);
real64_t rx,ry,rz;
SpatialBasis_extractZyxOrientation(fishLocation,&rx,&ry,&rz);
//-- avoid ground --
if(ry>0.0)
  {
  distance=pz/sin(ry);
  pitch1=-M_PI_2;
  s1=0.5*FISH_SPEED_MIN;
  wp1=ws1=distance<4.0*FISH_SIZE ? 100.0 : 2.0/distance;
  }
//-- avoid rocks --
Rock *rock;
detectFirstRock(fishLocation,2.0*ROCK_SIZE,
                &rock,&distance,&yaw2,&pitch2);
if(rock)
  {
  yaw2=yaw2<0.0 ? yaw2+M_PI_2 : yaw2-M_PI_2;
  pitch2=pitch2<0.0 ? pitch2+M_PI_2 : pitch2-M_PI_2;
  s2=0.5*FISH_SPEED_MIN;
  wy2=wp2=ws2=10.0;
  }
//-- deal with neighbours --
if(fish->memoCount&&PtrVector_size(&fish->memo))
  {
  --fish->memoCount;
  }
else
  {
  if(!collectFish(fishLocation,&simu.fineFishGrid,
#if defined ASYNC_BOIDS
                  simu.fineFishCells,
#else
                  simu.fineFishCells[fromState],
#endif
                  fish->range,&fish->memo))
    {
#if defined ASYNC_BOIDS
    SpinSynchro_unlockW(&fish->sync);
#endif
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
  fish->memoCount=simu.memo;
  }
bool following=false;
int seenCount=0;
intptr_t candidateCount=PtrVector_size(&fish->memo);
if(candidateCount)
  {
  real64_t speed=0.0;
  real64_t posX=0.0,posY=0.0,posZ=0.0;
  real64_t dirX=0.0,dirY=0.0,dirZ=0.0;
  real64_t r2=fish->range*fish->range;
  real64_t minD=FISH_SIZE*FISH_SIZE;
  for(intptr_t i=0;i<candidateCount;++i)
    {
    Fish *other=(Fish *)PtrVector_get(&fish->memo,i);
    if(other==fish)
      {
      continue;
      }
    if(simu.neighbourAffinity)
      {
      const int otherCpu=Activity_cpuCtx(&other->act);
      Activity_adjustAffinity(&fish->act,s,otherCpu,
                              Platform_cpuProximity(cpuCtx,otherCpu));
      }
    if(other->school!=fish->school)
      {
      continue;
      }
#if defined ASYNC_BOIDS
    if(!SpinSynchro_tryLockR(&other->sync))
      {
      SpinSynchro_unlockW(&fish->sync);
      return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
      }
    const SpatialBasis *otherLocation=&other->location;
#else
    const SpatialBasis *otherLocation=&other->location[fromState];
#endif
    real64_t x,y,z;
    SpatialBasis_getPosition(otherLocation,&x,&y,&z);
    SpatialBasis_globalToLocalPosition(fishLocation,&x,&y,&z);
    real64_t d=x*x+y*y+z*z;
    if(d<=r2)
      {
      real64_t dxy=sqrt(d-z*z);
      real64_t yaw=atan2(y,x);
      if(fabs(yaw)<=0.4*M_2PI)
        {
        ++seenCount;
        speed+=other->speed;
        posX+=x; posY+=y; posZ+=z;
        real64_t dx,dy,dz;
        SpatialBasis_getLocalX(otherLocation,&dx,&dy,&dz);
        dirX+=dx; dirY+=dy; dirZ+=dz;
        if(x>0.0)
          {
          real64_t pitch=-atan2(z,dxy);
          if((fabs(yaw)<M_PI_4)&&(fabs(pitch)<M_PI_4))
            {
            following=true;
            }
          if(d<minD)
            {
            minD=d;
            yaw3=yaw<0.0 ? yaw+M_PI_2 : yaw-M_PI_2;
            pitch=pitch<0.0 ? pitch+M_PI_2 : pitch-M_PI_2;
            wy3=wp3=FISH_SIZE/d;
            }
          }
        }
      }
#if defined ASYNC_BOIDS
    SpinSynchro_unlockR(&other->sync);
#endif
    }
  if(seenCount)
    {
    // ensure alignment
    real64_t div=1.0/seenCount;
    dirX*=div; dirY*=div; dirZ*=div;
    SpatialBasis_globalToLocalVector(fishLocation,&dirX,&dirY,&dirZ);
    cart3DToAngles64(dirX,dirY,dirZ,&yaw4,&pitch4);
    s4=speed*div;
    wy4=wp4=ws4=1.0;
    // ensure cohesion
    cart3DToAngles64(posX,posY,posZ,&yaw5,&pitch5);
    wy5=wp5=1.0;
    }
  }
if(seenCount<2)
  {
  fish->range*=2.0;
  if(fish->range>FISH_RANGE_MAX)
    {
    fish->range=FISH_RANGE_MAX;
    }
  fish->memoCount=0;
  }
else if(seenCount>8)
  {
  fish->range*=0.5;
  if(fish->range<FISH_RANGE_MIN)
    {
    fish->range=FISH_RANGE_MIN;
    }
  fish->memoCount=0;
  }
//-- escape from sharks --
Shark *shark;
if(!detectFirstShark(fishLocation,&fish->act,SHARK_SIZE,M_2PI,
                     &shark,&distance,&yaw7,&pitch7))
  {
#if defined ASYNC_BOIDS
  SpinSynchro_unlockW(&fish->sync);
  return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
#endif
  }
bool purchased=false;
if(shark)
  {
  yaw7=yaw7<0.0 ? yaw7+M_PI_2 : yaw7-M_PI_2;
  pitch7=pitch3<0.0 ? pitch7+M_PI_2 : pitch7-M_PI_2;
  s7=FISH_SPEED_MAX;
  wy7=wp7=ws7=SHARK_SIZE/distance;
  }
else
  {
  s7=FISH_SPEED_MIN;
  ws7=0.5;
  }
//-- stay around --
if(!following&&!purchased)
  {
  real64_t x,y,z;
  x=R64Vector_get(&simu.schools,3*fish->school);
  y=R64Vector_get(&simu.schools,3*fish->school+1);
  z=R64Vector_get(&simu.schools,3*fish->school+2);
  SpatialBasis_globalToLocalPosition(fishLocation,&x,&y,&z);
  cart3DToSpherical64(x,y,z,&distance,&yaw6,&pitch6);
  wy6=wp6=distance;
  }
//-- compute final direction and move --
#if !defined ASYNC_BOIDS
  int toState=Scheduler_toState(&simu.sched);
  fish->location[toState]=*fishLocation; // bitwise copy is correct here
  fishLocation=&fish->location[toState];
#endif
real64_t sum=wy1+wy2+wy3+wy4+wy5+wy6+wy7;
if(sum)
  {
  SpatialBasis_yaw(fishLocation,
                   simu.dt*(yaw1*wy1+yaw2*wy2+yaw3*wy3+yaw4*wy4+
                            yaw5*wy5+yaw6*wy6+yaw7*wy7)/sum);
  }
sum=wp1+wp2+wp3+wp4+wp5+wp6+wy7;
if(sum)
  {
  SpatialBasis_pitch(fishLocation,
                     simu.dt*(pitch1*wp1+pitch2*wp2+pitch3*wp3+
                              pitch4*wp4+pitch5*wp5+pitch6*wp6+
                              pitch7*wp7)/sum);
  }
sum=ws1+ws2+ws3+ws4+ws5+ws6+ws7;
if(sum)
  {
  fish->speed+=simu.dt*(((s1*ws1+s2*ws2+s3*ws3+s4*ws4+s5*ws5+
                          s6*ws6+s7*ws7)/sum)-fish->speed);
  if(fish->speed>FISH_SPEED_MAX)
    {
    fish->speed=FISH_SPEED_MAX;
    }
  else if(fish->speed<FISH_SPEED_MIN)
    {
    fish->speed=FISH_SPEED_MIN;
    }
  }
SpatialBasis_roll(fishLocation,-simu.dt*rx);
SpatialBasis_translate(fishLocation,
                       simu.dt*fish->speed,0.0,0.0);
SpatialBasis_getPosition(fishLocation,&px,&py,&pz);
intptr_t fineKey=SpatialGrid_coordToKey(&simu.fineFishGrid,px,py,pz);
intptr_t coarseKey=SpatialGrid_coordToKey(&simu.coarseFishGrid,px,py,pz);
#if defined ASYNC_BOIDS
  if(fineKey!=fish->fineKey)
    {
    Cell_add(simu.fineFishCells+fineKey,fish);
    Cell_remove(simu.fineFishCells+fish->fineKey,fish);
    fish->fineKey=fineKey;
    }
  if(coarseKey!=fish->coarseKey)
    {
    Cell_add(simu.coarseFishCells+coarseKey,fish);
    Cell_remove(simu.coarseFishCells+fish->coarseKey,fish);
    fish->coarseKey=coarseKey;
    }
#else
  if(fineKey!=fish->fineKey[toState])
    {
    Cell_add(simu.fineFishCells[toState]+fineKey,fish);
    if(fish->fineKey[toState]>=0)
      {
      Cell_remove(simu.fineFishCells[toState]+fish->fineKey[toState],fish);
      }
    fish->fineKey[toState]=fineKey;
    }
  if(coarseKey!=fish->coarseKey[toState])
    {
    Cell_add(simu.coarseFishCells[toState]+coarseKey,fish);
    if(fish->coarseKey[toState]>=0)
      {
      Cell_remove(simu.coarseFishCells[toState]+fish->coarseKey[toState],fish);
      }
    fish->coarseKey[toState]=coarseKey;
    }
#endif
// if(Fish_numaMigration(fish)) return ACT_DEAD;
#if defined ASYNC_BOIDS
  SpinSynchro_unlockW(&fish->sync);
#endif
return ACT_DONE;
}

Fish *
Fish_new(int school)
{
int cpuCtx=Platform_cpuCtx();
Fish *fish=Memory_alloc(1,Fish);
fish->dead=false;
fish->speed=FISH_SPEED_MIN;
fish->range=FISH_RANGE_MIN;
fish->school=school;
fish->memoCount=simu.memo ? Random_int(cpuCtx,simu.memo) : 0;
PtrVector_init(&fish->memo);
#if defined ASYNC_BOIDS
  SpatialBasis *location=&fish->location;
#else
  int toState=Scheduler_toState(&simu.sched);
  SpatialBasis *location=&fish->location[toState];
#endif
SpatialBasis_init(location);
real64_t x=R64Vector_get(&simu.schools,3*school);
real64_t y=R64Vector_get(&simu.schools,3*school+1);
real64_t z=R64Vector_get(&simu.schools,3*school+2);
x+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
y+=0.8*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5);
z+=0.5*LIMIT_SIZE*(Random_real64(cpuCtx)-0.5)+0.5*LIMIT_SIZE;
SpatialBasis_setPosition(location,x,y,z);
SpatialBasis_yaw(location,2.0*M_PI*(Random_real64(cpuCtx)-0.5));
SpatialBasis_pitch(location,M_PI*(Random_real64(cpuCtx)-0.5));
#if defined ASYNC_BOIDS
  fish->fineKey=SpatialGrid_coordToKey(&simu.fineFishGrid,x,y,z);
  Cell_add(simu.fineFishCells+fish->fineKey,fish);
  fish->coarseKey=SpatialGrid_coordToKey(&simu.coarseFishGrid,x,y,z);
  Cell_add(simu.coarseFishCells+fish->coarseKey,fish);
  SpinSynchro_init(&fish->sync);
#else
  int fromState=Scheduler_fromState(&simu.sched);
  fish->location[fromState]=*location; // bitwise copy is correct here
  fish->fineKey[fromState]=-1;
  fish->coarseKey[fromState]=-1;
  fish->fineKey[toState]=SpatialGrid_coordToKey(&simu.fineFishGrid,x,y,z);
  Cell_add(simu.fineFishCells[toState]+fish->fineKey[toState],fish);
  fish->coarseKey[toState]=SpatialGrid_coordToKey(&simu.coarseFishGrid,x,y,z);
  Cell_add(simu.coarseFishCells[toState]+fish->coarseKey[toState],fish);
#endif
Activity_init(&fish->act,&Fish_activate,&simu.sched,cpuCtx);
SpinSynchro_lockW(&simu.fishSync);
PtrVector_push(&simu.fish,fish);
SpinSynchro_unlockW(&simu.fishSync);
return fish;
}

//----------------------------------------------------------------------------
