//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef BOIDS_H
#define BOIDS_H 1

#include "TransProg/random.h"
#include "TransProg/scheduler.h"
#include "TransProg/spatialUtils.h"
#include "TransProg/coordIndex.h"
#include "TransProg/r64Vector.h"

#define ASYNC_BOIDS
#define SHOW_AFFINITY
// #define AVOID_RENDERING

#define MAX_DT 0.1

// #define LIMIT_SIZE 4.0
#define LIMIT_SIZE 6.0
#define FISH_SIZE 0.125
#define ROCK_SIZE 0.25
#define SHARK_SIZE 1.0

#define FISH_RANGE_MIN FISH_SIZE
// #define FISH_RANGE_MAX (32.0*FISH_SIZE)
#define FISH_RANGE_MAX (4.0*FISH_SIZE)
#define FISH_SPEED_MIN (2.0*FISH_SIZE)
#define FISH_SPEED_MAX (16.0*FISH_SIZE)

// #define SHARK_RANGE (2.0*SHARK_SIZE)
#define SHARK_RANGE (1.5*SHARK_SIZE)
#define SHARK_SPEED (1.5*SHARK_SIZE)

//----------------------------------------------------------------------------

typedef struct _Rock
{
SpatialBasis location;
intptr_t key;
} Rock;

Rock *
Rock_new(int school);

//----------------------------------------------------------------------------

typedef struct _Shark
{
Activity act;
bool dead;
int school;
int memoCount;
PtrVector memo;
#if defined ASYNC_BOIDS
  SpatialBasis location;
  intptr_t key;
  SpinSynchro sync;
#else
  SpatialBasis location[2];
  intptr_t key[2];
#endif
} Shark;

Shark *
Shark_new(int school);

//----------------------------------------------------------------------------

typedef struct _Fish
{
Activity act;
bool dead;
real64_t speed;
real64_t range;
int school;
int memoCount;
PtrVector memo;
#if defined ASYNC_BOIDS
  SpatialBasis location;
  intptr_t fineKey;
  intptr_t coarseKey;
  SpinSynchro sync;
#else
  SpatialBasis location[2];
  intptr_t fineKey[2];
  intptr_t coarseKey[2];
#endif
} Fish;

Fish *
Fish_new(int school);

//----------------------------------------------------------------------------

typedef struct _Cell
{
SpinSynchro sync;
PtrVector content;
} Cell;

TRANSPROG_INLINE
void
Cell_add(Cell *cell,
         void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_push(&cell->content,element);
SpinSynchro_unlockW(&cell->sync);
}

TRANSPROG_INLINE
void
Cell_remove(Cell *cell,
            void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_fastErase(&cell->content,
                    PtrVector_find(&cell->content,
                                   (PtrVector_CmpFnct)0,
                                   element));
SpinSynchro_unlockW(&cell->sync);
}

//----------------------------------------------------------------------------

typedef struct _Simulation
{
Scheduler sched;
int memo;
int schoolCount;
int rockCount;
int fishCount;
int sharkCount;
real64_t dt;
bool neighbourAffinity;
#if !defined AVOID_RENDERING
  bool needsRedraw[2];
#endif
bool mustQuit;
int numaMigration;
R64Vector schools;
PtrVector rocks;
SpinSynchro sharksSync;
PtrVector sharks;
SpinSynchro fishSync;
PtrVector fish;
SpatialGrid rockGrid;
SpatialGrid fineFishGrid;
SpatialGrid coarseFishGrid;
SpatialGrid sharkGrid;
Cell *rockCells;
#if defined ASYNC_BOIDS
  Cell *fineFishCells;
  Cell *coarseFishCells;
  Cell *sharkCells;
#else
  Cell *fineFishCells[2];
  Cell *coarseFishCells[2];
  Cell *sharkCells[2];
#endif
} Simulation;

extern Simulation simu;

#if !defined AVOID_RENDERING

void
Rendering_showFish(Fish *f);

void
Rendering_activate(int cpuCtx,
                   void *data);

void
Rendering_init(void);

#endif

#endif // BOIDS_H

//----------------------------------------------------------------------------
