//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef MECASIMU_H
#define MECASIMU_H 1

#include "TransCuda/cuda.h"
#include "TransProg/math.h"

#define BLOCK_DIM_X 16 // 32
#define BLOCK_DIM_Y 4  // 8
#define BLOCK_DIM_Z 3  // 4
#define UNIDIMENSIONAL
#define GLOBAL_SPLIT_XYZ
#define SHARED_COPY
// #define SHARED_SPLIT_XYZ

// #define AVOID_RENDERING

#define NX 64
#define NY 32
#define NZ 8
#define L0 (1.0f/NX)

#if !BLOCK_DIM_Y || !BLOCK_DIM_Z
# if !defined UNIDIMENSIONAL
#   define UNIDIMENSIONAL
# endif
# if defined SHARED_COPY
#   undef SHARED_COPY
# endif
#endif

#define USE_RK4
// #define USE_VERLET
#define GPU_SOURCE_FILE "mecaSimu.gpu"

#define DENSITY 1.15e3f // (kg/m^3) rubber
#define ELASTIC_MODULUS 0.02e9f // (N/m^2) rubber
#define DT (1.0e-4f) // < L0*sqrt(DENSITY/ELASTIC_MODULUS) for explicit Euler
#define DAMP 10.0f
#define MASS (DENSITY*L0*L0*L0)
#define K_ELAST (ELASTIC_MODULUS*L0*L0)
#define MAX_FORCE 5000.0f

#if defined GLOBAL_SPLIT_XYZ
# define MAKE_ID(x,y,z) ((x)+NX*((y)+NY*(z)))
# define OX 0
# define OY (NX*NY*NZ)
# define OZ (OY+OY)
#else
# define MAKE_ID(x,y,z) (3*((x)+NX*((y)+NY*(z))))
# define OX 0
# define OY 1
# define OZ 2
#endif

typedef struct _AppliData
{
bool mustQuit,mustRedraw;
real32_t delta[3];
int anchorX,anchorY,anchorZ;
real32_t forceX,forceY,forceZ;
struct
  {
  CudaProgram *prg;
  int gridDimX,gridDimY,gridDimZ;
  int shMem;
  CudaBuffer *spheres;
  CudaLockedMem *sphereMem;
  CudaBuffer *colors;
  CudaLockedMem *colorMem;
  CudaBuffer *lines;
  CudaLockedMem *lineMem;
  CudaBuffer *p0;
  CudaLockedMem *p0mem;
  CudaBuffer *s0;
  CudaLockedMem *s0mem;
  CudaBuffer *p1;
  CudaLockedMem *p1mem;
#if defined USE_RK4
  CudaBuffer *s1;
  CudaLockedMem *s1mem;
  CudaBuffer *p2;
  CudaLockedMem *p2mem;
  CudaBuffer *s2;
  CudaLockedMem *s2mem;
  CudaBuffer *p3;
  CudaLockedMem *p3mem;
  CudaBuffer *s3;
  CudaLockedMem *s3mem;
#endif
  } gpu;
real32_t *spheres;
real32_t *colors;
real32_t *lines;
real32_t *p0;
real32_t *s0;
real32_t *p1;
#if defined USE_RK4
real32_t *s1;
real32_t *p2;
real32_t *s2;
real32_t *p3;
real32_t *s3;
volatile int rk;
#endif
int *threadLimits;
} AppliData;

extern AppliData ad;

void
AppliData_reset(void);

void
AppliData_addForce(real64_t dx,
                   real64_t dy,
                   real64_t dz);

void
AppliData_moveAnchor(real64_t dx,
                     real64_t dy,
                     real64_t dz);

#if !defined AVOID_RENDERING

void
Rendering_init(void);

int
Rendering_sphereBuffer(void);

int
Rendering_colorBuffer(void);

int
Rendering_lineBuffer(void);

void
Rendering_fillBuffers(const void *speheres,
                      const void *colors,
                      const void *lines);

bool
Rendering_visible(void);

void
Rendering_update(void);

#endif

#endif // MECASIMU_H

//----------------------------------------------------------------------------
