//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "mecaSimu.h"

#if !defined AVOID_RENDERING

#include "TransProg/ptrVector.h"

#include "TransGl/glDraw.h"
#include "TransGl/viewer.h"

typedef struct _RenderingData
{
PtrVector viewers;
int viewerCount;
GlShape lines;
GlPointSphere spheres;
} RenderingData;

static RenderingData rd;

static
Viewer *
addViewer(ViewerCallback callback)
{
int w=800;
int h=600;
int sw=GlView_screenWidth()-w;
int sh=GlView_screenHeight()-h;
int x=sw-((rd.viewerCount*50)%sw);
int y=(rd.viewerCount*50)%sh;
Viewer *v=Viewer_new(x,y,w,h,false,callback,(void *)0);
char buffer[0x100];
sprintf(buffer,"Viewer %d",++rd.viewerCount);
GlView_setTitle(Viewer_view(v),buffer);
SpatialBasis_yaw(Viewer_viewPoint(v),5.0*M_PI/6.0);
SpatialBasis_translate(Viewer_viewPoint(v),-2.0,-0.25,1.0);
SpatialBasis_pitch(Viewer_viewPoint(v),M_PI/6.0);
Perspective_setNearDistance(Viewer_perspective(v),0.01);
Perspective_setFarDistance(Viewer_perspective(v),10.0);
Perspective_setStereoEyeSeparation(Viewer_perspective(v),0.01);
Perspective_setStereoViewDistance(Viewer_perspective(v),1.0);
PtrVector_push(&rd.viewers,v);
return v;
}

int
Rendering_sphereBuffer(void)
{
return GlPointSphere_vertexBuffer(&rd.spheres);
}

int
Rendering_colorBuffer(void)
{
return GlPointSphere_colorBuffer(&rd.spheres);
}

int
Rendering_lineBuffer(void)
{
return GlShape_vertexBuffer(&rd.lines);
}

void
Rendering_fillBuffers(const void *spheres,
                      const void *colors,
                      const void *lines)
{
glBindBuffer(GL_ARRAY_BUFFER,Rendering_sphereBuffer());
glBufferSubData(GL_ARRAY_BUFFER,
                0,(NX*NY*NZ)*3*sizeof(real32_t),
                spheres);
glBindBuffer(GL_ARRAY_BUFFER,Rendering_colorBuffer());
glBufferSubData(GL_ARRAY_BUFFER,
                0,(NX*NY*NZ)*3*sizeof(real32_t),
                colors);
glBindBuffer(GL_ARRAY_BUFFER,Rendering_lineBuffer());
glBufferSubData(GL_ARRAY_BUFFER,
                0,(27*2)*3*sizeof(real32_t),
                lines);
}

static
void
viewerCB(Viewer *v)
{
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
Viewer_setRefDistanceFromPoint(v,L0*NX*0.5,L0*NY*0.5,L0*NZ*0.5);
bool close=false;
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"v"))
        {
        addViewer(&viewerCB);
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"Home"))
        {
        AppliData_reset();
        }
      else if(!strcmp(GfxState_key(state),"Down"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalZ(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(-dx,-dy,-dz);
          }
        else
          {
          AppliData_addForce(-dx,-dy,-dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Up"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalZ(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(dx,dy,dz);
          }
        else
          {
          AppliData_addForce(dx,dy,dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Right"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalY(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(-dx,-dy,-dz);
          }
        else
          {
          AppliData_addForce(-dx,-dy,-dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Left"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalY(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(dx,dy,dz);
          }
        else
          {
          AppliData_addForce(dx,dy,dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Next"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalX(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(-dx,-dy,-dz);
          }
        else
          {
          AppliData_addForce(-dx,-dy,-dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Prior"))
        {
        real64_t dx,dy,dz;
        SpatialBasis_getLocalX(&v->viewPoint,&dx,&dy,&dz);
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          {
          AppliData_moveAnchor(dx,dy,dz);
          }
        else
          {
          AppliData_addForce(dx,dy,dz);
          }
        }
      else if(!strcmp(GfxState_key(state),"Delete"))
        {
        ad.forceX=0.0;
        ad.forceY=0.0;
        ad.forceZ=0.0;
        }
      else if(!strcmp(GfxState_key(state),"\x1B"))
        {
        close=true;
        }
      else if(!strcmp(GfxState_key(state),"q"))
        {
        ad.mustQuit=true;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    const real32_t headLight[3]={ 0.0f, 0.0f, 0.0f };
    const real32_t *fog=(const real32_t *)0;
    const real32_t white[4]={ 1.0f, 1.0f, 1.0f, 1.0f };
    const real32_t ambient=0.2f;
    const real32_t specular=0.75f;
    const real32_t shininess=20.0f;
    GlShape_draw(&rd.lines,
                 white,ambient,specular,shininess,
                 headLight,fog,0,projMtx,viewMtx);
    GlPointSphere_draw(&rd.spheres,
                       white,ambient,specular,shininess,0.5f*L0,
                       headLight,fog,projMtx,viewMtx);
    break;
    }
  default:
    {
    break;
    }
  }
if(close)
  {
  for(intptr_t i=PtrVector_size(&rd.viewers);i--;)
    {
    if(PtrVector_get(&rd.viewers,i)==v)
      {
      Viewer_delete(v);
      PtrVector_fastErase(&rd.viewers,i);
      }
    }
  if(!PtrVector_size(&rd.viewers))
    {
    ad.mustQuit=true;
    }
  }
}

bool
Rendering_visible(void)
{
for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
  {
  Viewer *v=(Viewer *)PtrVector_get(&rd.viewers,i);
  if(GfxState_visible(GlView_state(Viewer_view(v))))
    {
    return true;
    }
  }
return false;
}

void
Rendering_update(void)
{
for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
  {
  Viewer *v=(Viewer *)PtrVector_get(&rd.viewers,i);
  GlView_postRedraw(Viewer_view(v));
  }
GlView_sync();
GlView_processEvents(0);
}

void
Rendering_init(void)
{
GlView_init();
GlDraw_init();
PtrVector_init(&rd.viewers);
rd.viewerCount=0;
real32_t lineVertices[(27+3)*2*3];
real32_t lineColors[(27+3)*2*3];
int i;
for(i=0;i<27*2;++i)
  {
  lineVertices[i*3+0]=0.0f;
  lineVertices[i*3+1]=0.0f;
  lineVertices[i*3+2]=0.0f;
  lineColors[i*3+0]=0.8f;
  lineColors[i*3+1]=0.0f;
  lineColors[i*3+2]=0.8f;
  }
lineVertices[i*3+0]=0.0f; // X axis (red)
lineVertices[i*3+1]=0.0f;
lineVertices[i*3+2]=0.0f;
lineColors[i*3+0]=0.8f;
lineColors[i*3+1]=0.0f;
lineColors[i*3+2]=0.0f;
++i;
lineVertices[i*3+0]=1.0f;
lineVertices[i*3+1]=0.0f;
lineVertices[i*3+2]=0.0f;
lineColors[i*3+0]=0.8f;
lineColors[i*3+1]=0.0f;
lineColors[i*3+2]=0.0f;
++i;
lineVertices[i*3+0]=0.0f; // Y axis (green)
lineVertices[i*3+1]=0.0f;
lineVertices[i*3+2]=0.0f;
lineColors[i*3+0]=0.0f;
lineColors[i*3+1]=0.8f;
lineColors[i*3+2]=0.0f;
++i;
lineVertices[i*3+0]=0.0f;
lineVertices[i*3+1]=1.0f;
lineVertices[i*3+2]=0.0f;
lineColors[i*3+0]=0.0f;
lineColors[i*3+1]=0.8f;
lineColors[i*3+2]=0.0f;
++i;
lineVertices[i*3+0]=0.0f; // Z axis (blue)
lineVertices[i*3+1]=0.0f;
lineVertices[i*3+2]=0.0f;
lineColors[i*3+0]=0.0f;
lineColors[i*3+1]=0.0f;
lineColors[i*3+2]=0.8f;
++i;
lineVertices[i*3+0]=0.0f;
lineVertices[i*3+1]=0.0f;
lineVertices[i*3+2]=1.0f;
lineColors[i*3+0]=0.0f;
lineColors[i*3+1]=0.0f;
lineColors[i*3+2]=0.8f;
GlShape_init(&rd.lines,GLSHAPE_HAS_COLORS,GL_LINES,
             0,(const int *)0,(27+3)*2,lineVertices,
             (const real32_t *)0,(const real32_t *)0,lineColors);
GlPointSphere_init(&rd.spheres,GLPOINTSPHERE_HAS_COLORS,
                   NX*NY*NZ,(const real32_t *)0,
                   (const real32_t *)0,(const real32_t *)0);
addViewer(&viewerCB);
}

#endif // !AVOID_RENDERING

//----------------------------------------------------------------------------
