//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "microbio.h"

#if !defined AVOID_RENDERING

#include "TransGl/glDraw.h"
#include "TransGl/viewer.h"
#include "TransProg/r32Vector.h"
#include "TransProg/surface.h"

#define COLOR_COUNT (3*3*3)

typedef struct _RenderingData
{
PtrVector viewers;
int viewerCount;
struct
  {
  GlShape shape;
  int texture;
  int buffers[2];
  } substrate;
#if defined BACTERIUM_COUNT
  real32_t colors[COLOR_COUNT][4];
  GlPointSphere pointSphere;
#endif
} RenderingData;

static RenderingData rd;

typedef struct _ThreadViewerData
{
real64_t xSeen,ySeen,zSeen;
R32Vector seen;
} ThreadViewerData;

typedef struct _MyViewer
{
Viewer v;
bool close;
intptr_t seenCount;
real64_t xSeen,ySeen,zSeen;
ThreadViewerData **tvd[2];
} MyViewer;

static
void
addViewer(ViewerCallback callback)
{
MyViewer *mv=Memory_alloc(1,MyViewer);
int w=800;
int h=600;
int sw=GlView_screenWidth()-w;
int sh=GlView_screenHeight()-h;
int x=sw-((rd.viewerCount*50)%sw);
int y=(rd.viewerCount*50)%sh;
Viewer_init(&mv->v,x,y,w,h,false,callback,(void *)0);
char buffer[0x100];
sprintf(buffer,"Viewer %d",++rd.viewerCount);
GlView_setTitle(Viewer_view(&mv->v),buffer);
SpatialBasis_yaw(Viewer_viewPoint(&mv->v),M_PI_2);
SpatialBasis_pitch(Viewer_viewPoint(&mv->v),M_PI_2);
SpatialBasis_translate(Viewer_viewPoint(&mv->v),-0.8*SUBSTRATE_SIZE,0.0,0.0);
real64_t nearD=5.0*BACTERIUM_RENDER_SIZE;
real64_t farD=2.0*SUBSTRATE_SIZE;
Perspective_setNearDistance(Viewer_perspective(&mv->v),nearD);
Perspective_setFarDistance(Viewer_perspective(&mv->v),farD);
Perspective_setStereoEyeSeparation(Viewer_perspective(&mv->v),5.0*nearD);
Perspective_setStereoViewDistance(Viewer_perspective(&mv->v),0.5*(nearD+farD));
mv->close=false;
mv->seenCount=0;
mv->xSeen=0.0;
mv->ySeen=0.0;
mv->zSeen=0.0;
mv->tvd[0]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
mv->tvd[1]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  ThreadViewerData *tvd=Memory_alloc(1,ThreadViewerData);
  R32Vector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[0][i]=tvd;
  tvd=Memory_alloc(1,ThreadViewerData);
  R32Vector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[1][i]=tvd;
  }
PtrVector_push(&rd.viewers,mv);
}

int *
Rendering_buffers(void)
{
return rd.substrate.buffers;
}

void
Rendering_destroyBuffers(void)
{
glDeleteTextures(1,(const unsigned int *)&rd.substrate.texture);
glDeleteBuffers(2,(const unsigned int *)rd.substrate.buffers);
}

static
void
viewerCB(Viewer *v)
{
MyViewer *mv=STRUCT_OF_MEMBER(v,MyViewer,v);
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=mv->seenCount ? 1.0/mv->seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*mv->xSeen,d*mv->ySeen,d*mv->zSeen);
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"v"))
        {
        addViewer(&viewerCB);
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"Home"))
        {
        ad.needsReset=true;
        }
      else if(!strcmp(GfxState_key(state),"\x1B"))
        {
        mv->close=true;
        }
      else if(!strcmp(GfxState_key(state),"q"))
        {
        ad.mustQuit=true;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    mv->close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    real32_t headLight[3]={0.0f,0.0f,0.0f};
    const real32_t *fog=(const real32_t *)0;
    real32_t substrateColor[4]={0.8f,0.8f,0.8f,1.0f};
    real32_t ambient=0.2f;
    real32_t specular=0.75f;
    real32_t shininess=20.0f;
    GlShape_draw(&rd.substrate.shape,
                 substrateColor,ambient,specular,shininess,
                 headLight,fog,rd.substrate.texture,projMtx,viewMtx);
#if defined BACTERIUM_COUNT
    int fromState=Scheduler_fromState(&ad.sched);
    for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
      {
      ThreadViewerData *tvd=mv->tvd[fromState][i];
      mv->xSeen+=tvd->xSeen;
      mv->ySeen+=tvd->ySeen;
      mv->zSeen+=tvd->zSeen;
      R32Vector *seen=&tvd->seen;
      intptr_t seenCount=R32Vector_size(seen)/3;
      mv->seenCount+=seenCount;
      rd.pointSphere.vertexCount=seenCount;
      glBindBuffer(GL_ARRAY_BUFFER,
                   GlPointSphere_vertexBuffer(&rd.pointSphere));
      glBufferSubData(GL_ARRAY_BUFFER,0,
                      rd.pointSphere.vertexCount*3*sizeof(real32_t),
                      R32Vector_data(&tvd->seen));
      glBindBuffer(GL_ARRAY_BUFFER,0);
      GlPointSphere_draw(&rd.pointSphere,
                         rd.colors[i%COLOR_COUNT],
                         ambient,specular,shininess,0.5f*BACTERIUM_RENDER_SIZE,
                         headLight,fog,projMtx,viewMtx);
      }
#endif
    break;
    }
  default:
    {
    break;
    }
  }
}

void
Rendering_showBacterium(Bacterium *b)
{
int cpuCtx=Activity_cpuCtx(&b->act);
int toState=Scheduler_toState(&ad.sched);
real64_t x=b->x,y=b->y,z=b->z;
for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
  {
  const MyViewer *mv=(const MyViewer *)PtrVector_get(&rd.viewers,i);
  if(Viewer_view(&mv->v)&&
     GfxState_visible(Viewer_state(&mv->v))) // FIXME: race condition
    {
    real64_t lx=x,ly=y,lz=z;
    SpatialBasis_globalToLocalPosition(Viewer_viewPoint(&mv->v),&lx,&ly,&lz);
    if(Perspective_locateVertex(Viewer_perspective(&mv->v),lx,ly,lz)
       ==LOC_MIDDLE)
      {
      ThreadViewerData *tvd=mv->tvd[toState][cpuCtx];
      R32Vector_push(&tvd->seen,(real32_t)x);
      R32Vector_push(&tvd->seen,(real32_t)y);
      R32Vector_push(&tvd->seen,(real32_t)z);
      tvd->xSeen+=x;
      tvd->ySeen+=y;
      tvd->zSeen+=z;
      }
    }
  }
}

void
Rendering_activate(int cpuCtx,
                   void *data)
{
(void)data;
bool needsGfxSync=false;
int toState=Scheduler_toState(&ad.sched);
int64_t t0=Platform_time();
if(ad.needsRedraw[toState])
  {
  for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
    {
    MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
    ThreadViewerData *tvd=mv->tvd[toState][cpuCtx];
    R32Vector_clear(&tvd->seen);
    tvd->xSeen=0.0;
    tvd->ySeen=0.0;
    tvd->zSeen=0.0;
    if(!cpuCtx&&mv->close)
      {
      PtrVector_fastErase(&rd.viewers,i); // FIXME: race condition
      if(!PtrVector_size(&rd.viewers))
        {
        ad.mustQuit=true;
        }
      Viewer_destroy(&mv->v);
      for(int c=0,c_end=Platform_cpuCtxCount();c<c_end;++c)
        {
        R32Vector_destroy(&mv->tvd[0][c]->seen);
        Memory_free(mv->tvd[0][c]);
        R32Vector_destroy(&mv->tvd[1][c]->seen);
        Memory_free(mv->tvd[1][c]);
        }
      Memory_free(mv->tvd[0]);
      Memory_free(mv->tvd[1]);
      Memory_free(mv);
      }
    }
  }
if(!cpuCtx)
  {
  int fromState=Scheduler_fromState(&ad.sched);
  if(ad.needsRedraw[fromState])
    {
    glBindTexture(GL_TEXTURE_2D,rd.substrate.texture);
    if(ad.rendering[fromState])
      {
      glTexSubImage2D(GL_TEXTURE_2D,0,0,0,
                      SUBSTRATE_WIDTH,SUBSTRATE_HEIGHT,
                      GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,
                      ad.rendering[fromState]);
      }
    else
      {
      glBindBuffer(GL_PIXEL_UNPACK_BUFFER,rd.substrate.buffers[fromState]);
      glTexSubImage2D(GL_TEXTURE_2D,0,0,0,
                      SUBSTRATE_WIDTH,SUBSTRATE_HEIGHT,
                      GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,
                      (const void *)0);
      glBindBuffer(GL_PIXEL_UNPACK_BUFFER,0);
      }
    for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
      {
      MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
      GlView_postRedraw(Viewer_view(&mv->v));
      }
    GlView_processEvents(0);
    needsGfxSync=true;
    }
  }
int64_t t1=Platform_time();
AppliData_activate(cpuCtx,(void *)0);
int64_t t2=Platform_time();
if(needsGfxSync)
  {
  GlView_sync();
  int64_t t3=Platform_time();
  ad.renderingDuration+=(t1-t0)+(t3-t2);
  }
}

void
Rendering_init(void)
{
GlView_init();
GlDraw_init();
//-- Substrate --
int indexCount,vertexCount;
int *indices;
real32_t *vertices;
real32_t *normals;
real32_t *texCoords;
Surface_allocGrid(&indexCount,&indices,&vertexCount,
                  &vertices,&normals,&texCoords,
                  true,SUBSTRATE_SIZE,SUBSTRATE_SIZE,1,1);
GlShape_init(&rd.substrate.shape,GLSHAPE_HAS_NORMALS|GLSHAPE_HAS_TEXCOORDS,
             GL_TRIANGLES,indexCount,indices,vertexCount,vertices,
             normals,texCoords,(const real32_t *)0);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);
glGenTextures(1,(unsigned int *)&rd.substrate.texture);
glBindTexture(GL_TEXTURE_2D,rd.substrate.texture);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,SUBSTRATE_WIDTH,SUBSTRATE_HEIGHT,
             0,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,(const void *)0);
glGenBuffers(2,(unsigned int *)rd.substrate.buffers);
for(int i=0;i<2;++i)
  {
  glBindBuffer(GL_PIXEL_UNPACK_BUFFER,rd.substrate.buffers[i]);
  glBufferData(GL_PIXEL_UNPACK_BUFFER,
               SUBSTRATE_WIDTH*SUBSTRATE_HEIGHT*4,
               (const void *)0,GL_STATIC_DRAW);
  glBindBuffer(GL_PIXEL_UNPACK_BUFFER,0);
  }
#if defined BACTERIUM_COUNT
  //-- Bacteria --
  for(int blue=0;blue<3;++blue)
    {
    for(int green=0;green<3;++green)
      {
      for(int red=0;red<3;++red)
        {
        real32_t *c=rd.colors[blue+3*(green+3*red)];
        c[0]=0.2f+0.3f*red;
        c[1]=0.2f+0.3f*green;
        c[2]=0.8f-0.3f*blue;
        c[3]=1.0f;
        }
      }
    }
  GlPointSphere_init(&rd.pointSphere,0,BACTERIUM_COUNT,(const int *)0,
                     (const int *)0,(const int *)0);
#endif
//-- Viewers --
PtrVector_init(&rd.viewers);
rd.viewerCount=0;
addViewer(&viewerCB);
}

#endif // !AVOID_RENDERING

//----------------------------------------------------------------------------
