//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/glView.h"
#include "TransGl/glDraw.h"

/**
Approaching zero driver overhead
GAME developer conference 2014

STA_D STA_R INV_W SUB_R --> best send, compute, total
SUB_R --> best receive
**/

const char *
mappingStr(int mapping)
{
const char *str="?????";
switch(mapping)
  {
  case 0:
    { str="UNMAP"; break; }
  case GL_MAP_READ_BIT:
    { str="MAP_R"; break; }
  case GL_MAP_WRITE_BIT:
    { str="MAP_W"; break; }
  case GL_MAP_WRITE_BIT|GL_MAP_INVALIDATE_BUFFER_BIT:
    { str="INV_W"; break; }
  case GL_MAP_WRITE_BIT|GL_MAP_UNSYNCHRONIZED_BIT:
    { str="UNS_W"; break; }
  }
return str;
}

const char *
usageStr(int usage)
{
const char *str="?????";
switch(usage)
  {
  case GL_STATIC_DRAW:
    { str="STA_D"; break; }
  case GL_DYNAMIC_DRAW:
    { str="DYN_D"; break; }
  case GL_STREAM_DRAW:
    { str="STR_D"; break; }
  case GL_STATIC_READ:
    { str="STA_R";  break; }
  case GL_DYNAMIC_READ:
    { str="DYN_R"; break; }
  case GL_STREAM_READ:
    { str="STR_R"; break; }
  }
return str;
}

void
runTest(int arraySize,
        int inputUsage,
        int outputUsage,
        int inputMapping,
        int outputMapping)
{
const char *sourceCode=
" #version 430                                                            \n"
" uniform uint u_ArraySize;                                               \n"
" layout(binding=0) writeonly restrict buffer SSBO_0                      \n"
"   { int outArray[]; };                                                  \n"
" layout(binding=1) readonly restrict buffer SSBO_1                       \n"
"   { int inArray1[]; };                                                  \n"
" layout(binding=2) readonly restrict buffer SSBO_2                       \n"
"   { int inArray2[]; };                                                  \n"
" layout(local_size_x=1024) in;                                           \n"
" void                                                                    \n"
" main(void)                                                              \n"
" {                                                                       \n"
" uint id=gl_GlobalInvocationID.x;                                        \n"
" if(id<u_ArraySize)                                                      \n"
"   {                                                                     \n"
"   outArray[id]=inArray1[id]+inArray2[id];                               \n"
"   }                                                                     \n"
" }                                                                       \n";
int shader=glCreateShader(GL_COMPUTE_SHADER);
if(!shader)
  {
  OPENGL_CHECK;
  }
glShaderSource(shader,1,&sourceCode,(const int *)0);
OPENGL_CHECK;
glCompileShader(shader);
OPENGL_CHECK;
bool error;
char *log=CoreGl_shaderStatus(shader,&error);
if(log)
  {
  Platform_msg("COMPUTE SHADER SOURCE:\n%s\n",sourceCode);
  Platform_msg("COMPUTE SHADER LOG:\n%s\n",log);
  Memory_free(log);
  }
if(error)
  {
  DROP_DEAD("shader compilation failure");
  }
int program=glCreateProgram();
if(!program)
  {
  OPENGL_CHECK;
  }
glAttachShader(program,shader);
OPENGL_CHECK;
glLinkProgram(program);
OPENGL_CHECK;
log=CoreGl_shaderStatus(program,&error);
if(log)
  {
  Platform_msg("COMPUTE SHADER PROGRAM LOG:\n%s\n",log);
  Memory_free(log);
  }
if(error)
  {
  DROP_DEAD("shader link failure");
  }
int localSize[3];
glGetProgramiv(program,GL_COMPUTE_WORK_GROUP_SIZE,localSize);
OPENGL_CHECK;
int u_ArraySize=glGetUniformLocation(program,"u_ArraySize");
OPENGL_CHECK;
unsigned int buffers[3];
glGenBuffers(3,buffers);
OPENGL_CHECK;
for(int i=0;i<3;++i)
  {
  glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[i]);
  OPENGL_CHECK;
#if 0
  glBufferData(GL_SHADER_STORAGE_BUFFER,arraySize*sizeof(int),NULL,
               i ? inputUsage : outputUsage);
  OPENGL_CHECK;
#else
  int flag=0;
  if(i)
    {
    if(inputMapping)
      {
      flag|=GL_MAP_WRITE_BIT;
      }
    else
      {
      flag|=GL_DYNAMIC_STORAGE_BIT;
      }
    if(inputUsage==GL_STREAM_DRAW)
      {
      flag|=GL_CLIENT_STORAGE_BIT;
      }
    }
  else
    {
    if(outputMapping)
      {
      flag|=GL_MAP_READ_BIT;
      }
    else
      {
      flag|=GL_DYNAMIC_STORAGE_BIT;
      }
    if(outputUsage==GL_STREAM_READ)
      {
      flag|=GL_CLIENT_STORAGE_BIT;
      }
    }
  glBufferStorage(GL_SHADER_STORAGE_BUFFER,arraySize*sizeof(int),NULL,
                  flag);
  OPENGL_CHECK;
#endif
  glBindBufferBase(GL_SHADER_STORAGE_BUFFER,i,buffers[i]);
  OPENGL_CHECK;
  }
int data1[17]={37,50,54,50,56,15,51,32,12,20,43,56,99,50,25,15,18};
int data2[17]={35,51,54,58,55,17,20,48,73,12,44,55,15,58,75,18,14};
int *outArray=Memory_alloc(arraySize,int);
int *inArray1=Memory_alloc(arraySize,int);
int *inArray2=Memory_alloc(arraySize,int);
unsigned int queries[3];
glGenQueries(3,queries);
OPENGL_CHECK;
int sendDuration=0;
int computeDuration=0;
int receiveDuration=0;
int repeat=10;
for(int i=0;i<repeat;++i)
  {
  for(int j=0;j<arraySize;++j)
    {
    inArray1[j]=data1[(j+i)%17];
    inArray2[j]=data2[(j+i)%17];
    }
  glBeginQuery(GL_TIME_ELAPSED,queries[0]);
  OPENGL_CHECK;
  if(inputMapping)
    {
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[1]);
    OPENGL_CHECK;
    void *b1=glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
             0,arraySize*sizeof(int),inputMapping);
    OPENGL_CHECK;
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[2]);
    OPENGL_CHECK;
    void *b2=glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
             0,arraySize*sizeof(int),inputMapping);
    OPENGL_CHECK;
    memcpy(b1,inArray1,arraySize*sizeof(int));
    memcpy(b2,inArray2,arraySize*sizeof(int));
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[1]);
    OPENGL_CHECK;
    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
    OPENGL_CHECK;
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[2]);
    OPENGL_CHECK;
    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
    OPENGL_CHECK;
    }
  else
    {
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[1]);
    OPENGL_CHECK;
    glBufferSubData(GL_SHADER_STORAGE_BUFFER,
                    0,arraySize*sizeof(int),inArray1);
    OPENGL_CHECK;
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[2]);
    OPENGL_CHECK;
    glBufferSubData(GL_SHADER_STORAGE_BUFFER,
                    0,arraySize*sizeof(int),inArray2);
    OPENGL_CHECK;
    }
  glEndQuery(GL_TIME_ELAPSED);
  OPENGL_CHECK;
  glBeginQuery(GL_TIME_ELAPSED,queries[1]);
  OPENGL_CHECK;
  glUseProgram(program);
  OPENGL_CHECK;
  glUniform1ui(u_ArraySize,arraySize);
  OPENGL_CHECK;
  for(int iter=0;iter<20;++iter)
    glDispatchCompute((arraySize+localSize[0]-1)/localSize[0],1,1);
  OPENGL_CHECK;
  glEndQuery(GL_TIME_ELAPSED);
  OPENGL_CHECK;
  glBeginQuery(GL_TIME_ELAPSED,queries[2]);
  OPENGL_CHECK;
  if(outputMapping)
    {
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[0]);
    OPENGL_CHECK;
    const void *b0=glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
                   0,arraySize*sizeof(int),outputMapping);
    OPENGL_CHECK;
    memcpy(outArray,b0,arraySize*sizeof(int));
    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
    OPENGL_CHECK;
    }
  else
    {
    glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffers[0]);
    OPENGL_CHECK;
    glGetBufferSubData(GL_SHADER_STORAGE_BUFFER,
                       0,arraySize*sizeof(int),outArray);
    OPENGL_CHECK;
    }
  glEndQuery(GL_TIME_ELAPSED);
  OPENGL_CHECK;
  unsigned int durations[3];
  for(int j=0;j<3;++j)
    {
    glGetQueryObjectuiv(queries[j],GL_QUERY_RESULT,durations+j);
    OPENGL_CHECK;
    }
  if(i>=repeat/2)
    {
    sendDuration+=durations[0]/1000;
    computeDuration+=durations[1]/1000;
    receiveDuration+=durations[2]/1000;
    }
  if(i>repeat) // never true
    {
    Platform_msg("    ");
    for(int j=0;j<20;++j)
      Platform_msg("%c",(char)outArray[j]);
    Platform_msg(" ... ");
    for(int j=0;j<20;++j)
      Platform_msg("%c",(char)outArray[arraySize-20+j]);
    Platform_msg("\n");
    }
  }
Memory_free(inArray2);
Memory_free(inArray1);
Memory_free(outArray);
glDeleteQueries(3,queries);
OPENGL_CHECK;
glDeleteBuffers(3,buffers);
OPENGL_CHECK;
glDeleteProgram(program);
OPENGL_CHECK;
glDeleteShader(shader);
OPENGL_CHECK;
repeat-=repeat/2;
sendDuration/=repeat;
computeDuration/=repeat;
receiveDuration/=repeat;
Platform_msg("%s/%s %s/%s %d + %d + %d = %d\n",
             usageStr(inputUsage),mappingStr(inputMapping),
             usageStr(outputUsage),mappingStr(outputMapping),
             sendDuration,computeDuration,receiveDuration,
             sendDuration+computeDuration+receiveDuration);
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
GlView_init();
GlDraw_init();
#if 0
int inputUsage[]={ GL_STATIC_DRAW,
                   GL_DYNAMIC_DRAW,
                   GL_STREAM_DRAW };
int outputUsage[]={ GL_STATIC_READ,
                    GL_DYNAMIC_READ,
                    GL_STREAM_READ };
int inputMapping[]={ 0,
                     GL_MAP_WRITE_BIT,
                     GL_MAP_WRITE_BIT|GL_MAP_UNSYNCHRONIZED_BIT };
int outputMapping[]={ 0,
                      GL_MAP_READ_BIT };
#else
int inputUsage[]={ GL_STATIC_DRAW,
                   GL_STREAM_DRAW };
int outputUsage[]={ GL_STATIC_READ,
                    GL_STREAM_READ };
int inputMapping[]={ 0,
                     GL_MAP_WRITE_BIT,
                     GL_MAP_WRITE_BIT|GL_MAP_UNSYNCHRONIZED_BIT };
int outputMapping[]={ 0,
                      GL_MAP_READ_BIT };
#endif
for(int iu=0,iu_end=sizeof(inputUsage)/sizeof(inputUsage[0]);
    iu<iu_end;++iu)
  {
  for(int ou=0,ou_end=sizeof(outputUsage)/sizeof(outputUsage[0]);
      ou<ou_end;++ou)
    {
    for(int im=0,im_end=sizeof(inputMapping)/sizeof(inputMapping[0]);
        im<im_end;++im)
      {
      for(int om=0,om_end=sizeof(outputMapping)/sizeof(outputMapping[0]);
          om<om_end;++om)
        {
        runTest(16*1024*1024,
                inputUsage[iu],
                outputUsage[ou],
                inputMapping[im],
                outputMapping[om]);
        }
      }
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
