//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/viewer.h"
#include "TransProg/ptrVector.h"
#include "TransProg/random.h"
#include "TransProg/surface.h"

#include "TransGl/glDraw.h"

#define USE_TEXTURE

#if 0
  typedef real64_t coord_t;
# define BoundingBox_addVertices BoundingBox_addVertices64
#elif 1
  typedef real32_t coord_t;
# define BoundingBox_addVertices BoundingBox_addVertices32
#else
  typedef uint16_t coord_t;
# define BoundingBox_addVertices BoundingBox_addVertices16
#endif

typedef struct _Object
{
GlShape *sh;
real32_t rgba[4];
int textureId;
BoundingBox *bbox;
SpatialBasis location;
} Object;

Viewer *viewer;
int seenCount;
real64_t xSeen,ySeen,zSeen;
PtrVector objects;

void
viewerCB(Viewer *v)
{
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=seenCount ? 1.0/seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*xSeen,d*ySeen,d*zSeen);
bool close=false;
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"\x1B"))
        {
        close=true;
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    xSeen=ySeen=zSeen=0.0;
    seenCount=0;
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    real32_t ambient=0.2f;
    real32_t specular=0.75f;
    real32_t shininess=20.0f;
    const real32_t *fog=(const real32_t *)0;
    real32_t headLight[3]={0.0f,0.0f,0.0f};
    for(intptr_t i=0,i_end=PtrVector_size(&objects);i<i_end;++i)
      {
      Object *o=(Object *)PtrVector_get(&objects,i);
      if(Perspective_boxVisibility(Viewer_perspective(v),
                                   Viewer_viewPoint(v),
                                   o->bbox,&o->location,
                                   (SpatialTransform *)0))
        {
        ++seenCount;
        real64_t xc=BoundingBox_centerX(o->bbox);
        real64_t yc=BoundingBox_centerY(o->bbox);
        real64_t zc=BoundingBox_centerZ(o->bbox);
        SpatialBasis_localToGlobalPosition(&o->location,&xc,&yc,&zc);
        xSeen+=xc;
        ySeen+=yc;
        zSeen+=zc;
        real32_t modelViewMtx[16];
        Gfx_computeModelViewMatrix(modelViewMtx,
                                   viewMtx,o->location.matrix);
        GlShape_draw(o->sh,
                     o->rgba,ambient,specular,shininess,
                     headLight,fog,o->textureId,projMtx,modelViewMtx);
        }
      }
    break;
    }
  default:
    {
    break;
    }
  }
if(close)
  {
  Viewer_delete(v);
  viewer=(Viewer *)0;
  }
}

Object *
makeObject(GlShape *sh,
           BoundingBox *bbox,
           int textureId)
{
int cpuCtx=Platform_cpuCtx();
Object *o=Memory_alloc(1,Object);
o->sh=sh;
o->rgba[0]=Random_real32(cpuCtx);
o->rgba[1]=Random_real32(cpuCtx);
o->rgba[2]=Random_real32(cpuCtx);
real32_t n=sqrtf(o->rgba[0]*o->rgba[0]+
                 o->rgba[1]*o->rgba[1]+
                 o->rgba[2]*o->rgba[2]);
o->rgba[0]/=n;
o->rgba[1]/=n;
o->rgba[2]/=n;
o->rgba[3]=1.0f;
// o->rgba[2]=o->rgba[1]=o->rgba[0]=0.6f;
o->textureId=textureId;
o->bbox=bbox;
SpatialBasis_init(&o->location);
return o;
}

void
loadShape(int indexCount,
          const int *indices,
          int vertexCount,
          const coord_t *vertices,
          const coord_t *normals,
          const coord_t *texCoords,
          GlShape *sh,
          BoundingBox *bbox)
{
#if !defined USE_TEXTURE
texCoords=(const coord_t *)0;
#endif
GlShape_init(sh,GLSHAPE_HAS_NORMALS|GLSHAPE_HAS_TEXCOORDS,
             GL_TRIANGLES,indexCount,indices,
             vertexCount,vertices,normals,texCoords,(const coord_t*)0);
BoundingBox_init(bbox);
BoundingBox_addVertices(bbox,vertices,vertexCount);
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
int precision=20;
int iter=2;
if(argc>1)
  {
  sscanf(argv[1],"%d",&precision);
  }
if(argc>2)
  {
  sscanf(argv[2],"%d",&iter);
  }
GlView_init();
GlDraw_init();

int w=512;
int h=512;
viewer=Viewer_new(GlView_screenWidth()-w,0,w,h,false,&viewerCB,(void *)0);
SpatialBasis_translate(Viewer_viewPoint(viewer),-8.0,0.0,0.0);
PtrVector_init(&objects);

int textureId=0;
#if defined USE_TEXTURE
  int texWidth,texHeight,texComp;
  uint8_t *texData;
# if 0
    texWidth=512;
    texHeight=512;
    texComp=3;
    texData=Memory_alloc(texComp*texWidth*texHeight,uint8_t);
    int index=0;
    for(int y=0;y<texHeight;++y)
      {
      real32_t green=y/(texHeight-1.0f);
      for(int x=0;x<texWidth;++x)
        {
        real32_t red=x/(texWidth-1.0f);
        real32_t blue=red*green;
        texData[index++]=(uint8_t)(255.0f*(red-blue));
        texData[index++]=(uint8_t)(255.0f*(green-blue));
        texData[index++]=(uint8_t)(255.0f*blue);
        }
      }
    textureId=GlDraw_loadTexture(texWidth,texHeight,texComp,texData,false);
# else
    texData=Gfx_loadTgaImage("earth.tga",&texWidth,&texHeight,&texComp);
    textureId=GlDraw_loadTexture(texWidth,texHeight,texComp,texData,false);
# endif
  Memory_free(texData);
#endif

GlShape sh[10];
BoundingBox bbox[10];
int indexCount,vertexCount;
int *indices;
coord_t *vertices;
coord_t *normals;
coord_t *texCoords;

Surface_allocBox(&indexCount,&indices,&vertexCount,&vertices,
                 &normals,&texCoords,
                 true,2.0,1.5,1.0);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+0,bbox+0);
Surface_makeBox(indices,vertices,normals,texCoords,
                false,2.0,1.5,1.0);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+1,bbox+1);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);

Surface_allocCylinder(&indexCount,&indices,&vertexCount,&vertices,
                      &normals,&texCoords,
                      true,1.0,2.0,true,true,true,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+2,bbox+2);
Surface_makeCylinder(indices,vertices,normals,texCoords,
                     false,1.0,2.0,true,true,true,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+3,bbox+3);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);

Surface_allocCone(&indexCount,&indices,&vertexCount,&vertices,
                  &normals,&texCoords,
                  true,1.0,2.0,true,true,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+4,bbox+4);
Surface_makeCone(indices,vertices,normals,texCoords,
                 false,1.0,2.0,true,true,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+5,bbox+5);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);

Surface_allocSphere(&indexCount,&indices,&vertexCount,&vertices,
                    &normals,&texCoords,
                    true,1.0,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+6,bbox+6);
Surface_makeSphere(indices,vertices,normals,texCoords,
                   false,1.0,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+7,bbox+7);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);

Surface_allocTorus(&indexCount,&indices,&vertexCount,&vertices,
                   &normals,&texCoords,
                   true,0.6,0.25,precision,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+8,bbox+8);
Surface_makeTorus(indices,vertices,normals,texCoords,
                  false,0.6,0.25,precision,precision);
loadShape(indexCount,indices,vertexCount,vertices,
          normals,texCoords,sh+9,bbox+9);
Memory_free(indices); Memory_free(vertices);
Memory_free(normals); Memory_free(texCoords);

real64_t step=5.0/iter;
for(int i=-iter;i<=iter;++i)
  {
  Object *o=makeObject(sh+0,bbox+0,textureId);
  SpatialBasis_translate(&o->location,i*step,3.0,0.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+1,bbox+1,textureId);
  SpatialBasis_translate(&o->location,i*step,-3.0,0.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+2,bbox+2,textureId);
  SpatialBasis_translate(&o->location,i*step,3.0,2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+3,bbox+3,textureId);
  SpatialBasis_translate(&o->location,i*step,-3.0,2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+4,bbox+4,textureId);
  SpatialBasis_translate(&o->location,i*step,3.0,-2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+5,bbox+5,textureId);
  SpatialBasis_translate(&o->location,i*step,-3.0,-2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+6,bbox+6,textureId);
  SpatialBasis_translate(&o->location,i*step,1.0,2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+7,bbox+7,textureId);
  SpatialBasis_translate(&o->location,i*step,-1.0,2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+8,bbox+8,textureId);
  SpatialBasis_translate(&o->location,i*step,1.0,-2.0);
  PtrVector_push(&objects,o);

  o=makeObject(sh+9,bbox+9,textureId);
  SpatialBasis_translate(&o->location,i*step,-1.0,-2.0);
  PtrVector_push(&objects,o);
  }

int count=0;
int64_t lastTime=Platform_time();
for(;;)
  {
  GlView_processEvents(0);
  GlView_sync();
  if(!viewer) break;
  for(intptr_t i=0,i_end=PtrVector_size(&objects);i<i_end;++i)
    {
    Object *o=(Object *)PtrVector_get(&objects,i);
    real64_t angle=0.0*M_PI_180;
    SpatialBasis_roll(&o->location,-angle);
    SpatialBasis_pitch(&o->location,angle);
    SpatialBasis_yaw(&o->location,0.5*angle);
    }
  GlView_postRedraw(Viewer_view(viewer));
  ++count;
  int64_t dt=Platform_time()-lastTime;
  if(dt>1000000)
    {
    Platform_msg("%g\n",count*1000000.0/dt);
    count=0;
    lastTime+=dt;
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
