//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/iFifo.h"
#include "TransProg/iDeque.h"
#include "TransProg/iRbTree.h"
#include "TransProg/iVector.h"
#include "TransProg/random.h"
#include "TransProg/text.h"

static
void
showVector(const IVector *v)
{
for(intptr_t i=0,i_end=IVector_size(v);i<i_end;++i)
  {
  Platform_msg("%d ",IVector_get(v,i));
  }
Platform_msg("\n");
}

static
void
showDeque(const IDeque *d)
{
for(intptr_t i=0,i_end=IDeque_size(d);i<i_end;++i)
  {
  Platform_msg("%d ",IDeque_get(d,i));
  }
Platform_msg("\n");
}

void
task0(int cpuCtx)
{
(void)cpuCtx;
int i;
IVector *v=IVector_new();
for(i=10;i--;)
  {
  IVector_insert(v,(i&1) ? 0 : (int)IVector_size(v),2*i);
  }
Platform_msg("initial: "); showVector(v);
for(i=(int)IVector_size(v);i--;)
  {
  Platform_msg("%d at %d\n",
               2*i,(int)IVector_find(v,(IVector_CmpFnct)0,2*i));
  }
IVector_sort(v,(IVector_CmpFnct)0);
Platform_msg("sorted: "); showVector(v);
IVector_sort(v,&IVector_defaultCmpDownFnct);
Platform_msg("reverse sorted: "); showVector(v);
for(i=(int)IVector_size(v);i--;)
  {
  IVector_sortedInsert(v,&IVector_defaultCmpDownFnct,2*i+1);
  }
Platform_msg("reverse sorted insert: "); showVector(v);
for(i=(int)IVector_size(v);i--;)
  {
  Platform_msg("%d at %d\n",i,
               (int)IVector_sortedFind(v,IVector_defaultCmpDownFnct,i));
  }
IVector_delete(v);
}

void
task1(int cpuCtx)
{
(void)cpuCtx;
int storage[16];
IFifo f;
IFifo_init(&f,storage,16);
int i=0;
while(IFifo_push(&f,++i))
  {
  Platform_msg("push %d\n",i);
  }
while(IFifo_pop(&f,&i))
  {
  Platform_msg("pop %d\n",i);
  }
IFifo_destroy(&f);
}

void
task2(int cpuCtx)
{
(void)cpuCtx;
IDeque *d=IDeque_new();
while(IDeque_size(d)<40)
  {
  if(IDeque_size(d)%2)
    {
    IDeque_pushFront(d,(int)IDeque_size(d));
    }
  else
    {
    IDeque_pushBack(d,(int)IDeque_size(d));
    }
  showDeque(d);
  }
while(IDeque_size(d)>30)
  {
  Platform_msg("popBack: %d\n",IDeque_popBack(d));
  }
showDeque(d);
while(IDeque_size(d)>20)
  {
  Platform_msg("popFront: %d\n",IDeque_popFront(d));
  }
showDeque(d);
IDeque_shrink(d);
showDeque(d);
IDeque_delete(d);
}

void
visitTree(const int *elem,void *arg)
{
(void)arg;
Platform_msg(" %d",*elem);
}

void
task3(int cpuCtx)
{
IRbTree *t=IRbTree_new((IRbTree_CmpFnct)0);
while(IRbTree_size(t)<15)
  {
  IRbTree_insert(t,(int)IRbTree_size(t));
  Platform_msg("%d:",(int)IRbTree_size(t));
  IRbTree_forEach(t,&visitTree,(void *)0);
  Platform_msg("\n");
  }
Platform_msg("==================================================\n");
while(IRbTree_size(t))
  {
  IRbTreeNode *n=IRbTree_first(t);
  for(int i=Random_int(cpuCtx,(int)IRbTree_size(t));i--;)
    {
    n=IRbTree_next(n);
    }
  IRbTree_eraseNode(t,n);
  Platform_msg("%d:",(int)IRbTree_size(t));
  IRbTree_forEach(t,&visitTree,(void *)0);
  Platform_msg("\n");
  }
IRbTree_delete(t);
}

void
task4(int cpuCtx)
{
int range=100000000;
IRbTree *t=IRbTree_new((IRbTree_CmpFnct)0);
for(;;)
  {
  if(IRbTree_size(t)<range/100)
    {
    IRbTree_insert(t,Random_int(cpuCtx,range));
    }
  IRbTreeNode *n=IRbTree_find(t,Random_int(cpuCtx,range));
  if(n)
    {
    IRbTree_eraseNode(t,n);
    }
  }
// IRbTree_delete(t);
}

void
task5(int cpuCtx)
{
(void)cpuCtx;
Text t;
Text_init(&t);
for(int i=0;i<10;++i)
  {
  Text_insert(&t,0,"line %.8x %.8x %.8x %.8x %.8x %.8x\n",i,i,i,i,i,i);
  Platform_msg("---- %d ----\n%s",i,Text_get(&t));
  }
Text_destroy(&t);
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
void (*tasks[])(int)={&task0,&task1,&task2,&task3,&task4,&task5};
int taskId=0;
if(argc>1)
  {
  sscanf(argv[1],"%d",&taskId);
  taskId%=sizeof(tasks)/sizeof(tasks[0]);
  }
Platform_msg("task %d\n",taskId);
(*tasks[taskId])(Platform_cpuCtx());
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
