//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "simu.h"

#if !defined AVOID_RENDERING

#include "TransGl/viewer.h"
#include "TransGl/glDraw.h"
#include "TransProg/r32Vector.h"

#define COLOR_COUNT (3*3*3)

typedef struct _RenderingData
{
real32_t colors[COLOR_COUNT][4];
GlPointSphere ps;
PtrVector viewers;
int viewerCount;
} RenderingData;

static RenderingData rd;

typedef struct _ThreadViewerData
{
real64_t xSeen,ySeen,zSeen;
R32Vector seen;
} ThreadViewerData;

typedef struct _MyViewer
{
Viewer v;
bool close;
intptr_t seenCount;
real64_t xSeen,ySeen,zSeen;
ThreadViewerData **tvd[2];
} MyViewer;

static
void
addViewer(ViewerCallback callback)
{
MyViewer *mv=Memory_alloc(1,MyViewer);
int w=600;
int h=600;
int sw=GlView_screenWidth()-w;
int sh=GlView_screenHeight()-h;
int x=sw-((rd.viewerCount*50)%sw);
int y=(rd.viewerCount*50)%sh;
Viewer_init(&mv->v,x,y,w,h,false,callback,(void *)0);
char buffer[0x100];
sprintf(buffer,"Viewer %d",++rd.viewerCount);
GlView_setTitle(Viewer_view(&mv->v),buffer);
SpatialBasis_translate(Viewer_viewPoint(&mv->v),-2.0*simu.size,0.0,0.0);
Perspective_setNearDistance(Viewer_perspective(&mv->v),0.01*simu.size);
Perspective_setFarDistance(Viewer_perspective(&mv->v),10*simu.size);
mv->close=false;
mv->seenCount=0;
mv->xSeen=0.0;
mv->ySeen=0.0;
mv->zSeen=0.0;
mv->tvd[0]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
mv->tvd[1]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  ThreadViewerData *tvd=Memory_alloc(1,ThreadViewerData);
  R32Vector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[0][i]=tvd;
  tvd=Memory_alloc(1,ThreadViewerData);
  R32Vector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[1][i]=tvd;
  }
PtrVector_push(&rd.viewers,mv);
}

static
void
viewerCB(Viewer *v)
{
MyViewer *mv=STRUCT_OF_MEMBER(v,MyViewer,v);
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=mv->seenCount ? 1.0/mv->seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*mv->xSeen,d*mv->ySeen,d*mv->zSeen);
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"v"))
        {
        addViewer(&viewerCB);
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"\x1B"))
        {
        mv->close=true;
        }
      else if(!strcmp(GfxState_key(state),"q"))
        {
        simu.mustQuit=true;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    mv->close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    mv->xSeen=mv->ySeen=mv->zSeen=0.0;
    mv->seenCount=0;
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    real32_t ambient=0.2f;
    real32_t specular=0.2f;
    real32_t shininess=20.0f;
    const real32_t *fog=(const real32_t *)0;
    real32_t headLight[3]={0.0f,0.0f,0.0f};
    int fromState=Scheduler_fromState(&simu.sched);
    for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
      {
      ThreadViewerData *tvd=mv->tvd[fromState][i];
      mv->xSeen+=tvd->xSeen;
      mv->ySeen+=tvd->ySeen;
      mv->zSeen+=tvd->zSeen;
      R32Vector *seen=&tvd->seen;
      intptr_t seenCount=R32Vector_size(seen)/3;
      mv->seenCount+=seenCount;
      rd.ps.vertexCount=seenCount;
      glBindBuffer(GL_ARRAY_BUFFER,GlPointSphere_vertexBuffer(&rd.ps));
      glBufferSubData(GL_ARRAY_BUFFER,0,
                      rd.ps.vertexCount*3*sizeof(real32_t),
                      R32Vector_data(&tvd->seen));
      glBindBuffer(GL_ARRAY_BUFFER,0);
      GlPointSphere_draw(&rd.ps,
                         rd.colors[i%COLOR_COUNT],ambient,specular,shininess,
                         0.5f*(real32_t)DIAMETER,
                         headLight,fog,projMtx,viewMtx);
      }
    break;
    }
  default:
    {
    break;
    }
  }
}

void
Rendering_showEntity(MyEntity *e)
{
int cpuCtx=Activity_cpuCtx(&e->act);
int toState=Scheduler_toState(&simu.sched);
real64_t x=e->x,y=e->y,z=e->z;
for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
  {
  const MyViewer *mv=(const MyViewer *)PtrVector_get(&rd.viewers,i);
  if(Viewer_view(&mv->v)&&
     GfxState_visible(Viewer_state(&mv->v))) // FIXME: race condition
    {
    real64_t lx=x,ly=y,lz=z;
    SpatialBasis_globalToLocalPosition(Viewer_viewPoint(&mv->v),&lx,&ly,&lz);
    if(Perspective_locateVertex(Viewer_perspective(&mv->v),lx,ly,lz)
       ==LOC_MIDDLE)
      {
      ThreadViewerData *tvd=mv->tvd[toState][cpuCtx];
      R32Vector_push(&tvd->seen,(real32_t)x);
      R32Vector_push(&tvd->seen,(real32_t)y);
      R32Vector_push(&tvd->seen,(real32_t)z);
      tvd->xSeen+=x;
      tvd->ySeen+=y;
      tvd->zSeen+=z;
      }
    }
  }
}

void
Rendering_activate(int cpuCtx,
                   void *data)
{
(void)data;
bool needsGfxSync=false;
int toState=Scheduler_toState(&simu.sched);
if(simu.needsRedraw[toState])
  {
  for(intptr_t i=PtrVector_size(&rd.viewers);i--;)
    {
    MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
    ThreadViewerData *tvd=mv->tvd[toState][cpuCtx];
    R32Vector_clear(&tvd->seen);
    tvd->xSeen=0.0;
    tvd->ySeen=0.0;
    tvd->zSeen=0.0;
    if(!cpuCtx&&mv->close)
      {
      PtrVector_fastErase(&rd.viewers,i); // FIXME: race condition
      if(!PtrVector_size(&rd.viewers))
        {
        simu.mustQuit=true;
        }
      Viewer_destroy(&mv->v);
      for(int c=0,c_end=Platform_cpuCtxCount();c<c_end;++c)
        {
        R32Vector_destroy(&mv->tvd[0][c]->seen);
        Memory_free(mv->tvd[0][c]);
        R32Vector_destroy(&mv->tvd[1][c]->seen);
        Memory_free(mv->tvd[1][c]);
        }
      Memory_free(mv->tvd[0]);
      Memory_free(mv->tvd[1]);
      Memory_free(mv);
      }
    }
  }
if(!cpuCtx&&simu.needsRedraw[Scheduler_fromState(&simu.sched)])
  {
  for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
    {
    MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
    GlView_postRedraw(Viewer_view(&mv->v));
    }
  GlView_processEvents(0);
  needsGfxSync=true;
  }
Scheduler_activate(cpuCtx,&simu.sched);
if(needsGfxSync)
  {
  GlView_sync();
  }
}

void
Rendering_init(void)
{
GlView_init();
GlDraw_init();
//-- Colors --
for(int blue=0;blue<3;++blue)
  {
  for(int green=0;green<3;++green)
    {
    for(int red=0;red<3;++red)
      {
      real32_t *c=rd.colors[blue+3*(green+3*red)];
      c[0]=0.2f+0.3f*red;
      c[1]=0.2f+0.3f*green;
      c[2]=0.8f-0.3f*blue;
      c[3]=1.0f;
      }
    }
  }
//-- PointSpheres --
GlPointSphere_init(&rd.ps,0,simu.entityCount,(int *)0,(int *)0,(int *)0);
//-- Viewer --
PtrVector_init(&rd.viewers);
rd.viewerCount=0;
addViewer(&viewerCB);
}

#endif // !AVOID_RENDERING

//----------------------------------------------------------------------------
