//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "simu.h"

Simulation simu;

TRANSPROG_INLINE
void
generateSpeed(int cpuCtx,
              real64_t *out_sx,
              real64_t *out_sy,
              real64_t *out_sz)
{
real64_t speed=(0.5*DIAMETER/DT)*Random_real64(cpuCtx);
real64_t yaw=M_2PI*Random_real64(cpuCtx);
real64_t pitch=M_PI*(Random_real64(cpuCtx)-0.5);
sphericalToCart3D64(speed,yaw,pitch,out_sx,out_sy,out_sz);
}

ActivityResult MyEntity_activate(Activity *a,Scheduler *s);

void
MyEntity_init(MyEntity *e,
              real64_t x,
              real64_t y,
              real64_t z,
              real64_t sx,
              real64_t sy,
              real64_t sz,
              int cpuCtx)
{
SpinSynchro_init(&e->sync);
e->dead=false;
e->x=x;
e->y=y;
e->z=z;
e->sx=sx;
e->sy=sy;
e->sz=sz;
e->key=SpatialGrid_coordToKey(&simu.grid,x,y,z);
Activity_init(&e->act,&MyEntity_activate,&simu.sched,cpuCtx);
Cell_add(simu.cells+e->key,e);
}

MyEntity *
MyEntity_new(real64_t x,
             real64_t y,
             real64_t z,
             real64_t sx,
             real64_t sy,
             real64_t sz,
             int cpuCtx)
{
MyEntity *e=Memory_alloc(1,MyEntity);
MyEntity_init(e,x,y,z,sx,sy,sz,cpuCtx);
return e;
}

void
MyEntity_destroy(MyEntity *e)
{
Activity_destroy(&e->act,&simu.sched);
SpinSynchro_destroy(&e->sync);
Cell_remove(simu.cells+e->key,e);
}

void
MyEntity_delete(MyEntity *e)
{
MyEntity_destroy(e);
Memory_free(e);
}

TRANSPROG_INLINE
MyEntity * // clone or 0
MyEntity_numaClone(MyEntity *e)
{
MyEntity *clone=(MyEntity *)0;
int cpuCtx=Activity_cpuCtx(&e->act);
intptr_t *numaUsage=Scheduler_initNumaUsage(&simu.sched,cpuCtx);
Memory_numaUsage(e,numaUsage);
if(Memory_numaBalance(Platform_memNumaCtx(cpuCtx),numaUsage)<0.5)
  {
  clone=Memory_alloc(1,MyEntity);
  SpinSynchro_init(&clone->sync);
  clone->dead=e->dead;
  clone->x=e->x;
  clone->y=e->y;
  clone->z=e->z;
  clone->sx=e->sx;
  clone->sy=e->sy;
  clone->sz=e->sz;
  clone->key=e->key;
  Activity_initCopy(&clone->act,&simu.sched,&e->act);
  Cell_add(simu.cells+clone->key,clone);
  }
return clone;
}

ActivityResult
MyEntity_activate(Activity *a,
                  Scheduler *s)
{
MyEntity *e=STRUCT_OF_MEMBER(a,MyEntity,act);
if(!SpinSynchro_tryLockW(&e->sync))
  {
  return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
  }
if(e->dead)
  {
  MyEntity_delete(e);
  return ACT_DEAD;
  }
#if !defined AVOID_RENDERING
  Rendering_showEntity(e);
#endif
int cpuCtx=Activity_cpuCtx(a);
real64_t x=e->x+e->sx*DT;
real64_t y=e->y+e->sy*DT;
real64_t z=e->z+e->sz*DT;
real64_t simuSize=simu.size;
real64_t simuHalfSize=0.5*simuSize;
if(x>simuHalfSize) x-=simuSize;
else if(x<-simuHalfSize) x+=simuSize;
if(y>simuHalfSize) y-=simuSize;
else if(y<-simuHalfSize) y+=simuSize;
if(z>simuHalfSize) z-=simuSize;
else if(z<-simuHalfSize) z+=simuSize;
MyEntity *e2=(MyEntity *)0;
real64_t d2=DIAMETER*DIAMETER;
SpatialGridNeighbourhood sgn;
for(intptr_t key=SpatialGridNeighbourhood_start(&sgn,&simu.grid,
                                                x,y,z,DIAMETER);
    key>=0;key=SpatialGridNeighbourhood_next(&sgn))
  {
  Cell *cell=simu.cells+key;
  PtrVector *cellContent=&cell->content;
  intptr_t contentSize=PtrVector_size(cellContent);
  if(!contentSize)
    {
    continue;
    }
  if(!SpinSynchro_tryLockR(&cell->sync))
    {
    if(e2)
      {
      SpinSynchro_unlockR(&e2->sync);
      }
    SpinSynchro_unlockW(&e->sync);
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
  for(intptr_t i=0;i<contentSize;++i)
    {
    MyEntity *e3=(MyEntity *)PtrVector_get(cellContent,i);
    if((e3==e)||(e3==e2))
      {
      continue;
      }
    else
      {
      if(simu.neighbourAffinity)
        {
        const int e3Cpu=Activity_cpuCtx(&e3->act);
        Activity_adjustAffinity(&e->act,s,e3Cpu,
                                Platform_cpuProximity(cpuCtx,e3Cpu));
        }
      if(!SpinSynchro_tryLockR(&e3->sync))
        {
        SpinSynchro_unlockR(&cell->sync);
        if(e2)
          {
          SpinSynchro_unlockR(&e2->sync);
          }
        SpinSynchro_unlockW(&e->sync);
        return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
        }
      if(e3->dead)
        {
        SpinSynchro_unlockR(&e3->sync);
        }
      else
        {
        real64_t dx3=e3->x-x;
        real64_t dy3=e3->y-y;
        real64_t dz3=e3->z-z;
        real64_t d3=dx3*dx3+dy3*dy3+dz3*dz3;
        if((d3<d2)&&
           ((e3->sx*e->sx+e3->sy*e->sy+e3->sz*e->sz)<0.0))
          {
          if(e2)
            {
            SpinSynchro_unlockR(&e2->sync);
            }
          e2=e3;
          d2=d3;
          }
        else
          {
          SpinSynchro_unlockR(&e3->sync);
          }
        }
      }
    }
  SpinSynchro_unlockR(&cell->sync);
  }
if(e2)
  {
  if(!SpinSynchro_tryUpgrade(&e2->sync))
    {
    SpinSynchro_unlockR(&e2->sync);
    SpinSynchro_unlockW(&e->sync);
    return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
    }
  generateSpeed(cpuCtx,&e->sx,&e->sy,&e->sz);
  generateSpeed(cpuCtx,&e2->sx,&e2->sy,&e2->sz);
#if 0
  switch(Random_integer(cpuCtx,20))
    {
    case 0:
      {
      e2->dead=true;
      break;
      }
    case 1:
      {
      real64_t sx,sy,sz;
      generateSpeed(cpuCtx,&sx,&sy,&sz);
      MyEntity_new(0.5*(x+e2->x),
                   0.5*(y+e2->y),
                   0.5*(z+e2->z),
                   sx,sy,sz,
                   cpuCtx);
      break;
      }
    }
#endif
  SpinSynchro_unlockW(&e2->sync);
  }
e->x=x;
e->y=y;
e->z=z;
intptr_t key=SpatialGrid_coordToKey(&simu.grid,e->x,e->y,e->z);
if(e->key!=key)
  {
  Cell_add(simu.cells+key,e);
  Cell_remove(simu.cells+e->key,e);
  e->key=key;
  }
if(simu.spatialAffinity)
  {
  Activity_adjustAffinity(a,s,SpatialAffinity_chooseCpu(&simu.spAff,x,y,z),1);
  }
if(simu.numaMigration&&
   !Random_int(cpuCtx,simu.numaMigration)&&
   MyEntity_numaClone(e))
  {
  MyEntity_delete(e);
  return ACT_DEAD;
  }
SpinSynchro_unlockW(&e->sync);
return ACT_DONE;
}

void
createEntities(int cpuCtx,
               void *data)
{
(void)data;
intptr_t begin=Platform_splitBegin(0,simu.entityCount,
                                   Platform_cpuCtxCount(),cpuCtx);
intptr_t end=Platform_splitEnd(0,simu.entityCount,
                               Platform_cpuCtxCount(),cpuCtx);
for(intptr_t i=begin;i<end;++i)
  {
  real64_t x=simu.size*(Random_real64(cpuCtx)-0.5);
  real64_t y=simu.size*(Random_real64(cpuCtx)-0.5);
  real64_t z=simu.size*(Random_real64(cpuCtx)-0.5);
  real64_t sx,sy,sz;
  generateSpeed(cpuCtx,&sx,&sy,&sz);
  MyEntity_new(x,y,z,sx,sy,sz,
               (int)(simu.spatialAffinity
                     ? SpatialAffinity_chooseCpu(&simu.spAff,x,y,z)
                     : Random_int(cpuCtx,Platform_cpuCtxCount())));
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
simu.entityCount=100000;
simu.stepCount=INT64_MAX;
simu.neighbourAffinity=true;
simu.spatialAffinity=false;
#if !defined AVOID_RENDERING
  simu.needsRedraw[0]=false;
  simu.needsRedraw[1]=false;
#endif
simu.mustQuit=false;
simu.numaMigration=20;
simu.size=50.0;
Scheduler_init(&simu.sched);
int prefetch;
for(int i=1,i_end=Platform_argc();i<i_end;++i)
  {
  if(!strcmp(Platform_argv(i),"-e")&&
     (i+1<Platform_argc()))
    {
    sscanf(Platform_argv(++i),"%"FPTR"d",&simu.entityCount);
    }
  else if(!strcmp(Platform_argv(i),"-n")&&
          (i+1<Platform_argc()))
    {
    sscanf(Platform_argv(++i),"%"F64"d",&simu.stepCount);
    }
  else if(!strcmp(Platform_argv(i),"-v")&&
          (i+1<Platform_argc()))
    {
    sscanf(Platform_argv(++i),"%lg",&simu.size);
    }
  else if(!strcmp(Platform_argv(i),"-m")&&
          (i+1<Platform_argc()))
    {
    sscanf(Platform_argv(++i),"%d",&simu.numaMigration);
    }
  else if(!strcmp(Platform_argv(i),"-p")&&
          (i+1<Platform_argc())&&
          (sscanf(Platform_argv(i+1),"%d",&prefetch)==1))
    {
    Scheduler_setPrefetch(&simu.sched,prefetch);
    ++i;
    }
  else if(!strcmp(Platform_argv(i),"-ws"))
    {
    Scheduler_setWorkStealing(&simu.sched,false);
    }
  else if(!strcmp(Platform_argv(i),"-na"))
    {
    simu.neighbourAffinity=false;
    }
  else if(!strcmp(Platform_argv(i),"-sa"))
    {
    simu.spatialAffinity=true;
    }
  else
    {
    Platform_msg("ignoring argument '%s'\n",Platform_argv(i));
    }
  }
SpatialAffinity_init(&simu.spAff,-0.5*simu.size,0.5*simu.size,
                                 -0.5*simu.size,0.5*simu.size,
                                 -0.5*simu.size,0.5*simu.size);
Platform_msg("scheduler prefetch: %d\n",(int)Scheduler_prefetch(&simu.sched));
Platform_msg("work stealing: %d\n",Scheduler_workStealing(&simu.sched));
Platform_msg("neighbour affinity: %d\n",simu.neighbourAffinity);
Platform_msg("spatial affinity: %d\n",simu.spatialAffinity);
if(Platform_numaCtxCount()<2)
  {
  simu.numaMigration=0;
  }
Platform_msg("numa migration: %d\n",simu.numaMigration);
Platform_msg("cpu/mem: %d/%d\n",
             Platform_cpuCtxCount(),Platform_memCtxCount());
Platform_msg("volume: %g^3 (%dx%dx%d)\n",
             simu.size,
             SpatialAffinity_xSplit(&simu.spAff),
             SpatialAffinity_ySplit(&simu.spAff),
             SpatialAffinity_zSplit(&simu.spAff));
Platform_msg("entities/cell: %g\n",simu.entityCount/pow(simu.size/DIAMETER,3));
SpatialGrid_init(&simu.grid,DIAMETER,
                 (intptr_t)simu.size,
                 (intptr_t)simu.size,
                 (intptr_t)simu.size);
simu.cells=Memory_alloc(SpatialGrid_cellCount(&simu.grid),Cell);
for(intptr_t i=0,i_end=SpatialGrid_cellCount(&simu.grid);i<i_end;++i)
  {
  SpinSynchro_init(&simu.cells[i].sync);
  PtrVector_init(&simu.cells[i].content);
  }
Platform_msg("entities: %"FPTR"d\n",simu.entityCount);
#if 0
  int cpuCtx=Platform_cpuCtx();
  for(intptr_t i=0,i_end=simu.entityCount;i<i_end;++i)
    {
    real64_t x=simu.size*(Random_real64(cpuCtx)-0.5);
    real64_t y=simu.size*(Random_real64(cpuCtx)-0.5);
    real64_t z=simu.size*(Random_real64(cpuCtx)-0.5);
    real64_t sx,sy,sz;
    generateSpeed(0,&sx,&sy,&sz);
    MyEntity_new(x,y,z,sx,sy,sz,
                 simu.spatialAffinity
                 ? SpatialAffinity_chooseCpu(&simu.spAff,x,y,z)
                 : Random_int(cpuCtx,Platform_cpuCtxCount()));
    }
#else
  Platform_parallel(&createEntities,(void *)0);
#endif
Platform_msg("steps: %"F64"u\n",simu.stepCount);
#if !defined AVOID_RENDERING
  Rendering_init();
#endif
int64_t now=Platform_time();
int64_t lastPerf=now;
int perfCount=0;
#if !defined AVOID_RENDERING
  int64_t lastDraw=now;
#endif
while(!simu.mustQuit&&(Scheduler_step(&simu.sched)<simu.stepCount))
  {
  ++perfCount;
  now=Platform_time();
  int64_t dt=now-lastPerf;
  if(dt>=2000000)
    {
    Platform_msg("%g\n",perfCount/(1e-6*dt));
    lastPerf+=dt;
    perfCount=0;
    }
  Scheduler_advanceStep(&simu.sched);
#if defined AVOID_RENDERING
  Platform_parallel(&Scheduler_activate,&simu.sched);
#else
  dt=now-lastDraw;
  int toState=Scheduler_toState(&simu.sched);
  simu.needsRedraw[toState]=(dt>=40000) ?  true : false;
  if(simu.needsRedraw[toState])
    {
    lastDraw+=dt;
    }
  Platform_parallel(&Rendering_activate,(void *)0);
#endif
  }
for(intptr_t i=0,i_end=SpatialGrid_cellCount(&simu.grid);i<i_end;++i)
  {
  SpinSynchro_destroy(&simu.cells[i].sync);
  PtrVector_destroy(&simu.cells[i].content);
  }
Memory_free(simu.cells);
SpatialGrid_destroy(&simu.grid);
SpatialAffinity_destroy(&simu.spAff);
Scheduler_destroy(&simu.sched);
return 0;
}

//----------------------------------------------------------------------------
