//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSCUDA_CUDA_H
#define TRANSCUDA_CUDA_H 1

#include "TransCuda/api.h"

#include "TransProg/text.h"

#include <cuda.h>

#define CUDA_DROP_DEAD(msg,err)                                    \
        do                                                         \
          {                                                        \
          char __errBuf[0x100];                                    \
          Cuda_errorMsg((err),__errBuf,0x100);                     \
          Platform_msg("\n%s:%d:%s()\n  %s --- %s\n",              \
                       __FILE__,__LINE__,__func__,(msg),__errBuf); \
          Platform_dropDead();                                     \
          } while(0)

typedef struct _CudaDevice
{
CUdevice id;
CUcontext context;
CUstream stream;
uint64_t peerMask;
char *name;
int computeCapabilityMajor;
int computeCapabilityMinor;
size_t totalMemory;
int max_threads_per_block;
int max_block_dim_x;
int max_block_dim_y;
int max_block_dim_z;
int max_grid_dim_x;
int max_grid_dim_y;
int max_grid_dim_z;
int max_shared_memory_per_block;
int total_constant_memory;
int warp_size;
// int max_pitch;
// int maximum_texture1d_width;
// int maximum_texture1d_linear_width;
// int maximum_texture2d_width;
// int maximum_texture2d_height;
// int maximum_texture2d_linear_width;
// int maximum_texture2d_linear_height;
// int maximum_texture2d_linear_pitch;
// int maximum_texture3d_width;
// int maximum_texture3d_height;
// int maximum_texture3d_depth;
// int maximum_texture3d_width_alternate;
// int maximum_texture3d_height_alternate;
// int maximum_texture3d_depth_alternate;
// int maximum_texturecubemap_width;
// int maximum_texture1d_layered_width;
// int maximum_texture1d_layered_layers;
// int maximum_texture2d_layered_width;
// int maximum_texture2d_layered_height;
// int maximum_texture2d_layered_layers;
// int maximum_texturecubemap_layered_width;
// int maximum_texturecubemap_layered_layers;
// int maximum_surface1d_width;
// int maximum_surface2d_width;
// int maximum_surface2d_height;
// int maximum_surface3d_width;
// int maximum_surface3d_height;
// int maximum_surface3d_depth;
// int maximum_surface1d_layered_width;
// int maximum_surface1d_layered_layers;
// int maximum_surface2d_layered_width;
// int maximum_surface2d_layered_height;
// int maximum_surface2d_layered_layers;
// int maximum_surfacecubemap_width;
// int maximum_surfacecubemap_layered_width;
// int maximum_surfacecubemap_layered_layers;
int max_registers_per_block;
int clock_rate;
// int texture_alignment;
// int texture_pitch_alignment;
int gpu_overlap;
int multiprocessor_count;
int core_count;
int kernel_exec_timeout;
int integrated;
int can_map_host_memory;
// int compute_mode;
int concurrent_kernels;
int ecc_enabled;
int pci_bus_id;
int pci_device_id;
int tcc_driver;
int memory_clock_rate;
int global_memory_bus_width;
int l2_cache_size;
int max_threads_per_multiprocessor;
int unified_addressing;
} CudaDevice;

typedef struct _CudaMarker
{
const CudaDevice *device;
CUevent event;
} CudaMarker;

typedef struct _CudaProgram
{
const CudaDevice *device;
char *name;
char *options;
char *sourceCode;
uint8_t *binaryCode;
intptr_t binaryCodeSize;
bool prefersCacheToShared;
Text buildLog;
CUmodule module;
CUfunction kernel;
int max_threads_per_block;
int shared_size_bytes;
int const_size_bytes;
int local_size_bytes;
int num_regs;
int ptx_version;
int binary_version;
} CudaProgram;

typedef struct _CudaBuffer
{
const CudaDevice *device;
CUdeviceptr devptr;
size_t size;
CUgraphicsResource gfxResource;
int gl;
} CudaBuffer;

typedef struct _CudaLockedMem
{
void *data;
intptr_t size;
bool writeOnly;
CUdeviceptr devptr;
} CudaLockedMem;

TRANSCUDA_API
void
Cuda_init(void);

TRANSCUDA_API
void
Cuda_errorMsg(int errorCode,
              char *out_msg,
              int msgSize);

TRANSCUDA_API
const CudaDevice *
Cuda_devices(int *out_deviceCount);

TRANSCUDA_API
const CudaDevice * // device used for OpenGL rendering or 0
Cuda_glDevice(void);

TRANSPROG_INLINE
int
CudaDevice_id(const CudaDevice *device)
{
return device->id;
}

TRANSPROG_INLINE
const char *
CudaDevice_name(const CudaDevice *device)
{
return device->name;
}

TRANSPROG_INLINE
int
CudaDevice_computeCapabilityMajor(const CudaDevice *device)
{
return device->computeCapabilityMajor;
}

TRANSPROG_INLINE
int
CudaDevice_computeCapabilityMinor(const CudaDevice *device)
{
return device->computeCapabilityMinor;
}

TRANSPROG_INLINE
intptr_t
CudaDevice_totalMemory(const CudaDevice *device)
{
return device->totalMemory;
}

TRANSCUDA_API
intptr_t
CudaDevice_freeMemory(const CudaDevice *device);

TRANSPROG_INLINE
int
CudaDevice_maxThreadsPerBlock(const CudaDevice *device)
{
return device->max_threads_per_block;
}

TRANSPROG_INLINE
int
CudaDevice_maxBlockDimX(const CudaDevice *device)
{
return device->max_block_dim_x;
}

TRANSPROG_INLINE
int
CudaDevice_maxBlockDimY(const CudaDevice *device)
{
return device->max_block_dim_y;
}

TRANSPROG_INLINE
int
CudaDevice_maxBlockDimZ(const CudaDevice *device)
{
return device->max_block_dim_z;
}

TRANSPROG_INLINE
int
CudaDevice_maxGridDimX(const CudaDevice *device)
{
return device->max_grid_dim_x;
}

TRANSPROG_INLINE
int
CudaDevice_maxGridDimY(const CudaDevice *device)
{
return device->max_grid_dim_y;
}

TRANSPROG_INLINE
int
CudaDevice_maxGridDimZ(const CudaDevice *device)
{
return device->max_grid_dim_z;
}

TRANSPROG_INLINE
int
CudaDevice_maxSharedMemoryPerBlock(const CudaDevice *device)
{
return device->max_shared_memory_per_block;
}

TRANSPROG_INLINE
int
CudaDevice_totalConstantMemory(const CudaDevice *device)
{
return device->total_constant_memory;
}

TRANSPROG_INLINE
int
CudaDevice_warpSize(const CudaDevice *device)
{
return device->warp_size;
}

TRANSPROG_INLINE
int
CudaDevice_maxRegistersPerBlock(const CudaDevice *device)
{
return device->max_registers_per_block;
}

TRANSPROG_INLINE
int // kHz
CudaDevice_clockRate(const CudaDevice *device)
{
return device->clock_rate;
}

TRANSPROG_INLINE
bool
CudaDevice_gpuOverlap(const CudaDevice *device)
{
return device->gpu_overlap ? true : false;
}

TRANSPROG_INLINE
int
CudaDevice_multiprocessorCount(const CudaDevice *device)
{
return device->multiprocessor_count;
}

TRANSPROG_INLINE
int // core count or 0 if unknown
CudaDevice_coreCount(const CudaDevice *device)
{
return device->core_count;
}

TRANSPROG_INLINE
bool
CudaDevice_kernelExecTimeout(const CudaDevice *device)
{
return device->kernel_exec_timeout ? true : false;
}

TRANSPROG_INLINE
bool
CudaDevice_integrated(const CudaDevice *device)
{
return device->integrated ? true : false;
}

TRANSPROG_INLINE
bool
CudaDevice_canMapHostMemory(const CudaDevice *device)
{
return device->can_map_host_memory ? true : false;
}

TRANSPROG_INLINE
bool
CudaDevice_concurrentKernels(const CudaDevice *device)
{
return device->concurrent_kernels ? true : false;
}

TRANSPROG_INLINE
bool
CudaDevice_eccEnabled(const CudaDevice *device)
{
return device->ecc_enabled ? true : false;
}

TRANSPROG_INLINE
int
CudaDevice_pciBusId(const CudaDevice *device)
{
return device->pci_bus_id;
}

TRANSPROG_INLINE
int
CudaDevice_pciDeviceId(const CudaDevice *device)
{
return device->pci_device_id;
}

TRANSPROG_INLINE
bool
CudaDevice_tccDriver(const CudaDevice *device)
{
return device->tcc_driver ? true : false;
}

TRANSPROG_INLINE
int // kHz
CudaDevice_memoryClockRate(const CudaDevice *device)
{
return device->memory_clock_rate;
}

TRANSPROG_INLINE
int
CudaDevice_globalMemoryBusWidth(const CudaDevice *device)
{
return device->global_memory_bus_width;
}

TRANSPROG_INLINE
int
CudaDevice_l2CacheSize(const CudaDevice *device)
{
return device->l2_cache_size;
}

TRANSPROG_INLINE
int
CudaDevice_maxThreadsPerMultiprocessor(const CudaDevice *device)
{
return device->max_threads_per_multiprocessor;
}

TRANSPROG_INLINE
bool
CudaDevice_unifiedAddressing(const CudaDevice *device)
{
return device->unified_addressing ? true : false;
}

TRANSCUDA_API
void
CudaDevice_sync(const CudaDevice *device);

TRANSCUDA_API
void
CudaMarker_init(CudaMarker *marker,
                const CudaDevice *device);

TRANSPROG_INLINE
CudaMarker *
CudaMarker_new(const CudaDevice *device)
{
CudaMarker *marker=Memory_alloc(1,CudaMarker);
CudaMarker_init(marker,device);
return marker;
}

TRANSCUDA_API
void
CudaMarker_destroy(CudaMarker *marker);

TRANSPROG_INLINE
void
CudaMarker_delete(CudaMarker *marker)
{
CudaMarker_destroy(marker);
Memory_free(marker);
}

TRANSPROG_INLINE
const CudaDevice *
CudaMarker_device(const CudaMarker *marker)
{
return marker->device;
}

TRANSCUDA_API
void
CudaMarker_set(CudaMarker *marker);

TRANSCUDA_API
bool // previous work is done
CudaMarker_test(const CudaMarker *marker);

TRANSCUDA_API
void
CudaMarker_sync(CudaMarker *marker);

TRANSCUDA_API
int64_t // microseconds
CudaMarker_duration(const CudaMarker *marker,
                    const CudaMarker *previous);

TRANSCUDA_API
void
CudaProgram_init(CudaProgram *program,
                 const CudaDevice *device,
                 const char *name,
                 const char *options,
                 const char *sourceCode,
                 bool prefersCacheToShared);

TRANSPROG_INLINE
CudaProgram *
CudaProgram_new(const CudaDevice *device,
                const char *name,
                const char *options,
                const char *sourceCode,
                bool prefersCacheToShared)
{
CudaProgram *program=Memory_alloc(1,CudaProgram);
CudaProgram_init(program,device,name,options,sourceCode,prefersCacheToShared);
return program;
}

TRANSCUDA_API
void
CudaProgram_initFromBinary(CudaProgram *program,
                           const CudaDevice *device,
                           const char *name,
                           const void *binaryCode,
                           intptr_t binaryCodeSize,
                           bool prefersCacheToShared);

TRANSPROG_INLINE
CudaProgram *
CudaProgram_newFromBinary(const CudaDevice *device,
                          const char *name,
                          const void *binaryCode,
                          intptr_t binaryCodeSize,
                          bool prefersCacheToShared)
{
CudaProgram *program=Memory_alloc(1,CudaProgram);
CudaProgram_initFromBinary(program,device,name,binaryCode,binaryCodeSize,
                           prefersCacheToShared);
return program;
}

TRANSCUDA_API
void
CudaProgram_destroy(CudaProgram *program);

TRANSPROG_INLINE
void
CudaProgram_delete(CudaProgram *program)
{
CudaProgram_destroy(program);
Memory_free(program);
}

TRANSPROG_INLINE
const CudaDevice *
CudaProgram_device(const CudaProgram *program)
{
return program->device;
}

TRANSPROG_INLINE
const char *
CudaProgram_name(const CudaProgram *program)
{
return program->name;
}

TRANSPROG_INLINE
const char *
CudaProgram_options(const CudaProgram *program)
{
return program->options;
}

TRANSPROG_INLINE
const char *
CudaProgram_sourceCode(const CudaProgram *program)
{
return program->sourceCode;
}

TRANSPROG_INLINE
const void *
CudaProgram_binaryCode(const CudaProgram *program,
                       intptr_t *binaryCodeSize)
{
*binaryCodeSize=program->binaryCodeSize;
return (const void *)program->binaryCode;
}

TRANSPROG_INLINE
bool
CudaProgram_prefersCacheToShared(const CudaProgram *program)
{
return program->prefersCacheToShared;
}

TRANSPROG_INLINE
bool
CudaProgram_buildFailure(const CudaProgram *program)
{
return !program->kernel;
}

TRANSPROG_INLINE
const char * // build log or 0
CudaProgram_buildLog(const CudaProgram *program)
{
return Text_length(&program->buildLog) ? Text_get(&program->buildLog)
                                       : (const char *)0;
}

TRANSPROG_INLINE
int
CudaProgram_maxThreadsPerBlock(const CudaProgram *program)
{
return program->max_threads_per_block;
}

TRANSPROG_INLINE
int
CudaProgram_sharedSizeBytes(const CudaProgram *program)
{
return program->shared_size_bytes;
}

TRANSPROG_INLINE
int
CudaProgram_constSizeBytes(const CudaProgram *program)
{
return program->const_size_bytes;
}

TRANSPROG_INLINE
int
CudaProgram_localSizeBytes(const CudaProgram *program)
{
return program->local_size_bytes;
}

TRANSPROG_INLINE
int
CudaProgram_numRegs(const CudaProgram *program)
{
return program->num_regs;
}

TRANSPROG_INLINE
int
CudaProgram_ptxVersion(const CudaProgram *program)
{
return program->ptx_version;
}

TRANSPROG_INLINE
int
CudaProgram_binaryVersion(const CudaProgram *program)
{
return program->binary_version;
}

TRANSCUDA_API
void
CudaProgram_launch(const CudaProgram *program,
                   int gridDimX,
                   int gridDimY,
                   int gridDimZ,
                   int blockDimX,
                   int blockDimY,
                   int blockDimZ,
                   int sharedMemorySize,
                   const void * const *args);

TRANSCUDA_API
void
CudaBuffer_init(CudaBuffer *buffer,
                const CudaDevice *device,
                intptr_t size);

TRANSPROG_INLINE
CudaBuffer *
CudaBuffer_new(const CudaDevice *device,
               intptr_t size)
{
CudaBuffer *buffer=Memory_alloc(1,CudaBuffer);
CudaBuffer_init(buffer,device,size);
return buffer;
}

TRANSCUDA_API
void
CudaBuffer_initFromGl(CudaBuffer *buffer,
                      const CudaDevice *device,
                      int glBuffer);

TRANSPROG_INLINE
CudaBuffer *
CudaBuffer_newFromGl(const CudaDevice *device,
                     int glBuffer)
{
CudaBuffer *buffer=Memory_alloc(1,CudaBuffer);
CudaBuffer_initFromGl(buffer,device,glBuffer);
return buffer;
}

TRANSCUDA_API
void
CudaBuffer_destroy(CudaBuffer *buffer);

TRANSPROG_INLINE
void
CudaBuffer_delete(CudaBuffer *buffer)
{
CudaBuffer_destroy(buffer);
Memory_free(buffer);
}

TRANSPROG_INLINE
const CudaDevice *
CudaBuffer_device(const CudaBuffer *buffer)
{
return buffer->device;
}

TRANSPROG_INLINE
const void * // buffer as program argument
CudaBuffer_programArg(const CudaBuffer *buffer)
{
return (const void *)&buffer->devptr;
}

TRANSPROG_INLINE
const void * // null buffer as program argument
CudaBuffer_nullProgramArg(void)
{
extern TRANSCUDA_API_VAR const void *_CudaBuffer_nullArg;
return (const void *)&_CudaBuffer_nullArg;
}

TRANSPROG_INLINE
intptr_t
CudaBuffer_size(const CudaBuffer *buffer)
{
return buffer->size;
}

TRANSPROG_INLINE
int
CudaBuffer_gl(const CudaBuffer *buffer)
{
return buffer->gl;
}

TRANSCUDA_API
void
CudaBuffer_acquireGl(CudaBuffer *buffer);

TRANSCUDA_API
void
CudaBuffer_releaseGl(CudaBuffer *buffer);

TRANSCUDA_API
void
CudaBuffer_send(CudaBuffer *buffer,
                intptr_t offset,
                intptr_t size,
                const void *data);

TRANSPROG_INLINE
void
CudaBuffer_sendAll(CudaBuffer *buffer,
                   const void *data)
{
CudaBuffer_send(buffer,0,buffer->size,data);
}

TRANSCUDA_API
void
CudaBuffer_receive(CudaBuffer *buffer,
                   intptr_t offset,
                   intptr_t size,
                   void *data);

TRANSPROG_INLINE
void
CudaBuffer_receiveAll(CudaBuffer *buffer,
                      void *data)
{
CudaBuffer_receive(buffer,0,buffer->size,data);
}

TRANSPROG_INLINE
bool // copy from srcBuffer to destBuffer will not involve host
CudaBuffer_directCopyAvailable(const CudaBuffer *destBuffer,
                               const CudaBuffer *srcBuffer)
{
return (destBuffer->device==srcBuffer->device)||
       (destBuffer->device->peerMask&(1ULL<<srcBuffer->device->id));
}

TRANSCUDA_API
void
CudaBuffer_copy(CudaBuffer *destBuffer,
                const CudaBuffer *srcBuffer,
                intptr_t destOffset,
                intptr_t srcOffset,
                intptr_t size);

TRANSCUDA_API
void
CudaLockedMem_init(CudaLockedMem *mem,
                   bool writeOnly,
                   intptr_t size);

TRANSPROG_INLINE
CudaLockedMem *
CudaLockedMem_new(bool writeOnly,
                  intptr_t size)
{
CudaLockedMem *mem=Memory_alloc(1,CudaLockedMem);
CudaLockedMem_init(mem,writeOnly,size);
return mem;
}

TRANSCUDA_API
void
CudaLockedMem_destroy(CudaLockedMem *mem);

TRANSPROG_INLINE
void
CudaLockedMem_delete(CudaLockedMem *mem)
{
CudaLockedMem_destroy(mem);
Memory_free(mem);
}

TRANSPROG_INLINE
void *
CudaLockedMem_data(const CudaLockedMem *mem)
{
return mem->data;
}

TRANSPROG_INLINE
intptr_t
CudaLockedMem_size(const CudaLockedMem *mem)
{
return mem->size;
}

TRANSPROG_INLINE
bool
CudaLockedMem_writeOnly(const CudaLockedMem *mem)
{
return mem->writeOnly;
}

TRANSPROG_INLINE
const void * // memory as zero-copy buffer program argument
CudaLockedMem_programArg(const CudaLockedMem *mem)
{
return (const void *)&mem->devptr;
}

#endif // TRANSCUDA_CUDA_H

//----------------------------------------------------------------------------
