//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSGL_GLDRAW_H
#define TRANSGL_GLDRAW_H 1

#include "TransGl/coreGl.h"

#include "TransProg/text.h"
#include "TransProg/iPtrVector.h"

#define OPENGL_CHECK ON_GL_ERROR({ Platform_dropDead(); })

typedef struct _GlVertexArray
{
int vao;
IPtrVector info;
} GlVertexArray;

typedef enum { GLSHAPE_HAS_NORMALS  =1<<0,
               GLSHAPE_HAS_TEXCOORDS=1<<1,
               GLSHAPE_HAS_COLORS   =1<<2,
               GLSHAPE_HAS_FOG      =1<<3 } GlShapeKindBit;
typedef int GlShapeKind;

typedef struct _GlShape
{
GlShapeKind kind;
int glPrimitive;
intptr_t indexCount;
intptr_t vertexCount;
GlVertexArray vertexArray;
int colors;
int texCoords;
int normals;
int vertices;
int indices;
} GlShape;

typedef enum { GLPOINTSPHERE_HAS_COLORS  =1<<0,
               GLPOINTSPHERE_HAS_RADIUSES=1<<1,
               GLPOINTSPHERE_HAS_FOG     =1<<2 } GlPointSphereKindBit;
typedef int GlPointSphereKind;

typedef struct _GlPointSphere
{
GlPointSphereKind kind;
intptr_t vertexCount;
GlVertexArray vertexArray;
int radiuses;
int colors;
int vertices;
} GlPointSphere;

TRANSGL_API
void
GlDraw_init(void);

TRANSGL_API
int // texture id
GlDraw_loadTexture(int width,
                   int height,
                   int components,
                   const uint8_t *storage,
                   bool roundUp);

TRANSGL_API
int // cube-map texture id
GlDraw_loadCubeMapTextures(int size,
                           int components,
                           const uint8_t *storage[6],
                           bool roundUp);

TRANSGL_API
void
GlDraw_releaseTexture(int textureId);

TRANSGL_API
void
GlDraw_snapshot(int x,
                int y,
                int width,
                int height,
                uint8_t *rgbStorage);

TRANSGL_API
void
GlVertexArray_init(GlVertexArray *vertexArray);

TRANSPROG_INLINE
GlVertexArray *
GlVertexArray_new(void)
{
GlVertexArray *vertexArray=Memory_alloc(1,GlVertexArray);
GlVertexArray_init(vertexArray);
return vertexArray;
}

TRANSGL_API
void
GlVertexArray_destroy(GlVertexArray *vertexArray);

TRANSPROG_INLINE
void
GlVertexArray_delete(GlVertexArray *vertexArray)
{
GlVertexArray_destroy(vertexArray);
Memory_free(vertexArray);
}

TRANSGL_API
int // buffer index
GlVertexArray_manageBuffer(GlVertexArray *vertexArray,
                           int attribute,
                           int components,
                           int glType,
                           int byteStride,
                           int divisor,
                           int buffer,
                           intptr_t byteOffset);

TRANSGL_API
int // buffer index
GlVertexArray_createBufferGeneric(GlVertexArray *vertexArray,
                                  int attribute,
                                  int components,
                                  int divisor,
                                  intptr_t count,
                                  const void *data,
                                  int sizeOfData);

#define GlVertexArray_createBuffer(vertexArray,            \
                                   attribute,              \
                                   components,             \
                                   divisor,                \
                                   count,                  \
                                   data)                   \
        GlVertexArray_createBufferGeneric((vertexArray),   \
                                          (attribute),     \
                                          (components),    \
                                          (divisor),       \
                                          (count),         \
                                          (data),          \
                                          sizeof(*(data)))

TRANSGL_API
void
GlVertexArray_bind(const GlVertexArray *vertexArray);

TRANSGL_API
void
GlVertexArray_unbind(const GlVertexArray *vertexArray);

TRANSPROG_INLINE
int // number of managed buffers
GlVertexArray_count(const GlVertexArray *vertexArray)
{
return (int)IPtrVector_size(&vertexArray->info)/7;
}

TRANSPROG_INLINE
int
GlVertexArray_attribute(const GlVertexArray *vertexArray,
                        int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+0];
}

TRANSPROG_INLINE
int
GlVertexArray_components(const GlVertexArray *vertexArray,
                         int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+1];
}

TRANSPROG_INLINE
int
GlVertexArray_glType(const GlVertexArray *vertexArray,
                     int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+2];
}

TRANSPROG_INLINE
int
GlVertexArray_byteStride(const GlVertexArray *vertexArray,
                         int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+3];
}

TRANSPROG_INLINE
int
GlVertexArray_divisor(const GlVertexArray *vertexArray,
                      int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+4];
}

TRANSPROG_INLINE
int // OpenGL buffer
GlVertexArray_buffer(const GlVertexArray *vertexArray,
                     int bufferIndex)
{
return (int)IPtrVector_data(&vertexArray->info)[7*bufferIndex+5];
}

TRANSPROG_INLINE
intptr_t
GlVertexArray_byteOffset(const GlVertexArray *vertexArray,
                         int bufferIndex)
{
return IPtrVector_data(&vertexArray->info)[7*bufferIndex+6];
}

TRANSGL_API
void
GlShape_initGeneric(GlShape *shape,
                    GlShapeKind kind,
                    int glPrimitive,
                    intptr_t indexCount,
                    const void *indices,
                    int sizeOfIndex,
                    intptr_t vertexCount,
                    const void *vertices,
                    int sizeOfVertex,
                    const void *normals,
                    int sizeOfNormal,
                    const void *texCoords,
                    int sizeOfTexCoord,
                    const void *colors,
                    int sizeOfColor);

#define GlShape_init(shape,                       \
                     kind,                        \
                     glPrimitive,                 \
                     indexCount,                  \
                     indices,                     \
                     vertexCount,                 \
                     vertices,                    \
                     normals,                     \
                     texCoords,                   \
                     colors)                      \
        GlShape_initGeneric((shape),              \
                            (kind),               \
                            (glPrimitive),        \
                            (indexCount),         \
                            (indices),            \
                            sizeof(*(indices)),   \
                            (vertexCount),        \
                            (vertices),           \
                            sizeof(*(vertices)),  \
                            (normals),            \
                            sizeof(*(normals)),   \
                            (texCoords),          \
                            sizeof(*(texCoords)), \
                            (colors),             \
                            sizeof(*(colors)))

TRANSPROG_INLINE
GlShape *
GlShape_newGeneric(GlShapeKind kind,
                   int glPrimitive,
                   intptr_t indexCount,
                   const void *indices,
                   int sizeOfIndex,
                   intptr_t vertexCount,
                   const void *vertices,
                   int sizeOfVertex,
                   const void *normals,
                   int sizeOfNormal,
                   const void *texCoords,
                   int sizeOfTexCoord,
                   const void *colors,
                   int sizeOfColor)
{
GlShape *shape=Memory_alloc(1,GlShape);
GlShape_initGeneric(shape,kind,glPrimitive,indexCount,indices,sizeOfIndex,
                    vertexCount,vertices,sizeOfVertex,normals,sizeOfNormal,
                    texCoords,sizeOfTexCoord,colors,sizeOfColor);
return shape;
}

#define GlShape_new(kind,                        \
                    glPrimitive,                 \
                    indexCount,                  \
                    indices,                     \
                    vertexCount,                 \
                    vertices,                    \
                    normals,                     \
                    texCoords,                   \
                    colors)                      \
        GlShape_newGeneric((kind),               \
                           (glPrimitive),        \
                           (indexCount),         \
                           (indices),            \
                           sizeof(*(indices)),   \
                           (vertexCount),        \
                           (vertices),           \
                           sizeof(*(vertices)),  \
                           (normals),            \
                           sizeof(*(normals)),   \
                           (texCoords),          \
                           sizeof(*(texCoords)), \
                           (colors),             \
                           sizeof(*(colors)))

TRANSGL_API
void
GlShape_destroy(GlShape *shape);

TRANSPROG_INLINE
void
GlShape_delete(GlShape *shape)
{
GlShape_destroy(shape);
Memory_free(shape);
}

TRANSGL_API
void
GlShape_draw(const GlShape *shape,
             const real32_t rgba[4],
             real32_t ambient,
             real32_t specular,
             real32_t shininess,
             const real32_t headLight[3],
             const real32_t fog[4],
             int textureId,
             const real32_t *projectionMatrix,
             const real32_t *modelViewMatrix);

TRANSPROG_INLINE
int // OpenGL buffer for indices
GlShape_indexBuffer(const GlShape *shape)
{
return GlVertexArray_buffer(&shape->vertexArray,shape->indices);
}

TRANSPROG_INLINE
int // OpenGL buffer for vertices
GlShape_vertexBuffer(const GlShape *shape)
{
return GlVertexArray_buffer(&shape->vertexArray,shape->vertices);
}

TRANSPROG_INLINE
int // OpenGL buffer for normals
GlShape_normalBuffer(const GlShape *shape)
{
return GlVertexArray_buffer(&shape->vertexArray,shape->normals);
}

TRANSPROG_INLINE
int // OpenGL buffer for texCoords
GlShape_texCoordBuffer(const GlShape *shape)
{
return GlVertexArray_buffer(&shape->vertexArray,shape->texCoords);
}

TRANSPROG_INLINE
int // OpenGL buffer for colors
GlShape_colorBuffer(const GlShape *shape)
{
return GlVertexArray_buffer(&shape->vertexArray,shape->colors);
}

TRANSGL_API
void
GlPointSphere_initGeneric(GlPointSphere *pointSphere,
                          GlPointSphereKind kind,
                          intptr_t vertexCount,
                          const void *vertices,
                          int sizeOfVertex,
                          const void *colors,
                          int sizeOfColor,
                          const void *radiuses,
                          int sizeOfRadius);

#define GlPointSphere_init(pointSphere,                \
                           kind,                       \
                           vertexCount,                \
                           vertices,                   \
                           colors,                     \
                           radiuses)                   \
        GlPointSphere_initGeneric((pointSphere),       \
                                  (kind),              \
                                  (vertexCount),       \
                                  (vertices),          \
                                  sizeof(*(vertices)), \
                                  (colors),            \
                                  sizeof(*(colors)),   \
                                  (radiuses),          \
                                  sizeof(*(radiuses)))
TRANSGL_API
GlPointSphere *
GlPointSphere_newGeneric(GlPointSphereKind kind,
                         intptr_t vertexCount,
                         const void *vertices,
                         int sizeOfVertex,
                         const void *colors,
                         int sizeOfColor,
                         const void *radiuses,
                         int sizeOfRadius);

#define GlPointSphere_new(kind,                       \
                          vertexCount,                \
                          vertices,                   \
                          colors,                     \
                          radiuses)                   \
        GlPointSphere_newGeneric((kind),              \
                                 (vertexCount),       \
                                 (vertices),          \
                                 sizeof(*(vertices)), \
                                 (colors),            \
                                 sizeof(*(colors)),   \
                                 (radiuses),          \
                                 sizeof(*(radiuses)))

TRANSGL_API
void
GlPointSphere_destroy(GlPointSphere *pointSphere);

TRANSPROG_INLINE
void
GlPointSphere_delete(GlPointSphere *pointSphere)
{
GlPointSphere_destroy(pointSphere);
Memory_free(pointSphere);
}

TRANSGL_API
void
GlPointSphere_draw(const GlPointSphere *pointSphere,
                   const real32_t rgba[4],
                   real32_t ambient,
                   real32_t specular,
                   real32_t shininess,
                   real32_t radius,
                   const real32_t headLight[3],
                   const real32_t fog[4],
                   const real32_t *projectionMatrix,
                   const real32_t *modelViewMatrix);

TRANSPROG_INLINE
int // OpenGL buffer for vertices
GlPointSphere_vertexBuffer(const GlPointSphere *pointSphere)
{
return GlVertexArray_buffer(&pointSphere->vertexArray,pointSphere->vertices);
}

TRANSPROG_INLINE
int // OpenGL buffer for colors
GlPointSphere_colorBuffer(const GlPointSphere *pointSphere)
{
return GlVertexArray_buffer(&pointSphere->vertexArray,pointSphere->colors);
}

TRANSPROG_INLINE
int // OpenGL buffer for radiuses
GlPointSphere_radiusBuffer(const GlPointSphere *pointSphere)
{
return GlVertexArray_buffer(&pointSphere->vertexArray,pointSphere->radiuses);
}

#endif // TRANSGL_GLDRAW_H

//----------------------------------------------------------------------------
