//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_COORDINDEX_H
#define TRANSPROG_COORDINDEX_H 1

#include "TransProg/math.h"
#include "TransProg/memory.h"

//----------------------------------------------------------------------------

typedef struct _SpatialGrid
{
real64_t unitFactor;
intptr_t xCellCount;
intptr_t yCellCount;
intptr_t zCellCount;
} SpatialGrid;

TRANSPROG_INLINE
void
SpatialGrid_init(SpatialGrid *sg,
                 real64_t unitSize,
                 intptr_t xCellCount,
                 intptr_t yCellCount,
                 intptr_t zCellCount)
{
sg->unitFactor=1.0/unitSize;
sg->xCellCount=xCellCount;
sg->yCellCount=yCellCount;
sg->zCellCount=zCellCount;
}

TRANSPROG_INLINE
SpatialGrid *
SpatialGrid_new(real64_t unitSize,
                intptr_t xCellCount,
                intptr_t yCellCount,
                intptr_t zCellCount)
{
SpatialGrid *sg=Memory_alloc(1,SpatialGrid);
SpatialGrid_init(sg,unitSize,xCellCount,yCellCount,zCellCount);
return sg;
}

TRANSPROG_INLINE
void
SpatialGrid_destroy(SpatialGrid *sg)
{
(void)sg; // nothing to be done
}

TRANSPROG_INLINE
void
SpatialGrid_delete(SpatialGrid *sg)
{
SpatialGrid_destroy(sg);
Memory_free(sg);
}

TRANSPROG_INLINE
intptr_t
SpatialGrid_cellCount(const SpatialGrid *sg)
{
return sg->xCellCount*sg->yCellCount*sg->zCellCount;
}

TRANSPROG_INLINE
intptr_t
SpatialGrid_discreteDistance(const SpatialGrid *sg,
                             real64_t distance)
{
return (intptr_t)(distance*sg->unitFactor);
}

TRANSPROG_INLINE
intptr_t
SpatialGrid_coordToKey(const SpatialGrid *sg,
                       real64_t x,
                       real64_t y,
                       real64_t z)
{
const real64_t unitFactor=sg->unitFactor;
const real64_t xu=x*unitFactor;
const real64_t yu=y*unitFactor;
const real64_t zu=z*unitFactor;
const intptr_t xCellCount=sg->xCellCount;
const intptr_t yCellCount=sg->yCellCount;
const intptr_t zCellCount=sg->zCellCount;
const intptr_t xk=(intptr_t)(xu-xCellCount*floor(xu/xCellCount));
const intptr_t yk=(intptr_t)(yu-yCellCount*floor(yu/yCellCount));
const intptr_t zk=(intptr_t)(zu-zCellCount*floor(zu/zCellCount));
return (zk*yCellCount+yk)*xCellCount+xk;
}

TRANSPROG_INLINE
int // number of keys in neighbourhood
SpatialGrid_neighbourhood(const SpatialGrid *sg,
                          intptr_t *out_keys,
                          int capacity,
                          real64_t x,
                          real64_t y,
                          real64_t z,
                          real64_t range)
{
const real64_t unitFactor=sg->unitFactor;
const intptr_t zMin=(intptr_t)floor((z-range)*unitFactor);
const intptr_t zMax=(intptr_t)floor((z+range)*unitFactor);
const intptr_t yMin=(intptr_t)floor((y-range)*unitFactor);
const intptr_t yMax=(intptr_t)floor((y+range)*unitFactor);
const intptr_t xMin=(intptr_t)floor((x-range)*unitFactor);
const intptr_t xMax=(intptr_t)floor((x+range)*unitFactor);
const intptr_t zCellCount=sg->zCellCount;
const intptr_t yCellCount=sg->yCellCount;
const intptr_t xCellCount=sg->xCellCount;
int count=0;
for(intptr_t zi=zMin;zi<=zMax;++zi)
  {
  const intptr_t zk=yCellCount*((zi+zCellCount)%zCellCount);
  for(intptr_t yi=yMin;yi<=yMax;++yi)
    {
    const intptr_t yk=xCellCount*(zk+((yi+yCellCount)%yCellCount));
    for(intptr_t xi=xMin;xi<=xMax;++xi)
      {
      if(count<capacity)
        {
        out_keys[count]=yk+((xi+xCellCount)%xCellCount);
        }
      ++count;
      }
    }
  }
return count;
}

TRANSPROG_INLINE
int // number of keys in layer
SpatialGrid_layer(const SpatialGrid *sg,
                  intptr_t *out_keys,
                  int capacity,
                  real64_t x,
                  real64_t y,
                  real64_t z,
                  intptr_t layer)
{
const real64_t unitFactor=sg->unitFactor;
const intptr_t zMin=((intptr_t)floor(z*unitFactor))-layer;
const intptr_t zMax=zMin+2*layer;
const intptr_t yMin=((intptr_t)floor(y*unitFactor))-layer;
const intptr_t yMax=yMin+2*layer;
const intptr_t xMin=((intptr_t)floor(x*unitFactor))-layer;
const intptr_t xMax=xMin+2*layer;
const intptr_t zCellCount=sg->zCellCount;
const intptr_t yCellCount=sg->yCellCount;
const intptr_t xCellCount=sg->xCellCount;
int count=0;
for(intptr_t zi=zMin;zi<=zMax;++zi)
  {
  const intptr_t zk=yCellCount*((zi+zCellCount)%zCellCount);
  for(intptr_t yi=yMin;yi<=yMax;++yi)
    {
    const intptr_t yk=xCellCount*(zk+((yi+yCellCount)%yCellCount));
    const intptr_t xStep=((yi==yMax)||(yi==yMin)||(zi==zMax)||(zi==zMin))
                         ? 1 : xMax-xMin;
    for(intptr_t xi=xMin;xi<=xMax;xi+=xStep)
      {
      if(count<capacity)
        {
        out_keys[count]=yk+((xi+xCellCount)%xCellCount);
        }
      ++count;
      }
    }
  }
return count;
}

TRANSPROG_INLINE
intptr_t // first key of traversal
SpatialGrid_traversal(const SpatialGrid *sg,
                      intptr_t *out_keys,
                      int capacity,
                      real64_t xFrom,
                      real64_t yFrom,
                      real64_t zFrom,
                      real64_t xTo,
                      real64_t yTo,
                      real64_t zTo)
{
const real64_t unitFactor=sg->unitFactor;
intptr_t xi=(intptr_t)floor(xFrom*=unitFactor);
intptr_t yi=(intptr_t)floor(yFrom*=unitFactor);
intptr_t zi=(intptr_t)floor(zFrom*=unitFactor);
const real64_t dx=xTo*unitFactor-xFrom;
const real64_t dy=yTo*unitFactor-yFrom;
const real64_t dz=zTo*unitFactor-zFrom;
intptr_t xInc,yInc,zInc;
real64_t xStep,yStep,zStep;
real64_t xNext,yNext,zNext;
if(dx>0.0)      { xInc=+1; xStep=+1.0/dx; xNext=(xi+1-xFrom)*xStep; }
else if(dx<0.0) { xInc=-1; xStep=-1.0/dx; xNext=(xFrom-xi)*xStep;   }
else            { xInc=0;  xStep=DBL_MAX; xNext=DBL_MAX;            }
if(dy>0.0)      { yInc=+1; yStep=+1.0/dy; yNext=(yi+1-yFrom)*yStep; }
else if(dy<0.0) { yInc=-1; yStep=-1.0/dy; yNext=(yFrom-yi)*yStep;   }
else            { yInc=0;  yStep=DBL_MAX; yNext=DBL_MAX;            }
if(dz>0.0)      { zInc=+1; zStep=+1.0/dz; zNext=(zi+1-zFrom)*zStep; }
else if(dz<0.0) { zInc=-1; zStep=-1.0/dz; zNext=(zFrom-zi)*zStep;   }
else            { zInc=0;  zStep=DBL_MAX; zNext=DBL_MAX;            }
const intptr_t incCount=(intptr_t)(fabs(dx))+
                        (intptr_t)(fabs(dy))+
                        (intptr_t)(fabs(dz));
const intptr_t zCellCount=sg->zCellCount;
const intptr_t yCellCount=sg->yCellCount;
const intptr_t xCellCount=sg->xCellCount;
int count=0;
for(intptr_t i=0;i<=incCount;++i)
  {
  if(count<capacity)
    {
    const intptr_t zk=yCellCount*((zi+zCellCount)%zCellCount);
    const intptr_t yk=xCellCount*(zk+((yi+yCellCount)%yCellCount));
    const intptr_t xk=yk+((xi+xCellCount)%xCellCount);
    out_keys[count]=xk;
    }
  ++count;
  if(xNext<yNext)
    {
    if(xNext<zNext) { xNext=xNext+xStep; xi+=xInc; }
    else            { zNext=zNext+zStep; zi+=zInc; }
    }
  else
    {
    if(yNext<zNext) { yNext=yNext+yStep; yi+=yInc; }
    else            { zNext=zNext+zStep; zi+=zInc; }
    }
  }
return count;
}

//----------------------------------------------------------------------------

typedef struct _PlanarGrid
{
real64_t unitFactor;
intptr_t xCellCount;
intptr_t yCellCount;
} PlanarGrid;

TRANSPROG_INLINE
void
PlanarGrid_init(PlanarGrid *pg,
                real64_t unitSize,
                intptr_t xCellCount,
                intptr_t yCellCount)
{
pg->unitFactor=1.0/unitSize;
pg->xCellCount=xCellCount;
pg->yCellCount=yCellCount;
}

TRANSPROG_INLINE
PlanarGrid *
PlanarGrid_new(real64_t unitSize,
               intptr_t xCellCount,
               intptr_t yCellCount)
{
PlanarGrid *pg=Memory_alloc(1,PlanarGrid);
PlanarGrid_init(pg,unitSize,xCellCount,yCellCount);
return pg;
}

TRANSPROG_INLINE
void
PlanarGrid_destroy(PlanarGrid *pg)
{
(void)pg; // nothing to be done
}

TRANSPROG_INLINE
void
PlanarGrid_delete(PlanarGrid *pg)
{
PlanarGrid_destroy(pg);
Memory_free(pg);
}

TRANSPROG_INLINE
intptr_t
PlanarGrid_cellCount(const PlanarGrid *pg)
{
return pg->xCellCount*pg->yCellCount;
}

TRANSPROG_INLINE
intptr_t
PlanarGrid_discreteDistance(const PlanarGrid *pg,
                            real64_t distance)
{
return (intptr_t)(distance*pg->unitFactor);
}

TRANSPROG_INLINE
intptr_t
PlanarGrid_coordToKey(const PlanarGrid *pg,
                      real64_t x,
                      real64_t y)
{
const real64_t unitFactor=pg->unitFactor;
const real64_t xu=x*unitFactor;
const real64_t yu=y*unitFactor;
const intptr_t xCellCount=pg->xCellCount;
const intptr_t yCellCount=pg->yCellCount;
const intptr_t xk=(intptr_t)(xu-xCellCount*floor(xu/xCellCount));
const intptr_t yk=(intptr_t)(yu-yCellCount*floor(yu/yCellCount));
return yk*xCellCount+xk;
}

TRANSPROG_INLINE
int // number of keys in neighbourhood
PlanarGrid_neighbourhood(const PlanarGrid *pg,
                         intptr_t *out_keys,
                         int capacity,
                         real64_t x,
                         real64_t y,
                         real64_t range)
{
const real64_t unitFactor=pg->unitFactor;
const intptr_t yMin=(intptr_t)floor((y-range)*unitFactor);
const intptr_t yMax=(intptr_t)floor((y+range)*unitFactor);
const intptr_t xMin=(intptr_t)floor((x-range)*unitFactor);
const intptr_t xMax=(intptr_t)floor((x+range)*unitFactor);
const intptr_t yCellCount=pg->yCellCount;
const intptr_t xCellCount=pg->xCellCount;
int count=0;
for(intptr_t yi=yMin;yi<=yMax;++yi)
  {
  const intptr_t yk=xCellCount*((yi+yCellCount)%yCellCount);
  for(intptr_t xi=xMin;xi<=xMax;++xi)
    {
    if(count<capacity)
      {
      out_keys[count]=yk+((xi+xCellCount)%xCellCount);
      }
    ++count;
    }
  }
return count;
}

TRANSPROG_INLINE
int // number of keys in layer
PlanarGrid_layer(const PlanarGrid *pg,
                 intptr_t *out_keys,
                 int capacity,
                 real64_t x,
                 real64_t y,
                 intptr_t layer)
{
const real64_t unitFactor=pg->unitFactor;
const intptr_t yMin=((intptr_t)floor(y*unitFactor))-layer;
const intptr_t yMax=yMin+2*layer;
const intptr_t xMin=((intptr_t)floor(x*unitFactor))-layer;
const intptr_t xMax=xMin+2*layer;
const intptr_t yCellCount=pg->yCellCount;
const intptr_t xCellCount=pg->xCellCount;
int count=0;
for(intptr_t yi=yMin;yi<=yMax;++yi)
  {
  const intptr_t yk=xCellCount*((yi+yCellCount)%yCellCount);
  const intptr_t xStep=((yi==yMax)||(yi==yMin)) ? 1 : xMax-xMin;
  for(intptr_t xi=xMin;xi<=xMax;xi+=xStep)
    {
    if(count<capacity)
      {
      out_keys[count]=yk+((xi+xCellCount)%xCellCount);
      }
    ++count;
    }
  }
return count;
}

TRANSPROG_INLINE
intptr_t // first key of traversal
PlanarGrid_traversal(const PlanarGrid *pg,
                     intptr_t *out_keys,
                     int capacity,
                     real64_t xFrom,
                     real64_t yFrom,
                     real64_t xTo,
                     real64_t yTo)
{
const real64_t unitFactor=pg->unitFactor;
intptr_t xi=(intptr_t)floor(xFrom*=unitFactor);
intptr_t yi=(intptr_t)floor(yFrom*=unitFactor);
const real64_t dx=xTo*unitFactor-xFrom;
const real64_t dy=yTo*unitFactor-yFrom;
intptr_t xInc,yInc;
real64_t xStep,yStep;
real64_t xNext,yNext;
if(dx>0.0)      { xInc=+1; xStep=+1.0/dx; xNext=(xi+1-xFrom)*xStep; }
else if(dx<0.0) { xInc=-1; xStep=-1.0/dx; xNext=(xFrom-xi)*xStep;   }
else            { xInc=0;  xStep=DBL_MAX; xNext=DBL_MAX;            }
if(dy>0.0)      { yInc=+1; yStep=+1.0/dy; yNext=(yi+1-yFrom)*yStep; }
else if(dy<0.0) { yInc=-1; yStep=-1.0/dy; yNext=(yFrom-yi)*yStep;   }
else            { yInc=0;  yStep=DBL_MAX; yNext=DBL_MAX;            }
const intptr_t incCount=(intptr_t)(fabs(dx))+
                        (intptr_t)(fabs(dy));
const intptr_t yCellCount=pg->yCellCount;
const intptr_t xCellCount=pg->xCellCount;
int count=0;
for(intptr_t i=0;i<=incCount;++i)
  {
  if(count<capacity)
    {
    const intptr_t yk=xCellCount*((yi+yCellCount)%yCellCount);
    const intptr_t xk=yk+((xi+xCellCount)%xCellCount);
    out_keys[count]=xk;
    }
  ++count;
  if(xNext<yNext) { xNext=xNext+xStep; xi+=xInc; }
  else            { yNext=yNext+yStep; yi+=yInc; }
  }
return count;
}

//----------------------------------------------------------------------------

typedef struct _SpatialGridNeighbourhood
{
const SpatialGrid *sg;
intptr_t zMin,zMax,zi,zk;
intptr_t yMin,yMax,yi,yk;
intptr_t xMin,xMax,xi,xk;
} SpatialGridNeighbourhood;

TRANSPROG_INLINE
intptr_t // next key of neighbourhood or -1
SpatialGridNeighbourhood_next(SpatialGridNeighbourhood *sgn)
{
const SpatialGrid *sg=sgn->sg;
if(++sgn->xi<=sgn->xMax)
  {
  const intptr_t xCellCount=sg->xCellCount;
  return sgn->xk=sgn->yk+((sgn->xi+xCellCount)%xCellCount);
  }
if(++sgn->yi<=sgn->yMax)
  {
  const intptr_t yCellCount=sg->yCellCount;
  const intptr_t xCellCount=sg->xCellCount;
  const intptr_t yk=xCellCount*(sgn->zk+((sgn->yi+yCellCount)%yCellCount));
  sgn->yk=yk;
  const intptr_t xMin=sgn->xMin;
  sgn->xi=xMin;
  return sgn->xk=yk+((xMin+xCellCount)%xCellCount);
  }
if(++sgn->zi<=sgn->zMax)
  {
  const intptr_t zCellCount=sg->zCellCount;
  const intptr_t yCellCount=sg->yCellCount;
  const intptr_t xCellCount=sg->xCellCount;
  const intptr_t zk=yCellCount*((sgn->zi+zCellCount)%zCellCount);
  sgn->zk=zk;
  const intptr_t yMin=sgn->yMin;
  sgn->yi=yMin;
  const intptr_t yk=xCellCount*(zk+((yMin+yCellCount)%yCellCount));
  sgn->yk=yk;
  const intptr_t xMin=sgn->xMin;
  sgn->xi=xMin;
  return sgn->xk=yk+((xMin+xCellCount)%xCellCount);
  }
return -1;
}

TRANSPROG_INLINE
intptr_t // first key of neighbourhood
SpatialGridNeighbourhood_start(SpatialGridNeighbourhood *sgn,
                               const SpatialGrid *sg,
                               real64_t x,
                               real64_t y,
                               real64_t z,
                               real64_t range)
{
sgn->sg=sg;
const real64_t unitFactor=sg->unitFactor;
sgn->zMin=(intptr_t)floor((z-range)*unitFactor);
sgn->zMax=(intptr_t)floor((z+range)*unitFactor);
sgn->yMin=(intptr_t)floor((y-range)*unitFactor);
sgn->yMax=(intptr_t)floor((y+range)*unitFactor);
sgn->xMin=(intptr_t)floor((x-range)*unitFactor);
sgn->xMax=(intptr_t)floor((x+range)*unitFactor);
sgn->zi=sgn->zMin-1;
sgn->yi=sgn->yMax;
sgn->xi=sgn->xMax;
sgn->zk=sgn->yk=sgn->xk=0;
return SpatialGridNeighbourhood_next(sgn);
}

//----------------------------------------------------------------------------

typedef struct _PlanarGridNeighbourhood
{
const PlanarGrid *pg;
intptr_t yMin,yMax,yi,yk;
intptr_t xMin,xMax,xi,xk;
} PlanarGridNeighbourhood;

TRANSPROG_INLINE
intptr_t // next key of neighbourhood or -1
PlanarGridNeighbourhood_next(PlanarGridNeighbourhood *pgn)
{
const PlanarGrid *pg=pgn->pg;
if(++pgn->xi<=pgn->xMax)
  {
  const intptr_t xCellCount=pg->xCellCount;
  return pgn->xk=pgn->yk+((pgn->xi+xCellCount)%xCellCount);
  }
if(++pgn->yi<=pgn->yMax)
  {
  const intptr_t yCellCount=pg->yCellCount;
  const intptr_t xCellCount=pg->xCellCount;
  const intptr_t yk=xCellCount*((pgn->yi+yCellCount)%yCellCount);
  pgn->yk=yk;
  const intptr_t xMin=pgn->xMin;
  pgn->xi=xMin;
  return pgn->xk=yk+((xMin+xCellCount)%xCellCount);
  }
return -1;
}

TRANSPROG_INLINE
intptr_t // first key of neighbourhood
PlanarGridNeighbourhood_start(PlanarGridNeighbourhood *pgn,
                              const PlanarGrid *pg,
                              real64_t x,
                              real64_t y,
                              real64_t range)
{
pgn->pg=pg;
const real64_t unitFactor=pg->unitFactor;
pgn->yMin=(intptr_t)floor((y-range)*unitFactor);
pgn->yMax=(intptr_t)floor((y+range)*unitFactor);
pgn->xMin=(intptr_t)floor((x-range)*unitFactor);
pgn->xMax=(intptr_t)floor((x+range)*unitFactor);
pgn->yi=pgn->yMin-1;
pgn->xi=pgn->xMax;
pgn->yk=pgn->xk=0;
return PlanarGridNeighbourhood_next(pgn);
}

//----------------------------------------------------------------------------

typedef struct _SpatialGridLayer
{
const SpatialGrid *sg;
intptr_t zMin,zMax,zi,zk;
intptr_t yMin,yMax,yi,yk;
intptr_t xMin,xMax,xi,xk,xStep;
} SpatialGridLayer;

TRANSPROG_INLINE
intptr_t // next key of layer or -1
SpatialGridLayer_next(SpatialGridLayer *sgl)
{
const SpatialGrid *sg=sgl->sg;
const intptr_t xMax=sgl->xMax;
if((sgl->xi+=sgl->xStep)<=xMax)
  {
  const intptr_t xCellCount=sg->xCellCount;
  return sgl->xk=sgl->yk+((sgl->xi+xCellCount)%xCellCount);
  }
const intptr_t yMax=sgl->yMax;
if(++sgl->yi<=yMax)
  {
  const intptr_t yCellCount=sg->yCellCount;
  const intptr_t xCellCount=sg->xCellCount;
  const intptr_t yk=xCellCount*(sgl->zk+((sgl->yi+yCellCount)%yCellCount));
  sgl->yk=yk;
  const intptr_t xMin=sgl->xMin;
  sgl->xStep=((sgl->yi==yMax)||(sgl->yi==sgl->yMin)||
              (sgl->zi==sgl->zMax)||(sgl->zi==sgl->zMin))
             ? 1 : xMax-xMin;
  sgl->xi=xMin;
  return sgl->xk=yk+((xMin+xCellCount)%xCellCount);
  }
if(++sgl->zi<=sgl->zMax)
  {
  const intptr_t zCellCount=sg->zCellCount;
  const intptr_t yCellCount=sg->yCellCount;
  const intptr_t xCellCount=sg->xCellCount;
  const intptr_t zk=yCellCount*((sgl->zi+zCellCount)%sg->zCellCount);
  sgl->zk=zk;
  const intptr_t yMin=sgl->yMin;
  sgl->yi=yMin;
  const intptr_t yk=xCellCount*(zk+((yMin+yCellCount)%sg->yCellCount));
  sgl->yk=yk;
  sgl->xStep=1;
  const intptr_t xMin=sgl->xMin;
  sgl->xi=xMin;
  return sgl->xk=yk+((xMin+xCellCount)%xCellCount);
  }
return -1;
}

TRANSPROG_INLINE
intptr_t // first key of layer
SpatialGridLayer_start(SpatialGridLayer *sgl,
                       const SpatialGrid *sg,
                       real64_t x,
                       real64_t y,
                       real64_t z,
                       intptr_t layer)
{
sgl->sg=sg;
const real64_t unitFactor=sg->unitFactor;
sgl->zMin=((intptr_t)floor(z*unitFactor))-layer;
sgl->zMax=sgl->zMin+2*layer;
sgl->yMin=((intptr_t)floor(y*unitFactor))-layer;
sgl->yMax=sgl->yMin+2*layer;
sgl->xMin=((intptr_t)floor(x*unitFactor))-layer;
sgl->xMax=sgl->xMin+2*layer;
sgl->zi=sgl->zMin-1;
sgl->yi=sgl->yMax;
sgl->xi=sgl->xMax;
sgl->zk=sgl->yk=sgl->xk=0;
sgl->xStep=1;
return SpatialGridLayer_next(sgl);
}

//----------------------------------------------------------------------------

typedef struct _PlanarGridLayer
{
const PlanarGrid *pg;
intptr_t yMin,yMax,yi,yk;
intptr_t xMin,xMax,xi,xk,xStep;
} PlanarGridLayer;

TRANSPROG_INLINE
intptr_t // next key of layer or -1
PlanarGridLayer_next(PlanarGridLayer *pgl)
{
const PlanarGrid *pg=pgl->pg;
if((pgl->xi+=pgl->xStep)<=pgl->xMax)
  {
  const intptr_t xCellCount=pg->xCellCount;
  return pgl->xk=pgl->yk+((pgl->xi+xCellCount)%pg->xCellCount);
  }
const intptr_t yMax=pgl->yMax;
if(++pgl->yi<=yMax)
  {
  const intptr_t yCellCount=pg->yCellCount;
  const intptr_t xCellCount=pg->xCellCount;
  const intptr_t yk=xCellCount*((pgl->yi+yCellCount)%pg->yCellCount);
  pgl->yk=yk;
  const intptr_t xMin=pgl->xMin;
  pgl->xStep=((pgl->yi==yMax)||(pgl->yi==pgl->yMin))
             ? 1 : pgl->xMax-xMin;
  pgl->xi=xMin;
  return pgl->xk=yk+((xMin+xCellCount)%xCellCount);
  }
return -1;
}

TRANSPROG_INLINE
intptr_t // first key of layer
PlanarGridLayer_start(PlanarGridLayer *pgl,
                      const PlanarGrid *pg,
                      real64_t x,
                      real64_t y,
                      intptr_t layer)
{
pgl->pg=pg;
const real64_t unitFactor=pg->unitFactor;
pgl->yMin=((intptr_t)floor(y*unitFactor))-layer;
pgl->yMax=pgl->yMin+2*layer;
pgl->xMin=((intptr_t)floor(x*unitFactor))-layer;
pgl->xMax=pgl->xMin+2*layer;
pgl->yi=pgl->yMin-1;
pgl->xi=pgl->xMax;
pgl->yk=pgl->xk=0;
pgl->xStep=1;
return PlanarGridLayer_next(pgl);
}

//----------------------------------------------------------------------------

typedef struct _SpatialGridTraversal
{
const SpatialGrid *sg;
intptr_t xi,yi,zi;
intptr_t xInc,yInc,zInc;
real64_t xStep,yStep,zStep;
real64_t xNext,yNext,zNext;
intptr_t incCount;
} SpatialGridTraversal;

TRANSPROG_INLINE
intptr_t // first key of traversal
SpatialGridTraversal_start(SpatialGridTraversal *sgt,
                           const SpatialGrid *sg,
                           real64_t xFrom,
                           real64_t yFrom,
                           real64_t zFrom,
                           real64_t xTo,
                           real64_t yTo,
                           real64_t zTo)
{
sgt->sg=sg;
const real64_t unitFactor=sg->unitFactor;
const intptr_t xi=(intptr_t)floor(xFrom*=unitFactor);
const intptr_t yi=(intptr_t)floor(yFrom*=unitFactor);
const intptr_t zi=(intptr_t)floor(zFrom*=unitFactor);
sgt->xi=xi;
sgt->yi=yi;
sgt->zi=zi;
const real64_t dx=xTo*unitFactor-xFrom;
const real64_t dy=yTo*unitFactor-yFrom;
const real64_t dz=zTo*unitFactor-zFrom;
if(dx>0.0)
  { sgt->xInc=1;  sgt->xStep=1.0/dx; sgt->xNext=(xi+1-xFrom)*sgt->xStep; }
else if(dx<0.0)
  { sgt->xInc=-1; sgt->xStep=-1.0/dx; sgt->xNext=(xFrom-xi)*sgt->xStep; }
else
  { sgt->xInc=0;  sgt->xStep=DBL_MAX; sgt->xNext=DBL_MAX; }
if(dy>0.0)
  { sgt->yInc=1;  sgt->yStep=1.0/dy; sgt->yNext=(yi+1-yFrom)*sgt->yStep; }
else if(dy<0.0)
  { sgt->yInc=-1; sgt->yStep=-1.0/dy; sgt->yNext=(yFrom-yi)*sgt->yStep; }
else
  { sgt->yInc=0;  sgt->yStep=DBL_MAX; sgt->yNext=DBL_MAX; }
if(dz>0.0)
  { sgt->zInc=1;  sgt->zStep=1.0/dz; sgt->zNext=(zi+1-zFrom)*sgt->zStep; }
else if(dz<0.0)
  { sgt->zInc=-1; sgt->zStep=-1.0/dz; sgt->zNext=(zFrom-zi)*sgt->zStep; }
else
  { sgt->zInc=0;  sgt->zStep=DBL_MAX; sgt->zNext=DBL_MAX; }
sgt->incCount=(intptr_t)(fabs(dx))+
              (intptr_t)(fabs(dy))+
              (intptr_t)(fabs(dz));
const intptr_t zCellCount=sg->zCellCount;
const intptr_t yCellCount=sg->yCellCount;
const intptr_t xCellCount=sg->xCellCount;
const intptr_t zk=yCellCount*((zi+zCellCount)%zCellCount);
const intptr_t yk=xCellCount*(zk+((yi+yCellCount)%yCellCount));
const intptr_t xk=yk+((xi+xCellCount)%xCellCount);
return xk;
}

TRANSPROG_INLINE
intptr_t // next key of traversal or -1
SpatialGridTraversal_next(SpatialGridTraversal *sgt)
{
if(--sgt->incCount>=0)
  {
  const real64_t xNext=sgt->xNext;
  const real64_t yNext=sgt->yNext;
  const real64_t zNext=sgt->zNext;
  if(xNext<yNext)
    {
    if(xNext<zNext) { sgt->xNext=xNext+sgt->xStep; sgt->xi+=sgt->xInc; }
    else            { sgt->zNext=zNext+sgt->zStep; sgt->zi+=sgt->zInc; }
    }
  else
    {
    if(yNext<zNext) { sgt->yNext=yNext+sgt->yStep; sgt->yi+=sgt->yInc; }
    else            { sgt->zNext=zNext+sgt->zStep; sgt->zi+=sgt->zInc; }
    }
  const SpatialGrid *sg=sgt->sg;
  const intptr_t zCellCount=sg->zCellCount;
  const intptr_t yCellCount=sg->yCellCount;
  const intptr_t xCellCount=sg->xCellCount;
  const intptr_t zk=yCellCount*((sgt->zi+zCellCount)%zCellCount);
  const intptr_t yk=xCellCount*(zk+((sgt->yi+yCellCount)%yCellCount));
  const intptr_t xk=yk+((sgt->xi+xCellCount)%xCellCount);
  return xk;
  }
return -1;
}

//----------------------------------------------------------------------------

typedef struct _PlanarGridTraversal
{
const PlanarGrid *pg;
intptr_t xi,yi;
intptr_t xInc,yInc;
real64_t xStep,yStep;
real64_t xNext,yNext;
intptr_t incCount;
} PlanarGridTraversal;

TRANSPROG_INLINE
intptr_t // first key of traversal
PlanarGridTraversal_start(PlanarGridTraversal *pgt,
                          const PlanarGrid *pg,
                          real64_t xFrom,
                          real64_t yFrom,
                          real64_t xTo,
                          real64_t yTo)
{
pgt->pg=pg;
const real64_t unitFactor=pg->unitFactor;
const intptr_t xi=(intptr_t)floor(xFrom*=unitFactor);
const intptr_t yi=(intptr_t)floor(yFrom*=unitFactor);
pgt->xi=xi;
pgt->yi=yi;
const real64_t dx=xTo*unitFactor-xFrom;
const real64_t dy=yTo*unitFactor-yFrom;
if(dx>0.0)
  { pgt->xInc=1;  pgt->xStep=1.0/dx; pgt->xNext=(xi+1-xFrom)*pgt->xStep; }
else if(dx<0.0)
  { pgt->xInc=-1; pgt->xStep=-1.0/dx; pgt->xNext=(xFrom-xi)*pgt->xStep; }
else
  { pgt->xInc=0;  pgt->xStep=DBL_MAX; pgt->xNext=DBL_MAX; }
if(dy>0.0)
  { pgt->yInc=1;  pgt->yStep=1.0/dy; pgt->yNext=(yi+1-yFrom)*pgt->yStep; }
else if(dy<0.0)
  { pgt->yInc=-1; pgt->yStep=-1.0/dy; pgt->yNext=(yFrom-yi)*pgt->yStep; }
else
  { pgt->yInc=0;  pgt->yStep=DBL_MAX; pgt->yNext=DBL_MAX; }
pgt->incCount=(intptr_t)(fabs(dx))+
              (intptr_t)(fabs(dy));
const intptr_t yCellCount=pg->yCellCount;
const intptr_t xCellCount=pg->xCellCount;
const intptr_t yk=xCellCount*((yi+yCellCount)%yCellCount);
const intptr_t xk=yk+((xi+xCellCount)%xCellCount);
return xk;
}

TRANSPROG_INLINE
intptr_t // next key of traversal or -1
PlanarGridTraversal_next(PlanarGridTraversal *pgt)
{
if(--pgt->incCount>=0)
  {
  const real64_t xNext=pgt->xNext;
  const real64_t yNext=pgt->yNext;
  if(xNext<yNext)
    { pgt->xNext=xNext+pgt->xStep; pgt->xi+=pgt->xInc; }
  else
    { pgt->yNext=yNext+pgt->yStep; pgt->yi+=pgt->yInc; }
  const PlanarGrid *pg=pgt->pg;
  const intptr_t yCellCount=pg->yCellCount;
  const intptr_t xCellCount=pg->xCellCount;
  const intptr_t yk=xCellCount*((pgt->yi+yCellCount)%yCellCount);
  const intptr_t xk=yk+((pgt->xi+xCellCount)%xCellCount);
  return xk;
  }
return -1;
}

//----------------------------------------------------------------------------

typedef struct _SpatialAffinity
{
real64_t xMin,xSizeInv;
real64_t yMin,ySizeInv;
real64_t zMin,zSizeInv;
int xSplit;
int ySplit;
int zSplit;
int xySplit;
} SpatialAffinity;

TRANSPROG_API
void
SpatialAffinity_init(SpatialAffinity *sa,
                     real64_t xMin,
                     real64_t xMax,
                     real64_t yMin,
                     real64_t yMax,
                     real64_t zMin,
                     real64_t zMax);

TRANSPROG_INLINE
SpatialAffinity *
SpatialAffinity_new(real64_t xMin,
                    real64_t xMax,
                    real64_t yMin,
                    real64_t yMax,
                    real64_t zMin,
                    real64_t zMax)
{
SpatialAffinity *sa=Memory_alloc(1,SpatialAffinity);
SpatialAffinity_init(sa,xMin,xMax,yMin,yMax,zMin,zMax);
return sa;
}

TRANSPROG_INLINE
void
SpatialAffinity_destroy(SpatialAffinity *sa)
{
(void)sa; // nothing to be done
}

TRANSPROG_INLINE
void
SpatialAffinity_delete(SpatialAffinity *sa)
{
SpatialAffinity_destroy(sa);
Memory_free(sa);
}

TRANSPROG_INLINE
int
SpatialAffinity_xSplit(const SpatialAffinity *sa)
{
return sa->xSplit;
}

TRANSPROG_INLINE
int
SpatialAffinity_ySplit(const SpatialAffinity *sa)
{
return sa->ySplit;
}

TRANSPROG_INLINE
int
SpatialAffinity_zSplit(const SpatialAffinity *sa)
{
return sa->zSplit;
}

TRANSPROG_INLINE
int
SpatialAffinity_chooseCpu(const SpatialAffinity *sa,
                          real64_t x,
                          real64_t y,
                          real64_t z)
{
#if !defined NDEBUG
  if((x<sa->xMin)||(x>=sa->xMin+1.0/sa->xSizeInv)||
     (y<sa->yMin)||(y>=sa->yMin+1.0/sa->ySizeInv)||
     (z<sa->zMin)||(z>=sa->zMin+1.0/sa->zSizeInv))
    {
    DROP_DEAD("point is out of bounds");
    }
#endif
int xi=(int)(sa->xSplit*(x-sa->xMin)*sa->xSizeInv);
int yi=(int)(sa->ySplit*(y-sa->yMin)*sa->ySizeInv);
int zi=(int)(sa->zSplit*(z-sa->zMin)*sa->zSizeInv);
return xi+sa->xSplit*yi+sa->xySplit*zi;
}

//----------------------------------------------------------------------------

typedef struct _PlanarAffinity
{
real64_t xMin,xSizeInv;
real64_t yMin,ySizeInv;
int xSplit;
int ySplit;
} PlanarAffinity;

TRANSPROG_API
void
PlanarAffinity_init(PlanarAffinity *pa,
                    real64_t xMin,
                    real64_t xMax,
                    real64_t yMin,
                    real64_t yMax);

TRANSPROG_INLINE
PlanarAffinity *
PlanarAffinity_new(real64_t xMin,
                   real64_t xMax,
                   real64_t yMin,
                   real64_t yMax)
{
PlanarAffinity *pa=Memory_alloc(1,PlanarAffinity);
PlanarAffinity_init(pa,xMin,xMax,yMin,yMax);
return pa;
}

TRANSPROG_INLINE
void
PlanarAffinity_destroy(PlanarAffinity *pa)
{
(void)pa; // nothing to be done
}

TRANSPROG_INLINE
void
PlanarAffinity_delete(PlanarAffinity *pa)
{
PlanarAffinity_destroy(pa);
Memory_free(pa);
}

TRANSPROG_INLINE
int
PlanarAffinity_xSplit(const PlanarAffinity *pa)
{
return pa->xSplit;
}

TRANSPROG_INLINE
int
PlanarAffinity_ySplit(const PlanarAffinity *pa)
{
return pa->ySplit;
}

TRANSPROG_INLINE
int
PlanarAffinity_chooseCpu(const PlanarAffinity *pa,
                         real64_t x,
                         real64_t y)
{
#if !defined NDEBUG
  if((x<pa->xMin)||(x>=pa->xMin+1.0/pa->xSizeInv)||
     (y<pa->yMin)||(y>=pa->yMin+1.0/pa->ySizeInv))
    {
    DROP_DEAD("point is out of bounds");
    }
#endif
int xi=(int)(pa->xSplit*(x-pa->xMin)*pa->xSizeInv);
int yi=(int)(pa->ySplit*(y-pa->yMin)*pa->ySizeInv);
return xi+pa->xSplit*yi;
}

#endif // TRANSPROG_COORDINDEX_H

//----------------------------------------------------------------------------
