//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SYNCHRO_H
#define TRANSPROG_SYNCHRO_H 1

#include "TransProg/memory.h"

typedef volatile struct _AtomicPointer
{
void *ptr;
} AtomicPointer;

typedef volatile struct _SpinSynchro
{
int32_t flag;
} SpinSynchro;

typedef struct _SystemSynchro
{
#if defined _WIN32
  CRITICAL_SECTION lock;
  CONDITION_VARIABLE cond;
  HANDLE *waiting;
  int waitingCount;
#else
  pthread_mutex_t lock;
  pthread_cond_t cond;
#endif
} SystemSynchro;

//----------------------------------------------------------------------------

TRANSPROG_INLINE
void
AtomicPointer_init(AtomicPointer *ptr,
                   void *value)
{
ptr->ptr=value;
}

TRANSPROG_INLINE
void *
AtomicPointer_get(AtomicPointer ptr)
{
intptr_t tagMask=(intptr_t)(Platform_maxCacheLineSize()-1);
return (void *)(((intptr_t)ptr.ptr)&~tagMask);
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
AtomicPointer_set(AtomicPointer *address,
                  AtomicPointer expectedOldValue,
                  void *newValue)
{
intptr_t tagMask=(intptr_t)(Platform_maxCacheLineSize()-1);
#if !defined NDEBUG
  if(((intptr_t)newValue)|tagMask)
    {
    DROP_DEAD("newValue is not cache-line aligned");
    }
#endif
intptr_t tag=(((intptr_t)address->ptr)+1)&tagMask;
newValue=(void *)(((intptr_t)newValue)|tag);
#if defined _WIN32
  return InterlockedCompareExchangePointer((PVOID volatile *)&address->ptr,
                                           newValue,
                                           (PVOID)expectedOldValue.ptr)
         ==expectedOldValue.ptr;
#else
  return __sync_bool_compare_and_swap(&address->ptr,
                                      expectedOldValue.ptr,
                                      newValue);
#endif
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeI32(volatile int32_t *address,
                             int32_t expectedOldValue,
                             int32_t newValue)
{
#if defined _WIN32
  return _InterlockedCompareExchange((volatile long *)address,
                                     newValue,
                                     expectedOldValue)==expectedOldValue;
#else
  return __sync_bool_compare_and_swap(address,expectedOldValue,newValue);
#endif
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeUi32(volatile uint32_t *address,
                              uint32_t expectedOldValue,
                              uint32_t newValue)
{
return Atomic_compareAndExchangeI32((volatile int32_t *)address,
                                    (int32_t)expectedOldValue,
                                    (int32_t)newValue);
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeR32(volatile real32_t *address,
                             real32_t expectedOldValue,
                             real32_t newValue)
{
typedef union { int32_t i32; real32_t r32; } u32;
return Atomic_compareAndExchangeI32((volatile int32_t *)address,
                                    ((u32 *)&expectedOldValue)->i32,
                                    ((u32 *)&newValue)->i32);
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeI64(volatile int64_t *address,
                             int64_t expectedOldValue,
                             int64_t newValue)
{
#if defined _WIN32
  return _InterlockedCompareExchange64((volatile long long *)address,
                                       newValue,
                                       expectedOldValue)==expectedOldValue;
#else
  return __sync_bool_compare_and_swap(address,expectedOldValue,newValue);
#endif
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeUi64(volatile uint64_t *address,
                              uint64_t expectedOldValue,
                              uint64_t newValue)
{
return Atomic_compareAndExchangeI64((volatile int64_t *)address,
                                    (int64_t)expectedOldValue,
                                    (int64_t)newValue);
}

TRANSPROG_INLINE
bool // assignment done, value at address was conforming to expectation
Atomic_compareAndExchangeR64(volatile real64_t *address,
                             real64_t expectedOldValue,
                             real64_t newValue)
{
typedef union { int64_t i64; real64_t r64; } u64;
return Atomic_compareAndExchangeI64((volatile int64_t *)address,
                                    ((u64 *)&expectedOldValue)->i64,
                                    ((u64 *)&newValue)->i64);
}

TRANSPROG_INLINE
int32_t // value at address before addition
Atomic_postAddI32(volatile int32_t *address,
                  int32_t addend)
{
#if defined _WIN32
  return _InterlockedExchangeAdd((volatile long *)address,addend);
#else
  return __sync_fetch_and_add(address,addend);
#endif
}

TRANSPROG_INLINE
uint32_t // value at address before addition
Atomic_postAddUi32(volatile uint32_t *address,
                   uint32_t addend)
{
return (uint32_t)Atomic_postAddI32((volatile int32_t *)address,
                                   (int32_t)addend);
}

TRANSPROG_INLINE
real32_t // value at address before addition
Atomic_postAddR32(volatile real32_t *address,
                  real32_t addend)
{
real32_t oldValue;
do
  {
  oldValue=*address;
  } while(!Atomic_compareAndExchangeR32(address,oldValue,oldValue+addend));
return oldValue;
}

TRANSPROG_INLINE
int64_t // value at address before addition
Atomic_postAddI64(volatile int64_t *address,
                  int64_t addend)
{
#if defined _WIN32
  return _InterlockedExchangeAdd64((volatile long long *)address,addend);
#else
  return __sync_fetch_and_add(address,addend);
#endif
}

TRANSPROG_INLINE
uint64_t // value at address before addition
Atomic_postAddUi64(volatile uint64_t *address,
                   uint64_t addend)
{
return (uint64_t)Atomic_postAddI64((volatile int64_t *)address,
                                   (int64_t)addend);
}

TRANSPROG_INLINE
real64_t // value at address before addition
Atomic_postAddR64(volatile real64_t *address,
                  real64_t addend)
{
real64_t oldValue;
do
  {
  oldValue=*address;
  } while(!Atomic_compareAndExchangeR64(address,oldValue,oldValue+addend));
return oldValue;
}

TRANSPROG_INLINE
int32_t // value at address after incrementation
Atomic_preIncrementI32(volatile int32_t *address)
{
#if defined _WIN32
  return _InterlockedIncrement((volatile long *)address);
#else
  return __sync_add_and_fetch(address,1);
#endif
}

TRANSPROG_INLINE
uint32_t // value at address after incrementation
Atomic_preIncrementUi32(volatile uint32_t *address)
{
return (uint32_t)Atomic_preIncrementI32((volatile int32_t *)address);
}

TRANSPROG_INLINE
int64_t // value at address after incrementation
Atomic_preIncrementI64(volatile int64_t *address)
{
#if defined _WIN32
  return _InterlockedIncrement64((volatile long long *)address);
#else
  return __sync_add_and_fetch(address,1);
#endif
}

TRANSPROG_INLINE
uint64_t // value at address after incrementation
Atomic_preIncrementUi64(volatile uint64_t *address)
{
return (uint64_t)Atomic_preIncrementI64((volatile int64_t *)address);
}

TRANSPROG_INLINE
int32_t // value at address after decrementation
Atomic_preDecrementI32(volatile int32_t *address)
{
#if defined _WIN32
  return _InterlockedDecrement((volatile long *)address);
#else
  return __sync_add_and_fetch(address,-1);
#endif
}

TRANSPROG_INLINE
uint32_t // value at address after decrementation
Atomic_preDecrementUi32(volatile uint32_t *address)
{
return (uint32_t)Atomic_preDecrementI32((volatile int32_t *)address);
}

TRANSPROG_INLINE
int64_t // value at address after decrementation
Atomic_preDecrementI64(volatile int64_t *address)
{
#if defined _WIN32
  return _InterlockedDecrement64((volatile long long *)address);
#else
  return __sync_add_and_fetch(address,-1);
#endif
}

TRANSPROG_INLINE
uint64_t // value at address after decrementation
Atomic_preDecrementUi64(volatile uint64_t *address)
{
return (uint64_t)Atomic_preDecrementI64((volatile int64_t *)address);
}

#if INT_MAX==INT32_MAX
# define Atomic_compareAndExchangeInt Atomic_compareAndExchangeI32
# define Atomic_postAddInt            Atomic_postAddI32
# define Atomic_preIncrementInt       Atomic_preIncrementI32
# define Atomic_preDecrementInt       Atomic_preDecrementI32
#else
# error "weird size for int!"
#endif

#if INTPTR_MAX==INT32_MAX
# define Atomic_compareAndExchangeIntPtr Atomic_compareAndExchangeI32
# define Atomic_postAddIntPtr            Atomic_postAddI32
# define Atomic_preIncrementIntPtr       Atomic_preIncrementI32
# define Atomic_preDecrementIntPtr       Atomic_preDecrementI32
#elif INTPTR_MAX==INT64_MAX
# define Atomic_compareAndExchangeIntPtr Atomic_compareAndExchangeI64
# define Atomic_postAddIntPtr            Atomic_postAddI64
# define Atomic_preIncrementIntPtr       Atomic_preIncrementI64
# define Atomic_preDecrementIntPtr       Atomic_preDecrementI64
#else
# error "weird size for intptr_t!"
#endif

//----------------------------------------------------------------------------

TRANSPROG_INLINE
void
SpinSynchro_init(SpinSynchro *s)
{
s->flag=0x01000000;
}

TRANSPROG_INLINE
SpinSynchro *
SpinSynchro_new(void)
{
SpinSynchro *s=Memory_alloc(1,SpinSynchro);
SpinSynchro_init(s);
return s;
}

TRANSPROG_INLINE
void
SpinSynchro_destroy(SpinSynchro *s)
{
(void)s;
}

TRANSPROG_INLINE
void
SpinSynchro_delete(SpinSynchro *s)
{
SpinSynchro_destroy(s);
Memory_free((void *)s); // discard volatile qualifier
}

TRANSPROG_INLINE
bool // success
SpinSynchro_tryLockW(SpinSynchro *s)
{
return Atomic_compareAndExchangeI32(&s->flag,0x01000000,0);
}

TRANSPROG_INLINE
void
SpinSynchro_lockW(SpinSynchro *s)
{
while(!SpinSynchro_tryLockW(s))
  {
  while(s->flag!=0x01000000)
    {
    Platform_pause(); // nb: pause does not seem to help much
    }
  }
}

TRANSPROG_INLINE
void
SpinSynchro_unlockW(SpinSynchro *s)
{
Atomic_postAddI32(&s->flag,0x01000000);
}

TRANSPROG_INLINE
bool // success
SpinSynchro_tryLockR(SpinSynchro *s)
{
if(Atomic_preDecrementI32(&s->flag)<0)
  {
  Atomic_preIncrementI32(&s->flag);
  return false;
  }
else
  {
  return true;
  }
}

TRANSPROG_INLINE
void
SpinSynchro_lockR(SpinSynchro *s)
{
while(!SpinSynchro_tryLockR(s))
  {
  while(s->flag<=0)
    {
    Platform_pause(); // nb: pause does not seem to help much
    }
  }
}

TRANSPROG_INLINE
void
SpinSynchro_unlockR(SpinSynchro *s)
{
Atomic_preIncrementI32(&s->flag);
}

TRANSPROG_INLINE
bool // success
SpinSynchro_tryUpgrade(SpinSynchro *s)
{
return Atomic_compareAndExchangeI32(&s->flag,0x00FFFFFF,0);
}

TRANSPROG_INLINE
void
SpinSynchro_upgrade(SpinSynchro *s)
{
while(!SpinSynchro_tryUpgrade(s))
  {
  while(s->flag!=0x00FFFFFF)
    {
    Platform_pause(); // nb: pause does not seem to help much
    }
  }
}

TRANSPROG_INLINE
void
SpinSynchro_downgrade(SpinSynchro *s)
{
Atomic_postAddI32(&s->flag,0x00FFFFFF);
}

//----------------------------------------------------------------------------

TRANSPROG_API
void
SystemSynchro_init(SystemSynchro *s);

TRANSPROG_API
SystemSynchro *
SystemSynchro_new(void);

TRANSPROG_API
void
SystemSynchro_destroy(SystemSynchro *s);

TRANSPROG_API
void
SystemSynchro_delete(SystemSynchro *s);

TRANSPROG_API
bool // success
SystemSynchro_tryEnter(SystemSynchro *s);

TRANSPROG_API
void
SystemSynchro_enter(SystemSynchro *s);

TRANSPROG_API
void
SystemSynchro_leave(SystemSynchro *s);

TRANSPROG_API
void
SystemSynchro_wait(SystemSynchro *s);

TRANSPROG_API
void
SystemSynchro_broadcast(SystemSynchro *s);

#endif // TRANSPROG_SYNCHRO_H

//----------------------------------------------------------------------------
