#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import sys
import traceback
import math

class SkelTransform(object):
  def __init__(self,kind):
    object.__init__(self)
    self.kind=kind
    self.value=0.0
    self.animation=[]
    self.channel=-1

class SkelNode(object):
  def __init__(self,skel,parent):
    object.__init__(self)
    self.index=-1
    self.name=None
    self.parent=parent
    self.children=[]
    self.offset=[0.0,0.0,0.0]
    self.transformations=[]
    skel.nodes.append(self)

  def addTransformation(self,tr):
    for t in self.transformations:
      if t.kind==tr.kind:
        sys.stderr.write('multiple %s transformations for node %s\n'
                          %(tr.kind,self.name))
        sys.exit(1)
    self.transformations.append(tr)

class Skeleton(object):
  def __init__(self,input):
    object.__init__(self)
    self.input=input;
    self.line=1
    self.c=input.read(1)
    self.nodes=[]
    self.channelCount=0
    self.dt=0.0
    self.frameCount=0
    self.frames=[]
    self._expectWord('HIERARCHY')
    self._expectWord('ROOT')
    root=self._readNode(None,False)
    self._expectWord('MOTION')
    self._expectWord('FRAMES:')
    self.frameCount=self._expectInt('number of frames',True)
    self._expectWord('FRAME')
    self._expectWord('TIME:')
    self.dt=self._expectFloat('frame time value',True)
    for i in range(self.frameCount):
      channel=0
      for n in self.nodes:
        for t in n.transformations:
          t.animation.append(self._expectFloat('channel value %u of frame %u'
                                               %(channel,i)))
          channel+=1
    root.index=0
    nodes=[root]
    self._adjustIndexes(root,nodes)
    self.nodes=nodes
    toRadian=math.pi/180.0
    for n in self.nodes:
      tr=[]
      trKinds=[]
      for t in n.transformations:
        k=['TX','TY','TZ','RX','RY','RZ'].index(t.kind)
        if k<3:
          t.value=n.offset[k]
        else:
          for i in range(self.frameCount):
            t.animation[i]*=toRadian
        avg=sum(t.animation)/self.frameCount
        tolerance=math.fabs(0.001*avg)
        constant=True
        for i in range(self.frameCount):
          if math.fabs(t.animation[i]-avg)>tolerance:
            constant=False
            break
        if not constant:
          t.channel=self.channelCount
          self.channelCount+=1
          tr.append(t)
          trKinds.append(t.kind)
      n.transformations=tr
      for (k,i) in zip(['TZ','TY','TX'],[2,1,0]):
        if n.offset[i] and k not in trKinds:
          t=SkelTransform(k)
          t.value=n.offset[i]
          n.transformations.insert(0,t)

  def _expectError(self,expected):
    sys.stderr.write('%s expected on line %d\n'%(expected,self.line))
    sys.exit(1)

  def _expectWord(self,name):
    token=self._nextToken()
    if token.lower()==name.lower():
      return token
    self._expectError(name)

  def _expectInt(self,name,positive=False):
    token=self._nextToken()
    try:
      value=int(token)
      if not positive or value>=0:
        return value
    except:
      pass
    self._expectError(name)

  def _expectFloat(self,name,positive=False):
    token=self._nextToken()
    try:
      value=float(token)
      if not positive or value>=0.0:
        return value
    except:
      pass
    self._expectError(name)

  def _nextToken(self):
    token=''
    while self.c!='' and self.c.isspace():
      if self.c=='\n':
        self.line+=1
      self.c=input.read(1)
    if self.c!='':
      token+=self.c
      self.c=self.input.read(1)
      while self.c!='' and not self.c.isspace():
        token+=self.c
        self.c=self.input.read(1)
    return token

  def _readNode(self,parent,terminal):
    node=SkelNode(self,parent)
    if not terminal:
      node.name=self._nextToken()
      if not node.name:
        self._expectError('node name')
    self._expectWord('{')
    while True:
      token=self._nextToken()
      if token=='}':
        break
      if token.upper()=='OFFSET':
        node.offset[0]=self._expectFloat('x offset')
        node.offset[1]=self._expectFloat('y offset')
        node.offset[2]=self._expectFloat('z offset')
        continue
      if terminal:
        self._expectError('OFFSET or }')
      else:
        if token.upper()=='CHANNELS':
          count=self._expectInt('number of channels',True)
          d={'XPOSITION':'TX', 'YPOSITION':'TY', 'ZPOSITION':'TZ', 
             'XROTATION':'RX', 'YROTATION':'RY', 'ZROTATION':'RZ'}
          for i in range(count):
            kind=d.get(self._nextToken().upper())
            if not kind:
              self._expectError('{X|Y|Z}{position|rotation}')
            node.addTransformation(SkelTransform(kind))
          continue
        if token.upper()=='JOINT':
          node.children.append(self._readNode(node,False))
          continue
        if token.upper()=='END':
          self._expectWord('SITE')
          node.children.append(self._readNode(node,True))
          continue
        self._expectError('OFFSET, CHANNELS, JOINT, End Site or }')
    if not terminal and not node.children:
      sys.stderr.write('%s node should contain children nodes\n'%node.name)
      sys.exit(1)
    return node

  def _adjustIndexes(self,node,nodes):
    for n in node.children:
      n.index=len(nodes)
      nodes.append(n)
    for n in node.children:
      self._adjustIndexes(n,nodes)

  def show(self,output,node=None,indent=0):
    if not node:
      output.write('dt %g\n'%self.dt)
      output.write('frames %d\n'%self.frameCount)
      output.write('channels %d\n'%self.channelCount)
      output.write('nodes %d\n'%len(self.nodes))
      node=self.nodes[0]
    spc='    '*indent
    output.write('%s(%d) %s\n'%(spc,node.index,node.name))
    for t in node.transformations:
      output.write('%s    %s %g'%(spc,t.kind,t.value))
      if t.animation:
        output.write(' + %s'%t.animation)
      output.write('\n')
    for child in node.children:
      self.show(output,child,indent+1)

  def save(self,output):
    output.write('dt %g\n'%self.dt)
    output.write('frames %d\n'%self.frameCount)
    output.write('channels %d\n'%self.channelCount)
    output.write('nodes %d\n'%len(self.nodes))
    for n in self.nodes:
      output.write('name %s\n'%n.name)
      if n.parent:
        output.write('parent %d\n'%n.parent.index)
      else:
        output.write('parent -1\n')
      output.write('children %d'%len(n.children))
      if n.children:
        output.write(' %d\n'%n.children[0].index)
      else:
        output.write(' -1\n')
      output.write('transformations %d\n'%len(n.transformations))
      for t in n.transformations:
        output.write('kind %s value %g channel %d'
                     %(t.kind,t.value,t.channel))
        if t.animation:
          for v in t.animation:
            output.write(' %g'%v)
        output.write('\n')

if __name__=='__main__':
  input=sys.stdin
  output=sys.stdout
  showUsage=False
  try:
    args=sys.argv[1:]
    args.reverse()
    while args:
      arg=args.pop()
      if arg=='-i':
        input=open(args.pop(),'r')
      elif arg=='-o':
        output=open(args.pop(),'w')
      else:
        showUsage=True
  except:
    traceback.print_exc()
    showUsage=True
  if showUsage:
    sys.stderr.write('usage: %s -i input_file -o output_file\n'
                     %sys.argv[0])
    sys.exit(1)
  skel=Skeleton(input)
  skel.save(output)
  # skel.show(sys.stderr)

#-----------------------------------------------------------------------------
