#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import os
import sys
import traceback
import math
import xml.dom.pulldom

if hasattr(dict,'iteritems'):
  dict_items23=dict.iteritems # python2
else:
  dict_items23=dict.items # python3

# xml.dom.minidom loads the xml tree at once and thus can lead to memory
# exhaustion in case of large input file.
# Conversely, xml.dom.pulldom enables to consider each node one after
# another and does not require the entire xml tree to stand in memory
# at once.

class SkelAnimation(object):
  def __init__(self,input,output):
    object.__init__(self)
    self.input=input
    self.output=output
    self.channel=-1

class SkelTransform(object):
  def __init__(self,kind,name,value):
    object.__init__(self)
    self.kind=kind
    self.name=name
    self.value=value
    self.animation=None

class SkelNode(object):
  def __init__(self,name,parent):
    object.__init__(self)
    self.name=name
    self.parent=parent
    if parent:
      parent.children.append(self)
    self.children=[]
    self.transformations=[]
    self.index=-1

class Skeleton(object):
  def __init__(self,input,rootId,dt):
    object.__init__(self)
    self.animations={}
    self.rootNode=SkelNode('',None)
    self.nodes=[]
    self.dt=dt
    self.frames=[]
    self.channelCount=0
    self.unit=1.0
    self.upAxis='Z_UP'
    self.xmlDoc=xml.dom.pulldom.parse(input)
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='COLLADA':
          self._pullCollada()
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break
    self.rootNode=self._findRootId(rootId,self.rootNode)
    if self.rootNode:
      self.rootNode.parent=None
      self.rootNode.index=len(self.nodes)
      self.nodes.append(self.rootNode)
      self._adjustNodes(self.rootNode)
      self._adjustAnimations()

  def _findChild(self,xmlNode,localName,**kw):
    for xmlChild in xmlNode.childNodes:
      if xmlChild.localName==localName:
        found=True
        for (a,v) in dict_items23(kw):
          if xmlChild.getAttribute(a)!=v:
            found=False
            break
        if found:
          return xmlChild
    return None

  def _xmlData(self,xmlNode):
    return ''.join([child.data for child in xmlNode.childNodes])

  def _findRootId(self,rootId,node):
    if node.name==rootId:
      return node
    for child in node.children:
      found=self._findRootId(rootId,child)
      if found:
        return found
    return None

  def _pullIgnored(self):
    depth=0
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        depth+=1
      elif evt==xml.dom.pulldom.END_ELEMENT:
        if depth:
          depth-=1
        else:
          break

  def _pullCollada(self):
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='asset':
          self._pullAsset()
        elif xmlNode.localName=='library_animations':
          self._pullAnimationLib()
        elif xmlNode.localName=='library_visual_scenes':
          self._pullSceneLib()
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break

  def _pullAsset(self):
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='unit':
          self.xmlDoc.expandNode(xmlNode)
          self.unit=float(xmlNode.getAttribute('meter'))
        elif xmlNode.localName=='up_axis':
          self.xmlDoc.expandNode(xmlNode)
          self.upAxis=self._xmlData(xmlNode)
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break

  def _pullAnimationLib(self):
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='animation':
          self.xmlDoc.expandNode(xmlNode)
          channel=self._findChild(xmlNode,'channel')
          if channel:
            target=channel.getAttribute('target')
            sampler=self._findChild(xmlNode,'sampler',
                                    id=channel.getAttribute('source')[1:])
            if sampler:
              interp=self._findChild(sampler,'input',semantic='INTERPOLATION')
              if interp:
                interp=self._findChild(xmlNode,'source',
                                       id=interp.getAttribute('source')[1:])
                if interp:
                  interp=self._findChild(interp,'Name_array')
                  if interp:
                    interp=self._xmlData(interp).split()
              input=self._findChild(sampler,'input',semantic='INPUT')
              if input:
                input=self._findChild(xmlNode,'source',
                                      id=input.getAttribute('source')[1:])
                if input:
                  input=self._findChild(input,'float_array')
                  if input:
                    input=[float(x) for x in self._xmlData(input).split()]
              output=self._findChild(sampler,'input',semantic='OUTPUT')
              if output:
                output=self._findChild(xmlNode,'source',
                                       id=output.getAttribute('source')[1:])
                if output:
                  output=self._findChild(output,'float_array')
                  if output:
                    output=[float(x) for x in self._xmlData(output).split()]
              if interp:
                for v in interp:
                  if v!='LINEAR':
                    sys.stderr.write('non LINEAR interpolation for %s\n'
                                     %target)
                    interp=None
                    break
              if input and output:
                if len(input)==len(output):
                  avg=sum(output)/len(output)
                  tolerance=0.001*math.fabs(avg)
                  constant=True
                  for v in output:
                    if math.fabs(v-avg)>tolerance:
                      constant=False
                      break
                  if constant:
                    # sys.stderr.write('ignoring constant animation for %s\n'
                    #                  %target)
                    interp=None
                else:
                  sys.stderr.write('inconsistent INPUT/OUTPUT length for %s\n'
                                   %target)
                  interp=None
              if interp and input and output:
                if self.animations.get(target):
                  sys.stderr.write('ignoring multiple %s animations\n'%target)
                else:
                  self.animations[target]=SkelAnimation(input,output)
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break

  def _pullSceneLib(self):
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='visual_scene':
          for (evt,xmlNode) in self.xmlDoc:
            if evt==xml.dom.pulldom.START_ELEMENT:
              if xmlNode.localName=='node':
                self._pullNode(SkelNode(xmlNode.getAttribute('id'),
                                        self.rootNode))
              else:
                self._pullIgnored()
            elif evt==xml.dom.pulldom.END_ELEMENT:
              break
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break

  def _pullNode(self,node):
    for (evt,xmlNode) in self.xmlDoc:
      if evt==xml.dom.pulldom.START_ELEMENT:
        if xmlNode.localName=='node':
          self._pullNode(SkelNode(xmlNode.getAttribute('id'),node))
        elif xmlNode.localName=='translate':
          self.xmlDoc.expandNode(xmlNode)
          name=xmlNode.getAttribute('sid')
          trans=[float(x) for x in self._xmlData(xmlNode).split()]
          self._addTransformation(node,SkelTransform('TX',name,trans[0]))
          self._addTransformation(node,SkelTransform('TY',name,trans[1]))
          self._addTransformation(node,SkelTransform('TZ',name,trans[2]))
        elif xmlNode.localName=='rotate':
          self.xmlDoc.expandNode(xmlNode)
          name=xmlNode.getAttribute('sid')
          rot=[float(x) for x in self._xmlData(xmlNode).split()]
          axis=rot[0:3]
          angle=rot[3]
          if axis==[1.0,0.0,0.0]:
            self._addTransformation(node,SkelTransform('RX',name,angle))
          elif axis==[0.0,1.0,0.0]:
            self._addTransformation(node,SkelTransform('RY',name,angle))
          elif axis==[0.0,0.0,1.0]:
            self._addTransformation(node,SkelTransform('RZ',name,angle))
          else:
            sys.stderr.write('unknown rotation axis %s for node %s\n'
                             %(axis,node.name))
        elif xmlNode.localName=='scale':
          self.xmlDoc.expandNode(xmlNode)
          sc=[float(x) for x in self._xmlData(xmlNode).split()]
          if sc!=[1.0,1.0,1.0]:
            sys.stderr.write('ignoring scale %s for node %s\n'
                             %(sc,node.name))
        else:
          self._pullIgnored()
      elif evt==xml.dom.pulldom.END_ELEMENT:
        break

  def _addTransformation(self,node,tr):
    for t in node.transformations:
      if t.kind==tr.kind:
        sys.stderr.write('ignoring multiple %s transformations for node %s\n'
                          %(tr.kind,node.name))
        return
    node.transformations.append(tr)
    target='%s/%s.'%(node.name,tr.name)
    if tr.kind[0]=='R':
      target+='ANGLE'
    else:
      target+=tr.kind[1]
    tr.animation=self.animations.get(target)

  def _adjustNodes(self,node):
    for transform in node.transformations:
      if transform.kind[0]=='R':
        factor=math.pi/180.0
      else:
        factor=self.unit
      transform.value*=factor
      anim=transform.animation
      if anim:
        for t in anim.input:
          if t not in self.frames:
            self.frames.append(t)
        anim.output=[x*factor-transform.value for x in anim.output]
        anim.channel=self.channelCount
        self.channelCount+=1
    for child in node.children:
      child.index=len(self.nodes)
      self.nodes.append(child)
    for child in node.children:
      self._adjustNodes(child)

  def _adjustAnimations(self):
    self.frames.sort()
    if self.dt<=0.0:
      for i in range(len(self.frames)):
        self.dt=self.frames[i]
        if self.dt>0.0:
          break
    if self.dt<=0.0:
      self.frames=[]
      for node in self.nodes:
        for transform in node.transformations:
          transform.animation=None
    else:
      minFrame=self.frames[0]
      maxFrame=self.frames[-1]
      frameCount=1+(maxFrame-minFrame)/self.dt
      ceil=math.ceil(frameCount)
      floor=math.floor(frameCount)
      if frameCount-floor<ceil-frameCount:
        frameCount=int(floor)
      else:
        frameCount=int(ceil)
      self.frames=[i*self.dt for i in range(frameCount)]
      for node in self.nodes:
        for transform in node.transformations:
          anim=transform.animation
          if anim:
            output=[]
            for f in self.frames:
              f+=minFrame
              found=False
              for i in range(len(anim.input)-1):
                prev=anim.input[i]
                next=anim.input[i+1]
                if f>=prev and f<=next:
                  ratio=(next-f)/(next-prev)
                  output.append(anim.output[i]*ratio+
                                anim.output[i+1]*(1.0-ratio))
                  found=True
                  break
              if not found:
                if f<anim.input[0]:
                  output.append(anim.output[0])
                else:
                  output.append(anim.output[-1])
            anim.output=output
            anim.input=self.frames

  def show(self,output,node=None,indent=0):
    if not node:
      output.write('dt %g\n'%self.dt)
      output.write('frames %d\n'%len(self.frames))
      output.write('channels %d\n'%self.channelCount)
      output.write('nodes %d\n'%len(self.nodes))
      node=self.nodes[0]
    spc='    '*indent
    output.write('%s(%d) %s\n'%(spc,node.index,node.name))
    for t in node.transformations:
      output.write('%s    %s %g'%(spc,t.kind,t.value))
      if t.animation:
        output.write(' + %s'%t.animation.output)
      output.write('\n')
    for child in node.children:
      self.show(output,child,indent+1)

  def save(self,output):
    output.write('dt %g\n'%self.dt)
    output.write('frames %d\n'%len(self.frames))
    output.write('channels %d\n'%self.channelCount)
    output.write('nodes %d\n'%len(self.nodes))
    for n in self.nodes:
      output.write('name %s\n'%n.name)
      if n.parent:
        output.write('parent %d\n'%n.parent.index)
      else:
        output.write('parent -1\n')
      output.write('children %d'%len(n.children))
      if n.children:
        output.write(' %d\n'%n.children[0].index)
      else:
        output.write(' -1\n')
      output.write('transformations %d\n'%len(n.transformations))
      for t in n.transformations:
        output.write('kind %s value %g'%(t.kind,t.value))
        if t.animation:
          output.write(' channel %d'%t.animation.channel)
          for v in t.animation.output:
            output.write(' %g'%v)
          output.write('\n')
        else:
          output.write(' channel -1\n')

if __name__=='__main__':
  input=sys.stdin
  output=sys.stdout
  rootId=None
  dt=0.0
  showUsage=False
  try:
    args=sys.argv[1:]
    args.reverse()
    while args:
      arg=args.pop()
      if arg=='-r':
        rootId=args.pop()
      elif arg=='-dt':
        dt=float(args.pop())
      elif arg=='-i':
        input=args.pop()
        if rootId is None:
          rootId=os.path.splitext(os.path.basename(input))[0]
        input=open(input,'r')
      elif arg=='-o':
        output=open(args.pop(),'w')
      else:
        showUsage=True
  except:
    traceback.print_exc()
    showUsage=True
  if showUsage:
    sys.stderr.write(('usage: %s -r root_node -dt delta_t'+
                      ' -i input_file -o output_file\n')%sys.argv[0])
    sys.exit(1)
  skel=Skeleton(input,rootId,dt)
  if not skel.nodes:
    sys.stderr.write('no \'%s\' skeleton\n'%rootId)
    sys.exit(1)
  skel.save(output)
  # skel.show(sys.stderr)

#-----------------------------------------------------------------------------
