//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/dynamicCompiler.h"

#include "TransProg/path.h"

#include "dynamicCompilerDefaults.c"

#if defined _WIN32
# define OUT_EXT "dll"
#elif defined __APPLE__
# define OUT_EXT "bundle"
#else
# define OUT_EXT "so"
#endif

static
void
_DynamicCompiler_initOptions(StrVector *options,
                             const char * const *defaultOptions)
{
StrVector_init(options);
for(;*defaultOptions;++defaultOptions)
  {
  StrVector_push(options,Memory_allocStr(*defaultOptions));
  }
}

static
void
_DynamicCompiler_destroyOptions(StrVector *options)
{
while(StrVector_size(options))
  {
  Memory_free(StrVector_pop(options));
  }
StrVector_destroy(options);
}

void
DynamicCompiler_init(DynamicCompiler *dc)
{
_DynamicCompiler_initOptions(&dc->cc,
                             _DynamicCompiler_cc);
_DynamicCompiler_initOptions(&dc->cflags_only_I,
                             _DynamicCompiler_cflags_only_I);
_DynamicCompiler_initOptions(&dc->cflags_only_D,
                             _DynamicCompiler_cflags_only_D);
_DynamicCompiler_initOptions(&dc->cflags_only_other,
                             _DynamicCompiler_cflags_only_other);
_DynamicCompiler_initOptions(&dc->libs_only_L,
                             _DynamicCompiler_libs_only_L);
_DynamicCompiler_initOptions(&dc->libs_only_l,
                             _DynamicCompiler_libs_only_l);
_DynamicCompiler_initOptions(&dc->libs_only_other,
                             _DynamicCompiler_libs_only_other);
_DynamicCompiler_initOptions(&dc->oflags,
                             _DynamicCompiler_oflags);
Text_init(&dc->buildLog);
}

void
DynamicCompiler_destroy(DynamicCompiler *dc)
{
Text_destroy(&dc->buildLog);
_DynamicCompiler_destroyOptions(&dc->libs_only_other);
_DynamicCompiler_destroyOptions(&dc->oflags);
_DynamicCompiler_destroyOptions(&dc->libs_only_l);
_DynamicCompiler_destroyOptions(&dc->libs_only_L);
_DynamicCompiler_destroyOptions(&dc->cflags_only_other);
_DynamicCompiler_destroyOptions(&dc->cflags_only_D);
_DynamicCompiler_destroyOptions(&dc->cflags_only_I);
_DynamicCompiler_destroyOptions(&dc->cc);
}

static
void
_DynamicCompiler_appendCommand(Text *command,
                               const StrVector *options)
{
for(intptr_t i=0,i_end=StrVector_size(options);i<i_end;++i)
  {
  if(Text_length(command))
    {
    Text_appendStr(command," ");
    }
  Text_appendStr(command,StrVector_get(options,i));
  }
}

PluginHandle // built plugin or 0
DynamicCompiler_build(DynamicCompiler *dc,
                      const char *sourceCode)
{
PluginHandle plugin=(PluginHandle)0;
Text_clear(&dc->buildLog);
char inFile[0x100]="";
char outFile[0x100]="";
char logFile[0x100]="";
intptr_t sz=strlen(sourceCode);
if(!Path_createTmpFile(inFile,0x100,".","dynCode_??????.c"))
  {
  char msg[0x100];
  Platform_lastErrorMsg(msg,0x100);
  Text_append(&dc->buildLog,
              "creating source file: %s\n",msg);
  }
else if(!Path_createTmpFile(outFile,0x100,".","dynCode_??????." OUT_EXT))
  {
  char msg[0x100];
  Platform_lastErrorMsg(msg,0x100);
  Text_append(&dc->buildLog,
              "creating plugin file: %s\n",msg);
  }
else if(!Path_createTmpFile(logFile,0x100,".","dynCode_??????.log"))
  {
  char msg[0x100];
  Platform_lastErrorMsg(msg,0x100);
  Text_append(&dc->buildLog,
              "creating log file: %s\n",msg);
  }
else if(Path_writeFile(inFile,sourceCode,sz,false)!=sz)
  {
  char msg[0x100];
  Platform_lastErrorMsg(msg,0x100);
  Text_append(&dc->buildLog,
              "writing to source file %s: %s\n",inFile,msg);
  }
else
  {
  Text command;
  Text_init(&command);
  _DynamicCompiler_appendCommand(&command,&dc->cc);
  _DynamicCompiler_appendCommand(&command,&dc->cflags_only_other);
  _DynamicCompiler_appendCommand(&command,&dc->cflags_only_I);
  _DynamicCompiler_appendCommand(&command,&dc->cflags_only_D);
  Text_append(&command," %s",inFile);
  _DynamicCompiler_appendCommand(&command,&dc->libs_only_other);
  _DynamicCompiler_appendCommand(&command,&dc->libs_only_L);
  _DynamicCompiler_appendCommand(&command,&dc->libs_only_l);
  _DynamicCompiler_appendCommand(&command,&dc->oflags);
  Text_append(&command,"%s >%s 2>&1",outFile,logFile);
  int status=system(Text_get(&command));
#if defined _WIN32
  if(status==-1)
    {
    int errorCode=errno;
    Text_append(&dc->buildLog,
                "system(\"%s\") failure --- %s (%d)\n",
                Text_get(&command),strerror(errorCode),errorCode);
    }
  else if(status)
    {
    Text_append(&dc->buildLog,
                "system(\"%s\") returned %d\n",
                Text_get(&command),status);
    }
#else
  if(WIFSIGNALED(status))
    {
    Text_append(&dc->buildLog,
                "system(\"%s\") terminated by signal %d\n",
                Text_get(&command),WTERMSIG(status));
    }
  else if(WEXITSTATUS(status))
    {
    Text_append(&dc->buildLog,
                "system(\"%s\") returned %d\n",
                Text_get(&command),WEXITSTATUS(status));
    }
#endif
  Text_destroy(&command);
  sz=(intptr_t)Path_fileSize(logFile);
  if(sz>0)
    {
    char *buildLog=Memory_alloc(sz+1,char);
    if(Path_readFile(logFile,buildLog,sz,0)==sz)
      {
      buildLog[sz]='\0';
      Text_appendStr(&dc->buildLog,buildLog);
      }
    Memory_free(buildLog);
    }
  plugin=Plugin_open(outFile);
  if(!plugin)
    {
    char msg[0x100];
    Plugin_lastErrorMsg(msg,0x100);
    Text_append(&dc->buildLog,
                "Plugin_open() failure: %s\n",msg);
    }
  }
if(inFile[0])
  {
  Path_remove(inFile);
#if defined _WIN32
  int prefix=(int)strlen(inFile)-1;
  strcpy(inFile+prefix,"obj");
  Path_remove(inFile);
#endif
  }
if(outFile[0])
  {
  // FIXME: under Window$, a file cannot be removed while in use!
  Path_remove(outFile);
#if defined _WIN32
  int prefix=(int)strlen(outFile)-(int)strlen(OUT_EXT);
  strcat(outFile,".manifest");
  Path_remove(outFile);
  strcpy(outFile+prefix,"exp");
  Path_remove(outFile);
  strcpy(outFile+prefix,"lib");
  Path_remove(outFile);
#endif
  }
if(logFile[0])
  {
  Path_remove(logFile);
  }
return plugin;
}

//----------------------------------------------------------------------------
