//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/natnetClient.h"

// #define SHOW_MESSAGES

#define NATNET_COMMAND_PORT    1510
#define NATNET_DATA_PORT       1511
#define NATNET_MAX_PACKET_SIZE (1024*1024)

#define NATNET_PING                  0
#define NATNET_PING_RESPONSE         1
#define NATNET_REQUEST               2
#define NATNET_RESPONSE              3
#define NATNET_REQUEST_MODEL_DEF     4
#define NATNET_MODEL_DEF             5
#define NATNET_REQUEST_FRAME_OF_DATA 6
#define NATNET_FRAME_OF_DATA         7
#define NATNET_MESSAGE_STRING        8
#define NATNET_UNRECOGNIZED_REQUEST  100

#define NATNET_MARKER_SET 0
#define NATNET_RIGID_BODY 1
#define NATNET_SKELETON   2

struct _NatnetServerInfo
{
char name[256];
uint8_t version[4];
uint8_t natnet[4];
};

TRANSPROG_INLINE
struct _NatnetRigidBody *
NatnetRigidBody_new(const char *name)
{
int nameSize=(int)strlen(name)+1;
uintptr_t size=OFFSET_OF_MEMBER(struct _NatnetRigidBody,name)+nameSize;
struct _NatnetRigidBody *body=(struct _NatnetRigidBody *)
                              Memory_allocVoid(size);
body->id=-1;
body->parentId=-1;
body->xOffset=0.0; body->yOffset=0.0; body->zOffset=0.0;
body->tx=0.0; body->ty=0.0; body->tz=0.0;
body->qx=0.0; body->qy=0.0; body->qz=0.0; body->qw=1.0;
memcpy(body->name,name,nameSize);
return body;
}

TRANSPROG_INLINE
void
NatnetRigidBody_delete(struct _NatnetRigidBody *body)
{
Memory_free(body);
}

TRANSPROG_INLINE
struct _NatnetSkeleton *
NatnetSkeleton_new(const char *name)
{
int nameSize=(int)strlen(name)+1;
uintptr_t size=OFFSET_OF_MEMBER(struct _NatnetSkeleton,name)+nameSize;
struct _NatnetSkeleton *skeleton=(struct _NatnetSkeleton *)
                                 Memory_allocVoid(size);
skeleton->id=-1;
PtrVector_init(&skeleton->rigidBodies);
memcpy(skeleton->name,name,nameSize);
return skeleton;
}

TRANSPROG_INLINE
void
NatnetSkeleton_delete(struct _NatnetSkeleton *skeleton)
{
for(intptr_t i=0,i_end=PtrVector_size(&skeleton->rigidBodies);i<i_end;++i)
  {
  struct _NatnetRigidBody *body=(struct _NatnetRigidBody *)
                                PtrVector_get(&skeleton->rigidBodies,i);
  NatnetRigidBody_delete(body);
  }
PtrVector_destroy(&skeleton->rigidBodies);
Memory_free(skeleton);
}

TRANSPROG_INLINE
int16_t
Natnet_readI16(const uint8_t **ptr)
{
int16_t b0=(*ptr)[0];
int16_t b1=(*ptr)[1];
*ptr+=2;
return (int16_t)((b0<<0)|(b1<<8));
}

TRANSPROG_INLINE
void
Natnet_writeI16(uint8_t **ptr,
                int16_t value)
{
(*ptr)[0]=(uint8_t)((value>>0)&0x00FF);
(*ptr)[1]=(uint8_t)((value>>8)&0x00FF);
*ptr+=2;
}

TRANSPROG_INLINE
int32_t
Natnet_readI32(const uint8_t **ptr)
{
int32_t b0=(*ptr)[0];
int32_t b1=(*ptr)[1];
int32_t b2=(*ptr)[2];
int32_t b3=(*ptr)[3];
*ptr+=4;
return (b0<<0)|(b1<<8)|(b2<<16)|(b3<<24);
}

#if 0 // unused
  TRANSPROG_INLINE
  void
  Natnet_writeI32(uint8_t **ptr,
                  int32_t value)
  {
  (*ptr)[0]=(uint8_t)((value>>0)&0x000000FF);
  (*ptr)[1]=(uint8_t)((value>>8)&0x000000FF);
  (*ptr)[2]=(uint8_t)((value>>16)&0x000000FF);
  (*ptr)[3]=(uint8_t)((value>>24)&0x000000FF);
  *ptr+=4;
  }
#endif

TRANSPROG_INLINE
real32_t
Natnet_readR32(const uint8_t **ptr)
{
union { int32_t i; real32_t r; } u;
u.i=Natnet_readI32(ptr);
return u.r;
}

#if 0 // unused
  TRANSPROG_INLINE
  void
  Natnet_writeR32(uint8_t **ptr,
                  real32_t value)
  {
  union { int32_t i; real32_t r; } u;
  u.r=value;
  Natnet_writeI32(ptr,u.i);
  }
#endif

TRANSPROG_INLINE
const char *
Natnet_readStr0(const uint8_t **ptr)
{
const char *str=(const char *)*ptr;
*ptr+=strlen(str)+1;
return str;
}

#if 0 // unused
 TRANSPROG_INLINE
 void
 Natnet_writeStr0(uint8_t **ptr,
                  const char *str)
 {
 char c;
 do
   {
   c=*(str++);
   (*ptr)[0]=c;
   ++(*ptr);
   } while(c);
 }
#endif

TRANSPROG_INLINE
bool
Natnet_minVersion(const NatnetClient *client,
                  int major,
                  int minor)
{
return client->version>=((major<<8)|minor);
}

TRANSPROG_INLINE
bool
Natnet_maxVersion(const NatnetClient *client,
                  int major,
                  int minor)
{
return client->version<=((major<<8)|minor);
}

TRANSPROG_INLINE
void
Natnet_send(NatnetClient *client,
            int msgType)
{
uint8_t *ptr=client->buffer;
Natnet_writeI16(&ptr,(int16_t)msgType);
Natnet_writeI16(&ptr,0);
if(Socket_sendToIpv4(client->sock,client->buffer,4,0,
                     client->address,NATNET_COMMAND_PORT)!=4)
  {
  SYS_DROP_DEAD("Socket_sendToIpv4() failure");
  }
}

static
void
Natnet_analysePingResponse(NatnetClient *client,
                           int msgSize,
                           const uint8_t *ptr)
{
#if defined SHOW_MESSAGES
  Platform_msg("NATNET_PING_RESPONSE\n");
#endif
if(msgSize<(int)sizeof(struct _NatnetServerInfo))
  {
  return;
  }
const struct _NatnetServerInfo *info=(const struct _NatnetServerInfo *)ptr;
#if defined SHOW_MESSAGES
  Platform_msg("  name: %s\n",info->name);
  Platform_msg("  version: %d.%d.%d.%d\n",
               (int)info->version[0],
               (int)info->version[1],
               (int)info->version[2],
               (int)info->version[3]);
  Platform_msg("  natnet: %d.%d.%d.%d\n",
               (int)info->natnet[0],
               (int)info->natnet[1],
               (int)info->natnet[2],
               (int)info->natnet[3]);
#endif
int major=info->natnet[0];
int minor=info->natnet[1];
int version=(major<<8)|minor;
if(!client->version)
  {
  client->version=version;
  if(!Natnet_maxVersion(client,2,6))
    {
    char msg[0x100];
    sprintf(msg,"NATNET version %d.%d not supported\n",major,minor);
    DROP_DEAD(msg);
    }
  }
else if(version!=client->version)
  {
  DROP_DEAD("NATNET version changed");
  }
}

static
void
Natnet_analyseModelDef(NatnetClient *client,
                       int msgSize,
                       const uint8_t *ptr)
{
(void)msgSize; // FIXME: should check that ptr does not go too far
if(!client->version||
   PtrVector_size(&client->rigidBodies)||
   PtrVector_size(&client->skeletons))
  {
  return;
  }
#if defined SHOW_MESSAGES
  Platform_msg("NATNET_MODEL_DEF\n");
#endif
int dataSetCount=Natnet_readI32(&ptr);
for(int ds=0;ds<dataSetCount;++ds)
  {
  int dsType=Natnet_readI32(&ptr);
  switch(dsType)
    {
    case NATNET_MARKER_SET:
      {
      const char *name=Natnet_readStr0(&ptr);
#if defined SHOW_MESSAGES
      Platform_msg("  NATNET_MARKER_SET [%s]\n",name);
#else
      (void)name;
#endif
      int markerCount=Natnet_readI32(&ptr);
      for(int m=0;m<markerCount;++m)
        {
        const char *name=Natnet_readStr0(&ptr);
#if defined SHOW_MESSAGES
        Platform_msg("    %s\n",name);
#else
        (void)name;
#endif
        }
      break;
      }
    case NATNET_RIGID_BODY:
      {
      const char *name=Natnet_minVersion(client,2,0)
                       ? Natnet_readStr0(&ptr) : "";
      struct _NatnetRigidBody *body=NatnetRigidBody_new(name);
      PtrVector_push(&client->rigidBodies,body);
      body->id=Natnet_readI32(&ptr);
      body->parentId=Natnet_readI32(&ptr);
      body->xOffset=Natnet_readR32(&ptr);
      body->yOffset=Natnet_readR32(&ptr);
      body->zOffset=Natnet_readR32(&ptr);
#if defined SHOW_MESSAGES
      Platform_msg("  NATNET_RIGID_BODY [%s] %d %d (%g,%g,%g)\n",
                   body->name,body->id,body->parentId,
                   body->xOffset,body->yOffset,body->zOffset);
#endif
      break;
      }
    case NATNET_SKELETON:
      {
      const char *name=Natnet_readStr0(&ptr);
      struct _NatnetSkeleton *skeleton=NatnetSkeleton_new(name);
      PtrVector_push(&client->skeletons,skeleton);
      skeleton->id=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
      Platform_msg("  NATNET_SKELETON %d [%s]\n",skeleton->id,skeleton->name);
#endif
      int rigidBodyCount=Natnet_readI32(&ptr);
      for(int rb=0;rb<rigidBodyCount;++rb)
        {
        const char *name=Natnet_minVersion(client,2,0)
                         ? Natnet_readStr0(&ptr) : "";
        struct _NatnetRigidBody *body=NatnetRigidBody_new(name);
        PtrVector_push(&skeleton->rigidBodies,body);
        body->id=Natnet_readI32(&ptr);
        body->parentId=Natnet_readI32(&ptr);
        body->xOffset=Natnet_readR32(&ptr);
        body->yOffset=Natnet_readR32(&ptr);
        body->zOffset=Natnet_readR32(&ptr);
#if defined SHOW_MESSAGES
        Platform_msg("    rigid-body [%s] %d %d (%g,%g,%g)\n",
                     body->name,body->id,body->parentId,
                     body->xOffset,body->yOffset,body->zOffset);
#endif
        }
      break;
      }
    default:
      {
#if defined SHOW_MESSAGES
      Platform_msg("  unknown dataset type %d\n",dsType);
#endif
      }
    }
  }
}

static
void
Natnet_analyseFrameOfData(NatnetClient *client,
                          int msgSize,
                          const uint8_t *ptr)
{
(void)msgSize; // FIXME: should check that ptr does not go too far
if(!PtrVector_size(&client->rigidBodies)&&
   !PtrVector_size(&client->skeletons))
  {
  return;
  }
#if defined SHOW_MESSAGES
  int frameNumber=Natnet_readI32(&ptr);
  Platform_msg("NATNET_FRAME_OF_DATA %d\n",frameNumber);
#else
  ptr+=sizeof(int32_t);
#endif
int markerSetCount=Natnet_readI32(&ptr);
for(int ms=0;ms<markerSetCount;++ms)
  {
  const char *name=Natnet_readStr0(&ptr);
#if defined SHOW_MESSAGES
  Platform_msg("  marker set [%s]\n",name);
#else
  (void)name;
#endif
  int markerCount=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
  for(int m=0;m<markerCount;++m)
    {
    real32_t x=Natnet_readR32(&ptr);
    real32_t y=Natnet_readR32(&ptr);
    real32_t z=Natnet_readR32(&ptr);
    Platform_msg("    (%g,%g,%g)\n",x,y,z);
    }
#else
  ptr+=markerCount*3*sizeof(real32_t);
#endif
  }
int otherMarkerCount=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
  for(int om=0;om<otherMarkerCount;++om)
    {
    real32_t x=Natnet_readR32(&ptr);
    real32_t y=Natnet_readR32(&ptr);
    real32_t z=Natnet_readR32(&ptr);
    Platform_msg("  other marker (%g,%g,%g)\n",x,y,z);
    }
#else
  ptr+=otherMarkerCount*3*sizeof(real32_t);
#endif
int rigidBodyCount=Natnet_readI32(&ptr);
if(rigidBodyCount!=PtrVector_size(&client->rigidBodies))
  {
  DROP_DEAD("rigid-body count mismatch");
  }
for(int rb=0;rb<rigidBodyCount;++rb)
  {
  struct _NatnetRigidBody *body=(struct _NatnetRigidBody *)
                                PtrVector_get(&client->rigidBodies,rb);
  int id=Natnet_readI32(&ptr);
  if(body->id!=id)
    {
    DROP_DEAD("rigid-body id mismatch");
    }
  body->tx=Natnet_readR32(&ptr);
  body->ty=Natnet_readR32(&ptr);
  body->tz=Natnet_readR32(&ptr);
  body->qx=Natnet_readR32(&ptr);
  body->qy=Natnet_readR32(&ptr);
  body->qz=Natnet_readR32(&ptr);
  body->qw=Natnet_readR32(&ptr);
#if defined SHOW_MESSAGES
  Platform_msg("  rigid-body %d (%g,%g,%g) (%g,%g,%g,%g)\n",
               body->id,body->tx,body->ty,body->tz,
               body->qx,body->qy,body->qz,body->qw);
#endif
  int markerCount=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
  for(int m=0;m<markerCount;++m)
    {
    real32_t x=Natnet_readR32(&ptr);
    real32_t y=Natnet_readR32(&ptr);
    real32_t z=Natnet_readR32(&ptr);
    Platform_msg("    (%g,%g,%g)\n",x,y,z);
    }
#else
  ptr+=markerCount*3*sizeof(real32_t);
#endif
  if(Natnet_minVersion(client,2,0))
    {
#if defined SHOW_MESSAGES
    for(int m=0;m<markerCount;++m)
      {
      int id=Natnet_readI32(&ptr);
      Platform_msg("    id %d\n",id);
      }
    for(int m=0;m<markerCount;++m)
      {
      real32_t sz=Natnet_readR32(&ptr);
      Platform_msg("    size %g\n",sz);
      }
#else
    ptr+=markerCount*(sizeof(int32_t)+sizeof(real32_t));
#endif
    if(Natnet_minVersion(client,2,6))
      {
#if defined SHOW_MESSAGES
      int16_t params=Natnet_readI16(&ptr);
      Platform_msg("    params 0x%hx\n",params);
#else
      ptr+=sizeof(int16_t);
#endif
      }
#if defined SHOW_MESSAGES
    real32_t meanMarkerError=Natnet_readR32(&ptr);
    Platform_msg("    mean marker error %g\n",meanMarkerError);
#else
    ptr+=sizeof(real32_t);
#endif
    }
  }
if(Natnet_minVersion(client,2,1))
  {
  int skeletonCount=Natnet_readI32(&ptr);
  if(skeletonCount!=PtrVector_size(&client->skeletons))
    {
    DROP_DEAD("skeleton count mismatch");
    }
  for(int sk=0;sk<skeletonCount;++sk)
    {
    struct _NatnetSkeleton *skeleton=(struct _NatnetSkeleton *)
                                     PtrVector_get(&client->skeletons,sk);
    int id=Natnet_readI32(&ptr);
    if(skeleton->id!=id)
      {
      DROP_DEAD("skeleton id mismatch");
      }
#if defined SHOW_MESSAGES
    Platform_msg("  skeleton %d\n",skeleton->id);
#endif
    int rigidBodyCount=Natnet_readI32(&ptr);
    if(rigidBodyCount!=PtrVector_size(&skeleton->rigidBodies))
      {
      DROP_DEAD("skeleton rigid-body count mismatch");
      }
    for(int rb=0;rb<rigidBodyCount;++rb)
      {
      struct _NatnetRigidBody *body=(struct _NatnetRigidBody *)
                                    PtrVector_get(&skeleton->rigidBodies,rb);
      int id=Natnet_readI32(&ptr);
      if(body->id+327680!=id)
        {
        Platform_msg("rb=%d  rbid=%d id=%d\n",rb,body->id,id);
        DROP_DEAD("skeleton rigid-body id mismatch");
        }
      body->tx=Natnet_readR32(&ptr);
      body->ty=Natnet_readR32(&ptr);
      body->tz=Natnet_readR32(&ptr);
      body->qx=Natnet_readR32(&ptr);
      body->qy=Natnet_readR32(&ptr);
      body->qz=Natnet_readR32(&ptr);
      body->qw=Natnet_readR32(&ptr);
#if defined SHOW_MESSAGES
      Platform_msg("    rigid-body %d (%g,%g,%g) (%g,%g,%g,%g)\n",
                   body->id,body->tx,body->ty,body->tz,
                   body->qx,body->qy,body->qz,body->qw);
#endif
      int markerCount=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
      for(int m=0;m<markerCount;++m)
        {
        real32_t x=Natnet_readR32(&ptr);
        real32_t y=Natnet_readR32(&ptr);
        real32_t z=Natnet_readR32(&ptr);
        Platform_msg("      (%g,%g,%g)\n",x,y,z);
        }
      for(int m=0;m<markerCount;++m)
        {
        int id=Natnet_readI32(&ptr);
        Platform_msg("    id %d\n",id);
        }
      for(int m=0;m<markerCount;++m)
        {
        real32_t sz=Natnet_readR32(&ptr);
        Platform_msg("    size %g\n",sz);
        }
      real32_t meanMarkerError=Natnet_readR32(&ptr);
      Platform_msg("    mean marker error %g\n",meanMarkerError);
#else
      ptr+=markerCount*(4*sizeof(real32_t)+sizeof(int32_t))+
           sizeof(real32_t);
#endif
      }
    }
  }
if(Natnet_minVersion(client,2,3))
  {
  int labeledMarkerCount=Natnet_readI32(&ptr);
#if defined SHOW_MESSAGES
  for(int m=0;m<labeledMarkerCount;++m)
    {
    int id=Natnet_readI32(&ptr);
    real32_t x=Natnet_readR32(&ptr);
    real32_t y=Natnet_readR32(&ptr);
    real32_t z=Natnet_readR32(&ptr);
    real32_t size=Natnet_readR32(&ptr);
    int16_t params=(int16_t)(Natnet_minVersion(client,2,6)
                             ? Natnet_readI16(&ptr) : 0);
    Platform_msg("  labeled marker %d (%g,%g,%g) %g 0x%hx\n",
                 id,x,y,z,size,params);
    }
#else
  ptr+=labeledMarkerCount*(sizeof(int32_t)+4*sizeof(real32_t));
  if(Natnet_minVersion(client,2,6))
    {
    ptr+=labeledMarkerCount*sizeof(int16_t);
    }
#endif
  }
#if defined SHOW_MESSAGES
  real32_t latency=Natnet_readR32(&ptr);
  Platform_msg("  latency %g\n",latency);
  int timecode=Natnet_readI32(&ptr);
  int timecodesub=Natnet_readI32(&ptr);
  Platform_msg("  timecode %d.%d\n",timecode,timecodesub);
  real32_t timestamp=Natnet_readR32(&ptr);
  Platform_msg("  timestamp %g\n",timestamp);
  int16_t params=Natnet_readI16(&ptr);
  Platform_msg("  params 0x%hx\n",params);
  int endOfData=Natnet_readI32(&ptr);
  Platform_msg("  end of data %d\n",endOfData);
#else
  ptr+=2*sizeof(real32_t)+3*sizeof(int32_t)+sizeof(int16_t);
#endif
}

static
void
Natnet_analyse(NatnetClient *client,
               const uint8_t *buffer,
               int msgType,
               int msgSize)
{
switch(msgType)
  {
  case NATNET_PING:
    {
#if defined SHOW_MESSAGES
    const char *txt=msgSize ? (const char *)buffer : "";
    Platform_msg("NATNET_PING [%s]\n",txt);
#endif
    break;
    }
  case NATNET_PING_RESPONSE:
    {
    Natnet_analysePingResponse(client,msgSize,buffer);
    break;
    }
  case NATNET_REQUEST:
    {
#if defined SHOW_MESSAGES
    const char *txt=msgSize ? (const char *)buffer : "";
    Platform_msg("NATNET_REQUEST [%s]\n",txt);
#endif
    break;
    }
  case NATNET_RESPONSE:
    {
#if defined SHOW_MESSAGES
    const char *txt=msgSize ? (const char *)buffer : "";
    Platform_msg("NATNET_RESPONSE [%s]\n",txt);
#endif
    break;
    }
  case NATNET_REQUEST_MODEL_DEF:
    {
#if defined SHOW_MESSAGES
    Platform_msg("NATNET_REQUEST_MODEL_DEF\n");
#endif
    break;
    }
  case NATNET_MODEL_DEF:
    {
    Natnet_analyseModelDef(client,msgSize,buffer);
    break;
    }
  case NATNET_REQUEST_FRAME_OF_DATA:
    {
#if defined SHOW_MESSAGES
    Platform_msg("NATNET_REQUEST_FRAME_OF_DATA\n");
#endif
    break;
    }
  case NATNET_FRAME_OF_DATA:
    {
    Natnet_analyseFrameOfData(client,msgSize,buffer);
    break;
    }
  case NATNET_MESSAGE_STRING:
    {
#if defined SHOW_MESSAGES
    const char *txt=msgSize ? (const char *)buffer : "";
    Platform_msg("NATNET_MESSAGE_STRING [%s]\n",txt);
#endif
    break;
    }
  case NATNET_UNRECOGNIZED_REQUEST:
    {
#if defined SHOW_MESSAGES
    Platform_msg("NATNET_UNRECOGNIZED_REQUEST\n");
#endif
    break;
    }
  default:
    {
#if defined SHOW_MESSAGES
    Platform_msg("unknown NATNET message\n");
#endif
    break;
    }
  }
}

static
int // -1: nothing available  0: ignored  1: message received
Natnet_recv(NatnetClient *client,
            uint8_t *buffer,
            int *msgType,
            int *msgSize,
            int64_t usecTimeout)
{
SocketHandle rdSet[2]={client->sock,INVALID_SOCKET};
if(Socket_select(rdSet,(SocketHandle *)0,usecTimeout)==SOCKET_ERROR)
  {
  SYS_DROP_DEAD("Socket_select() failure");
  }
if(!Socket_inSet(client->sock,rdSet))
  {
  return -1;
  }
Ipv4Address address;
uint16_t port;
int r=Socket_recvFromIpv4(client->sock,buffer,
                          NATNET_MAX_PACKET_SIZE,0,&address,&port);
if(r==SOCKET_ERROR)
  {
  // SYS_DROP_DEAD("Socket_recvFromIpv4() failure");
  // FIXME: under windows recv() fails when server is not running
  return -1;
  }
if((address!=client->address)||
   ((port!=NATNET_COMMAND_PORT)&&(port!=NATNET_DATA_PORT))||
   (r<4))
  {
  return 0;
  }
const uint8_t *ptr=buffer;
*msgType=Natnet_readI16(&ptr);
*msgSize=Natnet_readI16(&ptr);
return r>=(4+*msgSize) ? 1 : 0;
}

static
bool // UDP datagram received
Natnet_recvAndAnalyse(NatnetClient *client,
                      int64_t usecTimeout)
{
int msgType,msgSize;
int r=Natnet_recv(client,client->buffer,&msgType,&msgSize,usecTimeout);
if(r==-1)
  {
  return false; // nothing available
  }
while(r!=1) // wait for first not ignored message
  {
  r=Natnet_recv(client,client->buffer,&msgType,&msgSize,0);
  if(r==-1)
    {
    return false; // nothing available
    }
  }
int nextType,nextSize;
bool useNext=false;
if(msgType==NATNET_FRAME_OF_DATA)
  {
  for(;;) // check for multiple occurences of NATNET_FRAME_OF_DATA
    {
    r=Natnet_recv(client,client->nextBuffer,&nextType,&nextSize,0);
    if(r==-1)
      {
      break; // nothing more available
      }
    if(r==1)
      {
      if(nextType==msgType) // ignore preceding occurence
        {
        uint8_t *tmp=client->buffer;
        client->buffer=client->nextBuffer;
        client->nextBuffer=tmp;
        msgType=nextType;
        msgSize=nextSize;
        }
      else
        {
        useNext=true;
        break;
        }
      }
    }
  }
Natnet_analyse(client,client->buffer+4,msgType,msgSize);
if(useNext)
  {
  Natnet_analyse(client,client->nextBuffer+4,nextType,nextSize);
  }
return true;
}

void
NatnetClient_init(NatnetClient *client,
                  const char *server)
{
client->address=(Ipv4Address)0;
client->sock=INVALID_SOCKET;
client->version=0;
client->nextBuffer=(uint8_t *)0;
client->buffer=(uint8_t *)0;
SpatialBasis_init(&client->calibrationOffset);
PtrVector_init(&client->rigidBodies);
PtrVector_init(&client->skeletons);
client->address=Socket_ipv4Address(server);
if(!client->address)
  {
  Platform_msg("unknown IPV4 address for NETNAT server '%s'\n",server);
  return;
  }
client->sock=Socket_open(PF_INET,SOCK_DGRAM,0);
if(client->sock==INVALID_SOCKET)
  {
  SYS_DROP_DEAD("Socket_open() failure");
  }
#if 0
  if(!Socket_setReuseAddrOption(client->sock))
    {
    SYS_DROP_DEAD("Socket_setReuseAddrOption() failure");
    }
  if(!Socket_bindIpv4(client->sock,INADDR_ANY,NATNET_DATA_PORT))
    {
    SYS_DROP_DEAD("Socket_bindIpv4() failure");
    }
#endif
if(!Socket_setRcvBufOption(client->sock,NATNET_MAX_PACKET_SIZE))
  {
  SYS_DROP_DEAD("Socket_setRcvBufOption() failure");
  }
client->buffer=Memory_alloc(NATNET_MAX_PACKET_SIZE,uint8_t);
client->nextBuffer=Memory_alloc(NATNET_MAX_PACKET_SIZE,uint8_t);
int64_t timeout=500000;
int64_t lastTime=-1;
int attempts=0;
while(!client->version)
  {
  int64_t now=Platform_time();
  if(now-lastTime>timeout)
    {
    if(++attempts>3)
      {
      Platform_msg("cannot detect NATNET server '%s'\n",server);
      return;
      }
    Natnet_send(client,NATNET_PING);
    lastTime=now;
    }
  Natnet_recvAndAnalyse(client,timeout-(now-lastTime));
  }
lastTime=-1;
attempts=0;
while(!PtrVector_size(&client->rigidBodies)&&
      !PtrVector_size(&client->skeletons))
  {
  int64_t now=Platform_time();
  if(now-lastTime>timeout)
    {
    if(++attempts>3)
      {
      Platform_msg("cannot obtain model from NATNET server '%s'\n",server);
      return;
      }
    Natnet_send(client,NATNET_REQUEST_MODEL_DEF);
    lastTime=now;
    }
  Natnet_recvAndAnalyse(client,timeout-(now-lastTime));
  }
//Natnet_send(client,NATNET_REQUEST_FRAME_OF_DATA);
}

void
NatnetClient_destroy(NatnetClient *client)
{
for(intptr_t i=0,i_end=PtrVector_size(&client->rigidBodies);i<i_end;++i)
  {
  struct _NatnetRigidBody *body=(struct _NatnetRigidBody *)
                                PtrVector_get(&client->rigidBodies,i);
  NatnetRigidBody_delete(body);
  }
PtrVector_destroy(&client->rigidBodies);
for(intptr_t i=0,i_end=PtrVector_size(&client->skeletons);i<i_end;++i)
  {
  struct _NatnetSkeleton *skeleton=(struct _NatnetSkeleton *)
                                   PtrVector_get(&client->skeletons,i);
  NatnetSkeleton_delete(skeleton);
  }
PtrVector_destroy(&client->skeletons);
SpatialBasis_destroy(&client->calibrationOffset);
if(client->nextBuffer)
  {
  Memory_free(client->nextBuffer);
  }
if(client->buffer)
  {
  Memory_free(client->buffer);
  }
if((client->sock!=INVALID_SOCKET)&&!Socket_close(client->sock))
  {
  SYS_DROP_DEAD("Socket_close() failure");
  }
}

int // rigid-body index or -1
NatnetClient_findRigidBody(const NatnetClient *client,
                           const char *name)
{
for(int i=0,i_end=NatnetClient_rigidBodyCount(client);i<i_end;++i)
  {
  if(!strcmp(name,NatnetClient_rigidBodyName(client,i)))
    {
    return i;
    }
  }
return -1;
}

int // skeleton index or -1
NatnetClient_findSkeleton(const NatnetClient *client,
                          const char *name)
{
for(int i=0,i_end=NatnetClient_skeletonCount(client);i<i_end;++i)
  {
  if(!strcmp(name,NatnetClient_skeletonName(client,i)))
    {
    return i;
    }
  }
return -1;
}

int32_t // skeleton rigid-body index or -1
NatnetClient_findSkeletonRigidBody(const NatnetClient *client,
                                   int skeleton,
                                   const char *name)
{
for(int i=0,i_end=NatnetClient_skeletonRigidBodyCount(client,skeleton);
    i<i_end;++i)
  {
  if(!strcmp(name,NatnetClient_skeletonRigidBodyName(client,skeleton,i)))
    {
    return i;
    }
  }
return -1;
}

static
void
Natnet_location(const NatnetClient *client,
                const struct _NatnetRigidBody* body,
                SpatialBasis *out_location)
{
// switch axis:
//   x --> -y
//   y --> +z
//   z --> -x
real64_t qx=-body->qz,qy=-body->qx,qz=+body->qy,qw=body->qw;
real64_t r=qx*qx+qy*qy+qz*qz+qw*qw;
if(r)
  {
  r=1.0/sqrt(r);
  qx*=r; qy*=r; qz*=r; qw*=r;
  }
else
  {
  qx=0.0; qy=0.0; qz=0.0; qw=1.0;
  }
real64_t qx2=qx*qx;
real64_t qy2=qy*qy;
real64_t qz2=qz*qz;
real64_t qw2=qw*qw;
real64_t qxy=qx*qy;
real64_t qyz=qy*qz;
real64_t qzx=qz*qx;
real64_t qxw=qx*qw;
real64_t qyw=qy*qw;
real64_t qzw=qz*qw;
SpatialBasis location;
real64_t *m=location.matrix;
m[MTX_R00]=qw2+qx2-qy2-qz2;
m[MTX_R10]=2.0*(qxy+qzw);
m[MTX_R20]=2.0*(qzx-qyw);
m[MTX_Z0]=0.0;
m[MTX_R01]=2.0*(qxy-qzw);
m[MTX_R11]=qw2-qx2+qy2-qz2;
m[MTX_R21]=2.0*(qyz+qxw);
m[MTX_Z1]=0.0;
m[MTX_R02]=2.0*(qzx+qyw);
m[MTX_R12]=2.0*(qyz-qxw);
m[MTX_R22]=qw2-qx2-qy2+qz2;
m[MTX_Z2]=0.0;
m[MTX_TX]=-body->tz;
m[MTX_TY]=-body->tx;
m[MTX_TZ]=+body->ty;
m[MTX_ONE]=1.0;
SpatialBasis_setMatrix(out_location,client->calibrationOffset.matrix);
SpatialBasis_move(out_location,&location);
SpatialBasis_destroy(&location);
}

void
NatnetClient_rigidBodyLocation(const NatnetClient *client,
                               int rigidBody,
                               SpatialBasis *out_location)
{
const struct _NatnetRigidBody *rb=
  (const struct _NatnetRigidBody *)
  PtrVector_get(&client->rigidBodies,rigidBody);
Natnet_location(client,rb,out_location);
}

void
NatnetClient_skeletonRigidBodyLocation(const NatnetClient *client,
                                       int skeleton,
                                       int rigidBody,
                                       SpatialBasis *out_location)
{
const struct _NatnetSkeleton *sk=
  (const struct _NatnetSkeleton *)
  PtrVector_get(&client->skeletons,skeleton);
const struct _NatnetRigidBody *rb=
  (const struct _NatnetRigidBody *)
  PtrVector_get(&sk->rigidBodies,rigidBody);
Natnet_location(client,rb,out_location);
}

void
NatnetClient_update(NatnetClient *client)
{
while(Natnet_recvAndAnalyse(client,0));
}

//----------------------------------------------------------------------------
