//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/platform.h"

extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
#if defined _WIN32
  extern TRANSPROG_API_VAR struct _PlatformW32Fncts _Platform_w32Fncts;
#endif

typedef struct _CpuGroup
{
int numaNode;
int cacheSize[ASSUMED_MAX_CACHE_LEVEL];
int cacheLineSize[ASSUMED_MAX_CACHE_LEVEL];
int subCount;
struct _CpuGroup **sub;
int cpuCount;
uint8_t *cpuMask;
int height;
int index;
} CpuGroup;

static
CpuGroup *
CpuGroup_new(void)
{
CpuGroup *grp=(CpuGroup *)malloc(sizeof(CpuGroup));
grp->numaNode=-1;
for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
  {
  grp->cacheSize[i]=0;
  grp->cacheLineSize[i]=0;
  }
grp->subCount=0;
grp->sub=(CpuGroup **)malloc((_Platform_cpuMemInfo.maxCpuId+1)*
                             sizeof(CpuGroup*));
grp->cpuCount=0;
int maskSize=(_Platform_cpuMemInfo.maxCpuId+8)>>3;
grp->cpuMask=(uint8_t *)malloc(maskSize);
memset(grp->cpuMask,0,maskSize);
grp->height=-1;
grp->index=-1;
return grp;
}

static
void
CpuGroup_delete(CpuGroup *grp)
{
for(int i=0,i_end=grp->subCount;i<i_end;++i)
  {
  CpuGroup_delete(grp->sub[i]);
  }
free(grp->sub);
free(grp->cpuMask);
free(grp);
}

static
bool // inserted
CpuGroup_setCpu(CpuGroup *grp,
                int cpuId)
{
if((!_Platform_cpuMemInfo.cpuMask)||
   (((int)strlen(_Platform_cpuMemInfo.cpuMask)>cpuId)&&
    (_Platform_cpuMemInfo.cpuMask[cpuId]!='0')))
  {
  grp->cpuMask[cpuId>>3]|=(uint8_t)(1<<(cpuId&0x07));
  ++grp->cpuCount;
  return true;
  }
else
  {
  return false;
  }
}

static
void
CpuGroup_clearCpu(CpuGroup *grp,
                  int cpuId)
{
grp->cpuMask[cpuId>>3]&=(uint8_t)~(1<<(cpuId&0x07));
--grp->cpuCount;
}

static
bool
CpuGroup_hasCpu(const CpuGroup *grp,
                int cpuId)
{
return (grp->cpuMask[cpuId>>3]&(1<<(cpuId&0x07))) ? true : false;
}

static
void
CpuGroup_addSub(CpuGroup *grp,
                CpuGroup *sub)
{
grp->sub[grp->subCount++]=sub;
}

static
void
CpuGroup_changeNuma(CpuGroup *grp,
                    int numaNode)
{
grp->numaNode=numaNode;
for(int i=0,i_end=grp->subCount;i<i_end;++i)
  {
  CpuGroup_changeNuma(grp->sub[i],numaNode);
  }
}

#if defined _WIN32
# include "platform_detect_windows.c"
#elif defined __linux
# include "platform_detect_linux.c"
#else

static
void *
CpuGroup_specificInit(void)
{ return (void *)0; }

static
void
CpuGroup_specificDestroy(void *specif)
{ (void)specif; }

static
bool
CpuGroup_cpuOnline(int cpuId,
                   void *specif)
{ (void)cpuId; (void)specif; return false; }

static
bool
CpuGroup_numaOnline(int numaId,
                    void *specif)
{ (void)numaId; (void)specif; return false; }

static
void
CpuGroup_collectNuma(CpuGroup *grp,
                     void *specif)
{ (void)grp; (void)specif; }

static
int
CpuGroup_maxCacheLevel(int cpuId,
                       void *specif)
{ (void)cpuId; (void)specif; return -1; }

static
int
CpuGroup_cacheSize(int cpuId,
                   int level,
                   void *specif)
{ (void)cpuId; (void)level; (void)specif; return 0; }

static
int
CpuGroup_cacheLineSize(int cpuId,
                       int level,
                       void *specif)
{ (void)cpuId; (void)level; (void)specif; return 0; }

static
void
CpuGroup_sharedCache(CpuGroup *grp,
                     int cpuId,
                     int level,
                     void *specif)
{ (void)grp; (void)cpuId; (void)level; (void)specif; }

#endif

static
void
CpuGroup_subDivide(CpuGroup *grp,
                   void *specif)
{
for(int c=0,c_last=_Platform_cpuMemInfo.maxCpuId;c<=c_last;++c)
  {
  if(!CpuGroup_hasCpu(grp,c))
    {
    continue;
    }
  int level=-1;
  for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
    {
    if(grp->cacheLineSize[i])
      {
      level=i+1;
      break;
      }
    }
  if(level==-1)
    {
    level=CpuGroup_maxCacheLevel(c,specif);
    if(level>ASSUMED_MAX_CACHE_LEVEL)
      {
      char buffer[0x100];
      sprintf(buffer,"ASSUMED_MAX_CACHE_LEVEL should be at least %d\n",
                     level);
      DROP_DEAD(buffer);
      }
    if(level>_Platform_cpuMemInfo.maxCacheLevel)
      {
      _Platform_cpuMemInfo.maxCacheLevel=level;
      }
    }
  else if(level>1)
    {
    --level;
    }
  else
    {
    continue;
    }
  CpuGroup *found=(CpuGroup *)0;
  for(int i=0,i_end=grp->subCount;i<i_end;++i)
    {
    if(CpuGroup_hasCpu(grp->sub[i],c))
      {
      found=grp->sub[i];
      break;
      }
    }
  if(!found)
    {
    CpuGroup *sub=CpuGroup_new();
    sub->numaNode=grp->numaNode;
    for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
      {
      sub->cacheSize[i]=grp->cacheSize[i];
      sub->cacheLineSize[i]=grp->cacheLineSize[i];
      }
    sub->cacheSize[level-1]=CpuGroup_cacheSize(c,level,specif);
    sub->cacheLineSize[level-1]=CpuGroup_cacheLineSize(c,level,specif);
    if(_Platform_cpuMemInfo.flattenCpu)
      {
      CpuGroup_setCpu(sub,c);
      }
    else
      {
      CpuGroup_sharedCache(sub,c,level,specif);
      }
    CpuGroup_addSub(grp,sub);
    }
  CpuGroup_clearCpu(grp,c);
  }
switch(grp->subCount)
  {
  case 0:
    {
    if(grp->cpuCount>1)
      {
      for(int c=0,c_last=_Platform_cpuMemInfo.maxCpuId;c<=c_last;++c)
        {
        if(CpuGroup_hasCpu(grp,c))
          {
          CpuGroup *sub=CpuGroup_new();
          sub->numaNode=grp->numaNode;
          for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
            {
            sub->cacheSize[i]=grp->cacheSize[i];
            sub->cacheLineSize[i]=grp->cacheLineSize[i];
            }
          CpuGroup_setCpu(sub,c);
          CpuGroup_addSub(grp,sub);
          CpuGroup_clearCpu(grp,c);
          }
        }
      }
    break;
    }
  case 1:
    {
    CpuGroup *sub=grp->sub[0];
    grp->subCount=0;
    for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
      {
      grp->cacheSize[i]=sub->cacheSize[i];
      grp->cacheLineSize[i]=sub->cacheLineSize[i];
      }
    grp->cpuCount=sub->cpuCount;
    uint8_t *tmp=grp->cpuMask;
    grp->cpuMask=sub->cpuMask;
    sub->cpuMask=tmp;
    CpuGroup_delete(sub);
    CpuGroup_subDivide(grp,specif);
    break;
    }
  default:
    {
    for(int i=0,i_end=grp->subCount;i<i_end;++i)
      {
      CpuGroup_subDivide(grp->sub[i],specif);
      }
    }
  }
}

static
void
CpuGroup_updateHeight(CpuGroup *grp)
{
for(int i=0,i_end=grp->subCount;i<i_end;++i)
  {
  CpuGroup_updateHeight(grp->sub[i]);
  if(grp->sub[i]->height>grp->height)
    {
    grp->height=grp->sub[i]->height;
    }
  }
for(int i=0,i_end=grp->subCount;i<i_end;++i) // nb: height should not be
  {                                          // unbalanced, but with a user
  while(grp->sub[i]->height<grp->height)     // provided cpu-mask this could
    {                                        // happen
    CpuGroup *sub=CpuGroup_new();
    sub->height=grp->sub[i]->height+1;
    CpuGroup_addSub(sub,grp->sub[i]);
    grp->sub[i]=sub;
    }
  }
if(!++grp->height)
  {
  ++_Platform_cpuMemInfo.cpuCount;
  }
}

static
void
CpuGroup_updateIndex(CpuGroup *grp,
                     int height)
{
if(grp->height==height)
  {
  grp->index=_Platform_cpuMemInfo.memCount++;
  }
else if(grp->height>height)
  {
  for(int i=0,i_end=grp->subCount;i<i_end;++i)
    {
    CpuGroup_updateIndex(grp->sub[i],height);
    }
  }
}

static
void
CpuGroup_dump(CpuGroup *grp)
{
struct _CpumemInfoNode *node=_Platform_cpuMemInfo.nodes+grp->index;
node->parent=-1;
node->childCount=grp->subCount;
if(grp->subCount)
  {
  node->firstChild=grp->sub[0]->index;
  node->cpuCount=0;
  for(int i=0,i_end=grp->subCount;i<i_end;++i)
    {
    CpuGroup_dump(grp->sub[i]);
    struct _CpumemInfoNode *subNode=_Platform_cpuMemInfo.nodes+
                                    grp->sub[i]->index;
    subNode->parent=grp->index;
    node->cpuCount+=subNode->cpuCount;
    }
  node->firstCpu=_Platform_cpuMemInfo.nodes[grp->sub[0]->index].firstCpu;
  node->cpuId=-1;
  }
else
  {
  node->firstChild=-1;
  node->cpuCount=1;
  node->firstCpu=grp->index;
  int id;
  for(id=0;!CpuGroup_hasCpu(grp,id);++id) {}
  node->cpuId=id;
  }
node->numa=grp->numaNode;
for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
  {
  node->cacheSize[i]=grp->cacheSize[i];
  node->cacheLineSize[i]=grp->cacheLineSize[i];
  }
}

//----------------------------------------------------------------------------

#if defined CTL_HW
  static
  bool // success
  _Platform_askHwSysctl(int hwName,
                        void *result,
                        size_t len)
  {
  int r;
  int name[2];
  name[0]=CTL_HW;
  name[1]=hwName;
  RESTART_SYSCALL(r,sysctl(name,2,result,&len,NULL,0));
  return r==0;
  }
#endif

static
int64_t
_Platform_detectMemorySize(void)
{
int64_t memorySize=0;
#if defined _SC_PHYS_PAGES && \
    defined _SC_PAGE_SIZE
  if(!memorySize)
    {
    int64_t pages,size;
    RESTART_SYSCALL(pages,sysconf(_SC_PHYS_PAGES));
    RESTART_SYSCALL(size,sysconf(_SC_PAGE_SIZE));
    if(pages>0&&size>0)
      {
      memorySize=pages*size;
      }
    }
#endif
#if defined HW_MEMSIZE
  if(!memorySize)
    {
    int64_t value;
    if(_Platform_askHwSysctl(HW_MEMSIZE,&value,sizeof(value)))
      {
      memorySize=value;
      }
    }
#endif
#if defined HW_PHYSMEM64
  if(!memorySize)
    {
    int64_t value;
    if(_Platform_askHwSysctl(HW_PHYSMEM64,&value,sizeof(value)))
      {
      memorySize=value;
      }
    }
#endif
#if defined HW_PHYSMEM
  if(!memorySize)
    {
    int value;
    if(_Platform_askHwSysctl(HW_PHYSMEM,&value,sizeof(value)))
      {
      memorySize=value;
      }
    }
#endif
#if defined _WIN32
  if(!memorySize)
    {
    MEMORYSTATUSEX status;
    status.dwLength=sizeof(status);
    if(GlobalMemoryStatusEx(&status))
      {
      memorySize=status.ullTotalPhys;
      }
    }
#endif
return memorySize;
}

static
intptr_t
_Platform_detectPageSize(void)
{
intptr_t pageSize=0;
#if !defined _SC_PAGESIZE
# if defined _SC_PAGE_SIZE
#   define _SC_PAGESIZE _SC_PAGE_SIZE
# endif
#endif
#if defined _SC_PAGESIZE
  if(!pageSize)
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_PAGESIZE));
    if(r>0)
      {
      pageSize=r;
      }
    }
#endif
#if defined _WIN32
  if(!pageSize)
    {
    SYSTEM_INFO info;
    GetSystemInfo(&info);
    pageSize=info.dwPageSize;
    }
#endif
if(!pageSize)
  {
  pageSize=4096; // default value for most systems
  }
return pageSize;
}

static
intptr_t
_Platform_detectHugePageSize(void)
{
intptr_t hugePageSize=0;
#if defined __linux__
  if(!hugePageSize)
    {
    FILE *f=fopen("/proc/meminfo","r");
    if(f)
      {
      for(;;)
        {
        char buffer[0x100];
        if(!fgets(buffer,0x100,f)) { break; }
        char unit[0x10];
        intptr_t hpsz;
        int r=sscanf(buffer," Hugepagesize: %"FPTR"d %s ",&hpsz,unit);
        if(r==1)
          {
          hugePageSize=hpsz; break;
          }
        if(r==2)
          {
          if(!strcmp(unit,"B"))       { hugePageSize=hpsz;           break; }
          else if(!strcmp(unit,"kB")) { hugePageSize=hpsz*1024;      break; }
          else if(!strcmp(unit,"MB")) { hugePageSize=hpsz*1024*1024; break; }
          else
            {
            char buffer[0x100];
            sprintf(buffer,"unknown suffix `%s' for Hugepagesize\n",
                           unit);
            DROP_DEAD(buffer);
            }
          }
        }
      fclose(f);
      }
    }
#endif
#if defined _WIN32
  if(!hugePageSize)
    {
    if(_Platform_w32Fncts.kernel32)
      {
      if(_Platform_w32Fncts.GetLargePageMinimum)
        {
        SIZE_T hpsz=(*_Platform_w32Fncts.GetLargePageMinimum)();
        if(hpsz)
          {
          HANDLE process=GetCurrentProcess();
          if(process)
            {
            HANDLE token;
            if(OpenProcessToken(process,TOKEN_ADJUST_PRIVILEGES,&token))
              {
              TOKEN_PRIVILEGES tp;
              tp.PrivilegeCount=1;
              tp.Privileges[0].Attributes=SE_PRIVILEGE_ENABLED;
              if(LookupPrivilegeValue(NULL,SE_LOCK_MEMORY_NAME,
                                      &tp.Privileges[0].Luid))
                {
                // FIXME: of course we cannot enable a privilege we don't
                //        have, but how can we obtain this privilege?
                //        (even the administrator does not have it!!!)
                if(AdjustTokenPrivileges(token,FALSE,&tp,sizeof(tp),NULL,NULL)
                   &&(Platform_lastError()==ERROR_SUCCESS))
                  {
                  hugePageSize=(intptr_t)hpsz;
                  }
                }
              }
            }
          if(!hugePageSize)
            {
            char err[0x100];
            Platform_lastErrorMsg(err,0x100);
            Platform_msg("warning: cannot use huge pages (%d bytes) --- %s\n",
                         (int)hpsz,err);
            }
          }
        }
      }
    }
#endif
return hugePageSize;
}

static
void
_Platform_detectIntelLikeCache(int *cacheSize,
                               int *cacheLineSize)
{
#if defined CALL_CPUID
  int *s1=cacheSize+0;
  int *l1=cacheLineSize+0;
  int *s2=cacheSize+1;
  int *l2=cacheLineSize+1;
  int *s3=cacheSize+2;
  int *l3=cacheLineSize+2;
  uint32_t a,b,c,d;
  CALL_CPUID(0x00000000,0,a,b,c,d);
  uint8_t buffer[0x10];
  memcpy(buffer,&b,4);
  memcpy(buffer+4,&d,4);
  memcpy(buffer+8,&c,4);
  if(!memcmp(buffer,"GenuineIntel",12))
    {
    // http://download.intel.com/design/processor/applnots/24161832.pdf
    uint32_t maxLevel=a;
    if(maxLevel>=0x00000002)
      {
      for(uint32_t i=0,iterCount=255;i<iterCount;++i)
        {
        int count=0;
        CALL_CPUID(0x00000002,0,a,b,c,d);
        iterCount=a&0x000000FF;
        if(!(a&0x80000000))
          {
          buffer[count++]=(uint8_t)((a>>8)&0x000000FF);
          buffer[count++]=(uint8_t)((a>>16)&0x000000FF);
          buffer[count++]=(uint8_t)((a>>24)&0x000000FF);
          }
        if(!(b&0x80000000))
          {
          buffer[count++]=(uint8_t)(b&0x000000FF);
          buffer[count++]=(uint8_t)((b>>8)&0x000000FF);
          buffer[count++]=(uint8_t)((b>>16)&0x000000FF);
          buffer[count++]=(uint8_t)((b>>24)&0x000000FF);
          }
        if(!(c&0x80000000))
          {
          buffer[count++]=(uint8_t)(c&0x000000FF);
          buffer[count++]=(uint8_t)((c>>8)&0x000000FF);
          buffer[count++]=(uint8_t)((c>>16)&0x000000FF);
          buffer[count++]=(uint8_t)((c>>24)&0x000000FF);
          }
        if(!(d&0x80000000))
          {
          buffer[count++]=(uint8_t)(d&0x000000FF);
          buffer[count++]=(uint8_t)((d>>8)&0x000000FF);
          buffer[count++]=(uint8_t)((d>>16)&0x000000FF);
          buffer[count++]=(uint8_t)((d>>24)&0x000000FF);
          }
        while(count--)
          {
          switch(buffer[count])
            {
#define SHOW_TLB 0
#if SHOW_TLB
            case 0x01: Platform_msg("ITLB: 4KB, 4 w, 32 e\n"); break;
            case 0x02: Platform_msg("ITLB: 4MB, full, 2 e\n"); break;
            case 0x03: Platform_msg("DTLB: 4KB, 4 w, 64 e\n"); break;
            case 0x04: Platform_msg("DTLB: 4MB, 4 w, 8 e\n"); break;
            case 0x05: Platform_msg("DTLB: 4MB, 4 w, 32 e\n"); break;
#endif
            case 0x0A: *s1=1024*8;     *l1=32;  break;
            case 0x0C: *s1=1024*16;    *l1=32;  break;
            case 0x0D: *s1=1024*16;    *l1=64;  break;
            case 0x0E: *s1=1024*24;    *l1=64;  break; // ?
            case 0x10: *s1=1024*16;    *l1=32;  break; // IA64
            case 0x1A: *s2=1024*96;    *l2=64;  break; // IA64
            case 0x21: *s2=1024*256;   *l2=64;  break;
            case 0x22: *s3=1024*512;   *l3=64;  break;
            case 0x23: *s3=1024*1024;  *l3=64;  break;
            case 0x25: *s3=1024*2048;  *l3=64;  break;
            case 0x29: *s3=1024*4096;  *l3=64;  break;
            case 0x2C: *s1=1024*32;    *l1=64;  break;
            case 0x39: *s2=1024*128;   *l2=64;  break;
            case 0x3A: *s2=1024*192;   *l2=64;  break;
            case 0x3B: *s2=1024*128;   *l2=64;  break;
            case 0x3C: *s2=1024*256;   *l2=64;  break;
            case 0x3D: *s2=1024*384;   *l2=64;  break;
            case 0x3E: *s2=1024*512;   *l2=64;  break;
            case 0x41: *s2=1024*128;   *l2=32;  break;
            case 0x42: *s2=1024*256;   *l2=32;  break;
            case 0x43: *s2=1024*512;   *l2=32;  break;
            case 0x44: *s2=1024*1024;  *l2=32;  break;
            case 0x45: *s2=1024*2048;  *l2=32;  break;
            case 0x46: *s3=1024*4096;  *l3=64;  break;
            case 0x47: *s3=1024*8192;  *l3=64;  break;
            case 0x48: *s2=1024*3072;  *l2=64;  break;
            case 0x49: *s2=1024*4096;  *l2=64;  break; // P4: L3, Core2: L2
            case 0x4A: *s3=1024*6144;  *l3=64;  break;
            case 0x4B: *s3=1024*8192;  *l3=64;  break;
            case 0x4C: *s3=1024*12288; *l3=64;  break;
            case 0x4D: *s3=1024*16384; *l3=64;  break;
            case 0x4E: *s2=1024*6144;  *l2=64;  break;
#if SHOW_TLB
            case 0x50: Platform_msg("ITLB: 4KB|2MB|4MB, full, 64 e\n"); break;
            case 0x51: Platform_msg("ITLB: 4KB|2MB|4MB, full, 128 e\n"); break;
            case 0x52: Platform_msg("ITLB: 4KB|2MB|4MB, full, 256 e\n"); break;
            case 0x55: Platform_msg("ITLB: 2MB|4MB, full, 7 e\n"); break;
            case 0x56: Platform_msg("L1DTLB: 4MB, 4 w, 16 e\n"); break;
            case 0x57: Platform_msg("L1DTLB: 4KB, 4 w, 16 e\n"); break;
            case 0x5A: Platform_msg("DTLB0: 2MB|4MB, 4 w, 32 e\n"); break;
            case 0x5B: Platform_msg("DTLB: 4KB|4MB, full, 64 e\n"); break;
            case 0x5C: Platform_msg("DTLB: 4KB|4MB, full, 128 e\n"); break;
            case 0x5D: Platform_msg("DTLB: 4KB|4MB, full, 256 e\n"); break;
#endif
            case 0x60: *s1=1024*16;    *l1=64;  break;
            case 0x66: *s1=1024*8;     *l1=64;  break;
            case 0x67: *s1=1024*16;    *l1=64;  break;
            case 0x68: *s1=1024*32;    *l1=64;  break;
            case 0x78: *s2=1024*1024;  *l2=64;  break;
            case 0x79: *s2=1024*128;   *l2=64;  break;
            case 0x7A: *s2=1024*256;   *l2=64;  break;
            case 0x7B: *s2=1024*512;   *l2=64;  break;
            case 0x7C: *s2=1024*1024;  *l2=64;  break;
            case 0x7D: *s2=1024*2048;  *l2=64;  break;
            case 0x7E: *s2=1024*256;   *l2=128; break; // IA64
            case 0x7F: *s2=1024*512;   *l2=64;  break;
            case 0x80: *s2=1024*512;   *l2=64;  break; // ?
            case 0x81: *s2=1024*128;   *l2=32;  break; // ?
            case 0x82: *s2=1024*256;   *l2=32;  break;
            case 0x83: *s2=1024*512;   *l2=32;  break;
            case 0x84: *s2=1024*1024;  *l2=32;  break;
            case 0x85: *s2=1024*2048;  *l2=32;  break;
            case 0x86: *s2=1024*512;   *l2=64;  break;
            case 0x87: *s2=1024*1024;  *l2=64;  break;
            case 0x88: *s3=1024*2048;  *l3=64;  break; // IA64
            case 0x89: *s3=1024*4096;  *l3=64;  break; // IA64
            case 0x8A: *s3=1024*8192;  *l3=64;  break; // IA64
            case 0x8D: *s3=1024*3072;  *l3=128; break; // IA64
#if SHOW_TLB
            case 0xB0: Platform_msg("ITLB: 4KB, 4 w, 128 e\n"); break;
            case 0xB1: Platform_msg("ITLB: 2MB, 4 w, 8 e\n"); break;
            case 0xB2: Platform_msg("ITLB: 4KB, 4 w, 64 e\n"); break;
            case 0xB3: Platform_msg("DTLB: 4KB, 4 w, 128 e\n"); break;
            case 0xB4: Platform_msg("DTLB: 4KB, 4 w, 256 e\n"); break;
            case 0xCA: Platform_msg("SL2TLB: 4KB, 4 w, 512 e\n"); break;
#endif
            case 0xD0: *s3=1024*512;   *l3=64;  break;
            case 0xD1: *s3=1024*1024;  *l3=64;  break;
            case 0xD2: *s3=1024*2048;  *l3=64;  break;
            case 0xD6: *s3=1024*1024;  *l3=64;  break;
            case 0xD7: *s3=1024*2048;  *l3=64;  break;
            case 0xD8: *s3=1024*4096;  *l3=64;  break;
            case 0xDC: *s3=1024*1536;  *l3=64;  break;
            case 0xDD: *s3=1024*3072;  *l3=64;  break;
            case 0xDE: *s3=1024*6144;  *l3=64;  break;
            case 0xE2: *s3=1024*2048;  *l3=64;  break;
            case 0xE3: *s3=1024*4096;  *l3=64;  break;
            case 0xE4: *s3=1024*8192;  *l3=64;  break;
            case 0xEA: *s3=1024*12288; *l3=64;  break;
            case 0xEB: *s3=1024*18432; *l3=64;  break;
            case 0xEC: *s3=1024*24576; *l3=64;  break;
            }
          }
        }
      }
    if(maxLevel>=0x00000004)
      {
      for(int cache=0;;++cache)
        {
        CALL_CPUID(0x00000004,cache,a,b,c,d);
        uint32_t type=a&0x0000001F;
        if(!type) break; // no more cache description
        if((type!=1)&&(type!=3)) continue; // not data nor unified cache
        uint32_t lvl=(a>>5)&0x00000007;
        uint32_t ways=1+((b>>22)&0x000003FF);
        uint32_t partitions=1+((b>>12)&0x000003FF);
        uint32_t line=1+(b&0x00000FFF);
        uint32_t sets=1+c;
        if(lvl)
          {
          if(lvl>ASSUMED_MAX_CACHE_LEVEL)
            {
            char buffer[0x100];
            sprintf(buffer,"ASSUMED_MAX_CACHE_LEVEL should be at least %u\n",
                           lvl);
            DROP_DEAD(buffer);
            }
          cacheSize[lvl-1]=ways*partitions*line*sets;
          cacheLineSize[lvl-1]=line;
          }
        }
      }
    }
  else if(!memcmp(buffer,"AuthenticAMD",12))
    {
    // http://www.amd.com/us-en/assets/content_type/
    //                    white_papers_and_tech_docs/25481.pdf
    CALL_CPUID(0x80000000,0,a,b,c,d);
    uint32_t maxLevel=a;
    if(maxLevel>=0x80000005)
      {
      CALL_CPUID(0x80000005,0,a,b,c,d);
      *s1=(c&0xFF000000)>>14;
      *l1=c&0x000000FF;
      }
    if(maxLevel>=0x80000006)
      {
      CALL_CPUID(0x80000006,0,a,b,c,d);
      *s2=(c&0xFFFF0000)>>6;
      *l2=c&0x000000FF;
      *s3=((uint32_t)(d&0xFC000000))<<1; // higher bit lost?
      *l3=d&0x000000FF;
      }
    }
#endif
}

static
void
_Platform_detectCache(int *cacheSize,
                      int *cacheLineSize)
{
for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i)
  {
  cacheSize[i]=0;
  cacheLineSize[i]=0;
  }
_Platform_detectIntelLikeCache(cacheSize,cacheLineSize);
#if defined _SC_LEVEL1_DCACHE_SIZE
  if(!cacheSize[0])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL1_DCACHE_SIZE));
    if(r>0)
      {
      cacheSize[0]=r;
      }
    }
#endif
#if defined _SC_LEVEL1_DCACHE_LINESIZE
  if(!cacheLineSize[0])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL1_DCACHE_LINESIZE));
    if(r>0)
      {
      cacheLineSize[0]=r;
      }
    }
#endif
#if defined _SC_LEVEL2_CACHE_SIZE
  if(!cacheSize[1])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL2_CACHE_SIZE));
    if(r>0)
      {
      cacheSize[1]=r;
      }
    }
#endif
#if defined _SC_LEVEL2_CACHE_LINESIZE
  if(!cacheLineSize[1])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL2_CACHE_LINESIZE));
    if(r>0)
      {
      cacheLineSize[1]=r;
      }
    }
#endif
#if defined _SC_LEVEL3_CACHE_SIZE
  if(!cacheSize[2])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL3_CACHE_SIZE));
    if(r>0)
      {
      cacheSize[2]=r;
      }
    }
#endif
#if defined _SC_LEVEL3_CACHE_LINESIZE
  if(!cacheLineSize[2])
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_LEVEL3_CACHE_LINESIZE));
    if(r>0)
      {
      cacheLineSize[2]=r;
      }
    }
#endif
#if defined HW_L1DCACHESIZE
  if(!cacheSize[0])
    {
    int value;
    if(_Platform_askHwSysctl(HW_L1DCACHESIZE,&value,sizeof(value)))
      {
      cacheSize[0]=value;
      }
    }
#endif
#if defined HW_L2CACHESIZE
  if(!cacheSize[1])
    {
    int value;
    if(_Platform_askHwSysctl(HW_L2CACHESIZE,&value,sizeof(value)))
      {
      cacheSize[1]=value;
      }
    }
#endif
#if defined HW_L3CACHESIZE
  if(!cacheSize[2])
    {
    int value;
    if(_Platform_askHwSysctl(HW_L3CACHESIZE,&value,sizeof(value)))
      {
      cacheSize[2]=value;
      }
    }
#endif
#if defined HW_CACHELINE
  for(int i=0;i<ASSUMED_MAX_CACHE_LEVEL;++i) // FIXME: same cache line size?
    {
    if(cacheSize[i]&&!cacheLineSize[i])
      {
      int value;
      if(_Platform_askHwSysctl(HW_CACHELINE,&value,sizeof(value)))
        {
        cacheLineSize[i]=value;
        }
      }
    }
#endif
}

static
int
_Platform_detectCpuCount(void)
{
int cpuCount=0;
#if defined _SC_NPROCESSORS_ONLN
  if(!cpuCount)
    {
    int r;
    RESTART_SYSCALL(r,sysconf(_SC_NPROCESSORS_ONLN));
    if(r>0)
      {
      cpuCount=r;
      }
    }
#endif
#if defined HW_NCPU
  if(!cpuCount)
    {
    int value;
    if(_Platform_askHwSysctl(HW_NCPU,&value,sizeof(value)))
      {
      cpuCount=value;
      }
    }
#endif
#if defined _WIN32
  if(!cpuCount)
    {
    SYSTEM_INFO info;
    GetSystemInfo(&info);
    cpuCount=info.dwNumberOfProcessors;
    }
#endif
return cpuCount;
}

void
_Platform_detectCpuMem(void)
{
const char *varName="TRANSPROG_CPUMASK";
_Platform_cpuMemInfo.cpuMask=(const char *)0;
char cpuMask[0x400];
if(Platform_environ(varName,cpuMask,0x400))
  {
  int maskLen=(int)strlen(cpuMask);
  if(maskLen)
    {
    bool goodMask=true;
    for(int i=0;i<maskLen;++i)
      {
      if((cpuMask[i]!='0')&&(cpuMask[i]!='1'))
        {
        goodMask=false;
        Platform_msg("warning: ignoring %s (%s) environment variable\n"
                     "         (should be made of 0 and 1)\n",
                     varName,cpuMask);
        break;
        }
      }
    if(goodMask)
      {
      char *s=(char *)malloc(maskLen+1);
      memcpy(s,cpuMask,maskLen+1);
      _Platform_cpuMemInfo.cpuMask=s;
      }
    }
  }
_Platform_cpuMemInfo.flattenCpu=Platform_environAsBool(
                                "TRANSPROG_FLATTEN_CPU",false);
_Platform_cpuMemInfo.attachCpu=Platform_environAsBool(
                               "TRANSPROG_ATTACH_CPU",true);
_Platform_cpuMemInfo.memSize=_Platform_detectMemorySize();
_Platform_cpuMemInfo.pageSize=_Platform_detectPageSize();
_Platform_cpuMemInfo.hugePageSize=
  Platform_environAsBool("TRANSPROG_HUGE_PAGES",true)
  ? _Platform_detectHugePageSize() : 0;
_Platform_cpuMemInfo.maxCacheLineSize=0;
_Platform_cpuMemInfo.maxCacheLevel=0;
_Platform_cpuMemInfo.maxCpuId=-1;
_Platform_cpuMemInfo.maxNumaId=-1;
_Platform_cpuMemInfo.memCount=0;
_Platform_cpuMemInfo.cpuCount=0;
_Platform_cpuMemInfo.numaCount=0;
for(int i=0;i<=ASSUMED_MAX_NUMA_COUNT;++i)
  {
  _Platform_cpuMemInfo.numaId[i]=-1;
  }
CpuGroup *grp=(CpuGroup *)0;
void *specif=CpuGroup_specificInit();
if(_Platform_cpuMemInfo.maxCpuId>=0)
  {
  grp=CpuGroup_new();
  if(_Platform_cpuMemInfo.maxNumaId>=0)
    {
    _Platform_cpuMemInfo.numaCount=0;
    for(int n=0,n_last=_Platform_cpuMemInfo.maxNumaId;n<=n_last;++n)
      {
      if(CpuGroup_numaOnline(n,specif))
        {
        if(_Platform_cpuMemInfo.flattenCpu)
          {
          CpuGroup *tmp=CpuGroup_new();
          tmp->numaNode=n;
          CpuGroup_collectNuma(tmp,specif);
          if(tmp->cpuCount)
            {
            ++_Platform_cpuMemInfo.numaCount;
            for(int c=0,c_last=_Platform_cpuMemInfo.maxCpuId;c<=c_last;++c)
              {
              if(CpuGroup_hasCpu(tmp,c))
                {
                CpuGroup *sub=CpuGroup_new();
                sub->numaNode=n;
                CpuGroup_setCpu(sub,c);
                CpuGroup_subDivide(sub,specif);
                CpuGroup_addSub(grp,sub);
                }
              }
            }
          CpuGroup_delete(tmp);
          }
        else
          {
          CpuGroup *sub=CpuGroup_new();
          sub->numaNode=n;
          CpuGroup_collectNuma(sub,specif);
          if(sub->cpuCount)
            {
            ++_Platform_cpuMemInfo.numaCount;
            CpuGroup_subDivide(sub,specif);
            CpuGroup_addSub(grp,sub);
            }
          else
            {
            CpuGroup_delete(sub);
            }
          }
        }
      }
    if(grp->subCount)
      {
      if(_Platform_cpuMemInfo.numaCount>=ASSUMED_MAX_NUMA_COUNT)
        {
        char buffer[0x100];
        sprintf(buffer,"ASSUMED_MAX_NUMA_COUNT should be at least %d\n",
                _Platform_cpuMemInfo.numaCount);
        DROP_DEAD(buffer);
        }
      grp->numaNode=_Platform_cpuMemInfo.numaCount;
      for(int i=0,i_end=grp->subCount;i<i_end;++i)
        {
        for(int j=0,j_end=_Platform_cpuMemInfo.numaCount;j<j_end;++j)
          {
          if(_Platform_cpuMemInfo.numaId[j]==-1)
            {
            _Platform_cpuMemInfo.numaId[j]=grp->sub[i]->numaNode;
            }
          if(_Platform_cpuMemInfo.numaId[j]==grp->sub[i]->numaNode)
            {
            CpuGroup_changeNuma(grp->sub[i],j);
            break;
            }
          }
        }
      }
    if(grp->subCount==1)
      {
      CpuGroup *sub=grp->sub[0];
      grp->subCount=0;
      CpuGroup_delete(grp);
      grp=sub;
      }
    }
  else
    {
    for(int c=0,c_last=_Platform_cpuMemInfo.maxCpuId;c<=c_last;++c)
      {
      if(CpuGroup_cpuOnline(c,specif))
        {
        CpuGroup_setCpu(grp,c);
        }
      }
    grp->numaNode=0;
    _Platform_cpuMemInfo.numaCount=1;
    _Platform_cpuMemInfo.numaId[0]=-1;
    CpuGroup_subDivide(grp,specif);
    }
  if(!grp->subCount&&!grp->cpuCount)
    {
    CpuGroup_delete(grp);
    grp=(CpuGroup *)0;
    }
  }
CpuGroup_specificDestroy(specif);
if(!grp)
  {
  Platform_msg("warning: cpu/cache topology not detected "
               "(assuming flat)\n");
  int defaultCacheSize[ASSUMED_MAX_CACHE_LEVEL];
  int defaultCacheLineSize[ASSUMED_MAX_CACHE_LEVEL];
  _Platform_detectCache(defaultCacheSize,defaultCacheLineSize);
  _Platform_cpuMemInfo.maxCpuId=_Platform_detectCpuCount()-1;
  _Platform_cpuMemInfo.maxNumaId=-1;
  _Platform_cpuMemInfo.numaCount=1;
  _Platform_cpuMemInfo.numaId[0]=-1;
  grp=CpuGroup_new();
  grp->numaNode=0;
  for(int i=0,i_last=_Platform_cpuMemInfo.maxCpuId;i<=i_last;++i)
    {
    CpuGroup *sub=CpuGroup_new();
    if(CpuGroup_setCpu(sub,i))
      {
      sub->numaNode=grp->numaNode;
      for(int j=0;j<ASSUMED_MAX_CACHE_LEVEL;++j)
        {
        if(defaultCacheSize[j])
          {
          sub->cacheSize[j]=defaultCacheSize[j];
          sub->cacheLineSize[j]=defaultCacheLineSize[j];
          if(j+1>_Platform_cpuMemInfo.maxCacheLevel)
            {
            _Platform_cpuMemInfo.maxCacheLevel=j+1;
            }
          }
        }
      CpuGroup_addSub(grp,sub);
      }
    else
      {
      CpuGroup_delete(sub);
      }
    }
  if(grp->subCount==1)
    {
    CpuGroup *sub=grp->sub[0];
    grp->subCount=0;
    CpuGroup_delete(grp);
    grp=sub;
    }
  if(!grp->subCount&&!grp->cpuCount)
    {
    CpuGroup_delete(grp);
    grp=(CpuGroup *)0;
    }
  }
if(!grp)
  {
  DROP_DEAD("no cpu enabled");
  }
CpuGroup_updateHeight(grp);
for(int h=0,h_last=grp->height;h<=h_last;++h)
  {
  CpuGroup_updateIndex(grp,h);
  }
if(_Platform_cpuMemInfo.memCount>ASSUMED_MAX_MEM_COUNT)
  {
  char buffer[0x100];
  sprintf(buffer,"ASSUMED_MAX_MEM_COUNT should be at least %d\n",
                 _Platform_cpuMemInfo.memCount);
  DROP_DEAD(buffer);
  }
CpuGroup_dump(grp);
CpuGroup_delete(grp);
for(int i=0,i_end=_Platform_cpuMemInfo.cpuCount;i<i_end;++i)
  {
  for(int j=0,j_end=_Platform_cpuMemInfo.maxCacheLevel;j<j_end;++j)
    {
    if(_Platform_cpuMemInfo.nodes[i].cacheLineSize[j]>
       _Platform_cpuMemInfo.maxCacheLineSize)
      {
      _Platform_cpuMemInfo.maxCacheLineSize=
        _Platform_cpuMemInfo.nodes[i].cacheLineSize[j];
      }
    }
  }
}

//----------------------------------------------------------------------------
