//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/plugin.h"

#include "TransProg/memory.h"
#include "TransProg/path.h"

#if !defined _WIN32
#  include <dlfcn.h>
#endif

static
PluginHandle
_Plugin_sysOpen(const char *pluginName)
{
#if defined _WIN32
  return pluginName ? (PluginHandle)LoadLibrary(pluginName)
                    : (PluginHandle)GetModuleHandle((LPCTSTR)0);
#else
  return (PluginHandle)dlopen(pluginName,RTLD_LAZY);
#endif
}

PluginHandle // plugin or 0
Plugin_open(const char *pluginName)
{
PluginHandle plugin=_Plugin_sysOpen(pluginName);
if(!plugin&&pluginName)
  {
  const char *base=Path_findBasename(pluginName);
  if(base)
    {
    const char *ext=Path_findExtension(base);
    bool startsWithLib=!strncasecmp("lib",base,3);
    if(!ext||!startsWithLib)
      {
#if defined _WIN32
      const char *stdExt=".dll";
#elif defined __APPLE__
      const char *stdExt=".bundle";
#else
      const char *stdExt=".so";
#endif
      int dirLength=(int)(base-pluginName);
      int baseLength=(int)strlen(base);
      char *convertedName=Memory_alloc(dirLength+baseLength+0x20,char);
      memcpy(convertedName,pluginName,dirLength);
      if(!ext)
        {
        sprintf(convertedName+dirLength,"%s%s",base,stdExt);
        plugin=_Plugin_sysOpen(convertedName);
        }
      if(!plugin&&!startsWithLib)
        {
        if(ext)
          {
          sprintf(convertedName+dirLength,"lib%s",base);
          }
        else
          {
          sprintf(convertedName+dirLength,"lib%s%s",base,stdExt);
          }
        plugin=_Plugin_sysOpen(convertedName);
        }
      Memory_free(convertedName);
      }
    }
  }
return plugin;
}

void * // symbol or 0
Plugin_symbol(PluginHandle plugin,
              const char *symbolName)
{
void *symbol;
#if defined _WIN32
  FARPROC procAddr=GetProcAddress((HMODULE)plugin,symbolName);
  memcpy(&symbol,&procAddr,sizeof(void *));
#else
  symbol=dlsym(plugin,symbolName);
  if(!symbol) // FIXME: is this really a good idea?
    {
    char buffer[0x100];                      // under some platforms,
    snprintf(buffer,0x100,"_%s",symbolName); // symbol names are added
    buffer[0xFF]='\0';                       // a leading underscore
    symbol=dlsym(plugin,buffer);
    }
#endif
return symbol;
}

void
Plugin_lastErrorMsg(char *out_msg,
                    int msgSize)
{
#if defined _WIN32
  Platform_lastErrorMsg(out_msg,msgSize);
#else
  const char *msg=dlerror();
  snprintf(out_msg,msgSize,"%s",msg ? msg : "plugin error");
  out_msg[msgSize-1]='\0';
#endif
}

void
Plugin_close(PluginHandle plugin)
{
#if defined _WIN32
  FreeLibrary((HMODULE)plugin);
#else
  dlclose(plugin);
#endif
}

//----------------------------------------------------------------------------
