//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/scheduler.h"
#include "TransProg/random.h"

TRANSPROG_INLINE
void
_Activity_insert(Activity *a,
                 struct _SchedulerActivities *to,
                 int cpuCtx)
{
#if 0
  SpinSynchro_lockW(&to->sync);
  intptr_t last=PtrVector_size(&to->act);
  PtrVector_push(&to->act,a);
  // FIXME: it does randomise on long term but not enough on short term
  // nb: randomise only the ends to avoid touching many cache lines
  if(Random_int(cpuCtx,2))
    {
    void **data=PtrVector_data(&to->act);
    data[last]=data[0];
    data[0]=a;
    }
  SpinSynchro_unlockW(&to->sync);
#else
  SpinSynchro_lockW(&to->sync);
  intptr_t last=PtrVector_size(&to->act);
  PtrVector_setSize(&to->act,last+1);
  intptr_t pos=Random_intptr(cpuCtx,last+1);
  void **data=PtrVector_data(&to->act);
  data[last]=data[pos];
  data[pos]=a;
  SpinSynchro_unlockW(&to->sync);
#endif
}

void
Activity_init(Activity *a,
              ActivityFunction fnct,
              Scheduler *s,
              int cpuCtx)
{
a->fnct=fnct;
a->cpuCtx=(NarrowCpuCtx)cpuCtx;
a->retryCount=0;
a->skipCount=0;
struct _SchedulerActivities *to=s->ctx[a->cpuCtx].activities+
                                Scheduler_toState(s);
_Activity_insert(a,to,a->cpuCtx);
}

void
Activity_initCopy(Activity *a,
                  Scheduler *s,
                  const Activity *ref)
{
a->fnct=ref->fnct;
a->cpuCtx=ref->cpuCtx;
a->retryCount=ref->retryCount;
a->skipCount=ref->skipCount;
struct _SchedulerActivities *to=s->ctx[a->cpuCtx].activities+
                                Scheduler_toState(s);
_Activity_insert(a,to,a->cpuCtx);
}

void
Activity_destroy(Activity *a,
                 Scheduler *s)
{
(void)a; // nothing to be done
(void)s;
}

//----------------------------------------------------------------------------

void
Scheduler_init(Scheduler *s)
{
s->step=0;
for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
  {
  struct _SchedulerContext *c=s->ctx+cpu;
  c->bestCtx=cpu;
  c->bestAffinity=0;
  memset(c->affinity,0,sizeof(c->affinity));
  memset(c->numaUsage,0,sizeof(c->numaUsage));
  for(int i=0;i<2;++i)
    {
    SpinSynchro_init(&c->activities[i].sync);
    PtrVector_init(&c->activities[i].act);
    }
  }
s->workStealing=true;
s->prefetch=1;
}

void
Scheduler_destroy(Scheduler *s)
{
for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
  {
  struct _SchedulerContext *c=s->ctx+cpu;
  for(int i=0;i<2;++i)
    {
    SpinSynchro_destroy(&c->activities[i].sync);
    PtrVector_destroy(&c->activities[i].act); // FIXME: what if not empty?
    }
  }
}

void
Scheduler_activate(int cpuCtx,
                   void *scheduler)
{
Scheduler *s=(Scheduler *)scheduler;
int fromState=Scheduler_fromState(s);
int toState=1-fromState;
intptr_t prefetch=s->prefetch>0 ? s->prefetch : INTPTR_MAX;
struct _SchedulerContext *contexts=s->ctx;
struct _SchedulerContext *c=contexts+cpuCtx;
const NarrowCpuCtx *stealMap=Platform_cpuRoundTrip(cpuCtx);
struct _SchedulerActivities *from=c->activities+fromState;
int cpuCtxCount=Platform_cpuCtxCount();
for(;;)
  {
  Activity *a=(Activity *)0;
  SpinSynchro_lockW(&from->sync);
  intptr_t sz=PtrVector_size(&from->act);
  if(sz--)
    {
    a=(Activity *)PtrVector_pop(&from->act);
    if(sz>=prefetch)
      {
      Platform_prefetch(PtrVector_get(&from->act,sz-prefetch));
      }
    }
  SpinSynchro_unlockW(&from->sync);
  if(!a)
    {
    intptr_t stolen=0;
    if(s->workStealing)
      {
      for(int m=1;m<cpuCtxCount;++m)
        {
        struct _SchedulerActivities *src=contexts[stealMap[m]].activities+
                                         fromState;
        SpinSynchro_lockW(&src->sync);
        intptr_t count=PtrVector_size(&src->act);
        stolen=count>>1; // FIXME: why this amount? (it just works well)
        if(stolen)
          {
          SpinSynchro_lockW(&from->sync);
          PtrVector_pushMany(&from->act,
                             PtrVector_data(&src->act)+count-stolen,stolen);
          PtrVector_popMany(&src->act,stolen);
          SpinSynchro_unlockW(&from->sync);
          }
        SpinSynchro_unlockW(&src->sync);
        if(stolen)
          {
          break;
          }
        }
      }
    if(!stolen)
      {
      break;
      }
    }
  else
    {
    c->bestCtx=a->cpuCtx=(NarrowCpuCtx)cpuCtx;
    c->bestAffinity=0;
    for(int i=0;i<cpuCtxCount;++i)
      {
      c->affinity[i]=0;
      }
    switch((*a->fnct)(a,s))
      {
      case ACT_DEAD:
        {
        break;
        }
      case ACT_RETRY:
        {
        if(a->retryCount==(1<<(8*sizeof(a->retryCount)))-1)
          {
          DROP_DEAD("too many retry attempts");
          }
        ++a->retryCount;
        _Activity_insert(a,from,cpuCtx);
        break;
        }
      case ACT_SKIP:
        {
        a->retryCount=0;
        if(a->skipCount==(1<<(8*sizeof(a->skipCount)))-1)
          {
          DROP_DEAD("too many skipped steps");
          }
        ++a->skipCount;
        struct _SchedulerActivities *to2=s->ctx[c->bestCtx].activities+
                                         toState;
        _Activity_insert(a,to2,cpuCtx);
        break;
        }
      case ACT_DONE:
        {
        a->retryCount=0;
        if(a->skipCount) // steps were skipped --> reactivate in same step
          {
          --a->skipCount;
          _Activity_insert(a,from,cpuCtx);
          }
        else
          {
          struct _SchedulerActivities *to2=s->ctx[c->bestCtx].activities+
                                           toState;
          _Activity_insert(a,to2,cpuCtx);
          }
        break;
        }
      }
    }
  }
}

//----------------------------------------------------------------------------
