//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/synchro.h"

void
SystemSynchro_init(SystemSynchro *s)
{
#if defined _WIN32
  InitializeCriticalSection(&s->lock);
  extern TRANSPROG_API_VAR struct _PlatformW32Fncts _Platform_w32Fncts;
  if(_Platform_w32Fncts.InitializeConditionVariable&&
     _Platform_w32Fncts.SleepConditionVariableCS&&
     _Platform_w32Fncts.WakeAllConditionVariable)
    {
    InitializeConditionVariable(&s->cond);
    s->waiting=(HANDLE *)0;
    }
  else
    {
    // nb: s->cond not initialised
    s->waiting=Memory_alloc(Platform_cpuCtxCount(),HANDLE);
    }
  s->waitingCount=0;
#else
  int r=pthread_mutex_init(&s->lock,(pthread_mutexattr_t *)0);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_mutex_init() failure",r);
    }
  r=pthread_cond_init(&s->cond,(pthread_condattr_t *)0);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_cond_init() failure",r);
    }
#endif
}

SystemSynchro *
SystemSynchro_new(void)
{
SystemSynchro *s=Memory_alloc(1,SystemSynchro);
SystemSynchro_init(s);
return s;
}

void
SystemSynchro_destroy(SystemSynchro *s)
{
#if defined _WIN32
  DeleteCriticalSection(&s->lock);
  if(s->waiting)
    {
    Memory_free(s->waiting);
    }
  else
    {
    // DeleteConditionVariable(&s->cond);
    }
#else
  int r=pthread_cond_destroy(&s->cond);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_cond_destroy() failure",r);
    }
  r=pthread_mutex_destroy(&s->lock);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_mutex_destroy() failure",r);
    }
#endif
}

void
SystemSynchro_delete(SystemSynchro *s)
{
SystemSynchro_destroy(s);
Memory_free(s);
}

bool // success
SystemSynchro_tryEnter(SystemSynchro *s)
{
#if defined _WIN32
  return TryEnterCriticalSection(&s->lock) ? true : false;
#else
  int r=pthread_mutex_trylock(&s->lock);
  switch(r)
    {
    case 0: return true;
    case EBUSY: return false;
    default: SYS_DROP_DEAD2("pthread_mutex_trylock() failure",r); return false;
    }
#endif
}

void
SystemSynchro_enter(SystemSynchro *s)
{
#if defined _WIN32
  EnterCriticalSection(&s->lock);
#else
  int r=pthread_mutex_lock(&s->lock);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_mutex_lock() failure",r);
    }
#endif
}

void
SystemSynchro_leave(SystemSynchro *s)
{
#if defined _WIN32
  LeaveCriticalSection(&s->lock);
#else
  int r=pthread_mutex_unlock(&s->lock);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_mutex_unlock() failure",r);
    }
#endif
}

void
SystemSynchro_wait(SystemSynchro *s)
{
#if defined _WIN32
  if(s->waiting)
    {
    HANDLE _Platform_suspendEvent(int);
    HANDLE evt=_Platform_suspendEvent(Platform_cpuCtx());
    s->waiting[s->waitingCount++]=evt;
    LeaveCriticalSection(&s->lock);
    if(WaitForSingleObject(evt,INFINITE)!=WAIT_OBJECT_0)
      {
      SYS_DROP_DEAD("WaitForSingleObject() failure");
      }
    EnterCriticalSection(&s->lock);
    }
  else
    {
    if(!SleepConditionVariableCS(&s->cond,&s->lock,INFINITE))
      {
      SYS_DROP_DEAD("SleepConditionVariableCS() failure");
      }
    }
#else
  int r=pthread_cond_wait(&s->cond,&s->lock);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_cond_wait() failure",r);
    }
#endif
}

void
SystemSynchro_broadcast(SystemSynchro *s)
{
#if defined _WIN32
  if(s->waiting)
    {
    for(int i=0,i_end=s->waitingCount;i<i_end;++i)
      {
      if(!SetEvent(s->waiting[i]))
        {
        SYS_DROP_DEAD("SetEvent() failure");
        }
      }
    s->waitingCount=0;
    }
  else
    {
    WakeAllConditionVariable(&s->cond);
    }
#else
  int r=pthread_cond_broadcast(&s->cond);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_cond_broadcast() failure",r);
    }
#endif
}

//----------------------------------------------------------------------------
